/**************************************************************************
 *  linux/arch/armnommu/mach-dm270/dma.c
 *
 *    Copyright (C) 2003 Cadenux.com <http://www.cadenux.com>
 *    Gregory Nutt <greg.nutt@cadenux.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **************************************************************************/

/**************************************************************************
 * Included Files
 **************************************************************************/

#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/mman.h>
#include <linux/init.h>

#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/hardware.h>

#include <asm/mach/dma.h>

/**************************************************************************
 * Definitions
 **************************************************************************/

#define DM270_REFCTL_EXTMEM1 0x0000 /* SDRAM<->External Memory I/F */
#define DM270_REFCTL_SERIAL1 0x2000 /* SDRAM<->Serial I/F */
#define DM270_REFCTL_USB1    0x4000 /* SDRAM<->USB controller */
#define DM270_REFCTL_MMC1    0x6000 /* SDRAM<->MMC/SD controller */
#define DM270_REFCTL_DSP1    0x8000 /* SDRAM<->DSP (HPIB) I/F */

#define DM270_REFCTL_MASK1   0xe000

#define DM270_REFCTL_EXTMEM2 0x0000 /* SDRAM<->External Memory I/F */
#define DM270_REFCTL_SERIAL2 0x0400 /* SDRAM<->Serial I/F */
#define DM270_REFCTL_USB2    0x0800 /* SDRAM<->USB controller */
#define DM270_REFCTL_MMC2    0x0c00 /* SDRAM<->MMC/SD controller */
#define DM270_REFCTL_DSP2    0x1000 /* SDRAM<->DSP (HPIB) I/F */
#define DM270_REFCTL_MASK2   0x1e00

/**************************************************************************
 * Private Function Prototypes
 **************************************************************************/

static void dm270_enable_dma1(dmach_t channel, dma_t *dma);
static void dm270_disable_dma1(dmach_t channel, dma_t *dma);
static void dm270_enable_dma2(dmach_t channel, dma_t *dma);
static void dm270_disable_dma2(dmach_t channel, dma_t *dma);

/**************************************************************************
 * Private Data
 **************************************************************************/

static struct dma_ops dm270_dma_ops1 =
{
  type:           "DMA1",
  enable:         dm270_enable_dma1,
  disable:        dm270_disable_dma1,
};

static struct dma_ops dm270_dma_ops2 =
{
  type:           "DMA2",
  enable:         dm270_enable_dma2,
  disable:        dm270_disable_dma2,
};

/**************************************************************************
 * Private Functions
 **************************************************************************/

static void dm270_enable_dma1(dmach_t channel, dma_t *dma)
{
  dmamode_t dma_endpoint = dma->dma_mode & DMA_ENDPOINT_MASK;
  __u16 setting;

  switch (dma_endpoint)
    {
    case DMA_MODE_EXTMEM: /* SDRAM<->External Memory I/F */
      setting = DM270_REFCTL_EXTMEM1;
      break;

    case DMA_MODE_SERIAL: /* SDRAM<->Serial I/F */
      setting = DM270_REFCTL_SERIAL1;
      break;

    case DMA_MODE_USB: /* SDRAM<->USB controller */
      setting = DM270_REFCTL_USB1;
      break;

    case DMA_MODE_MMCSD: /* SDRAM<->MMC/SD controller */
      setting = DM270_REFCTL_MMC1;
      break;

    default:
      printk(KERN_ERR, "Unrecognized DMA endpoint");
      return;
    }

  setting |= (inw( DM270_SDRAMC_REFCTL) & ~DM270_REFCTL_MASK1);
  outw(setting,  DM270_SDRAMC_REFCTL);
}

static void dm270_disable_dma1(dmach_t channel, dma_t *dma)
{
  __u16 setting;

  setting = (inw(DM270_SDRAMC_REFCTL) & ~DM270_REFCTL_MASK1);
  outw(setting, DM270_SDRAMC_REFCTL);
}

static void dm270_enable_dma2(dmach_t channel, dma_t *dma)
{
  dmamode_t dma_endpoint = dma->dma_mode & DMA_ENDPOINT_MASK;
  __u16 setting;

  switch (dma_endpoint)
    {
    case DMA_MODE_EXTMEM: /* SDRAM<->External Memory I/F */
      setting = DM270_REFCTL_EXTMEM2;
      break;

    case DMA_MODE_SERIAL: /* SDRAM<->Serial I/F */
      setting = DM270_REFCTL_SERIAL2;
      break;

    case DMA_MODE_USB: /* SDRAM<->USB controller */
      setting = DM270_REFCTL_USB2;
      break;

    case DMA_MODE_MMCSD: /* SDRAM<->MMC/SD controller */
      setting = DM270_REFCTL_MMC2;
      break;

    default:
      printk(KERN_ERR, "Unrecognized DMA endpoint");
      return;
    }

  setting |= (inw(DM270_SDRAMC_REFCTL) & ~DM270_REFCTL_MASK2);
  outw(setting, DM270_SDRAMC_REFCTL);
}

static void dm270_disable_dma2(dmach_t channel, dma_t *dma)
{
  __u16 setting;

  setting = (inw(DM270_SDRAMC_REFCTL) & ~DM270_REFCTL_MASK2);
  outw(setting, DM270_SDRAMC_REFCTL);
}

/**************************************************************************
 * Public Functions
 **************************************************************************/

void __init arch_dma_init(dma_t *dma)
{
  /* The DM270 supports two DMA channels */

  dma[0].d_ops   = &dm270_dma_ops1;
  dma[1].d_ops   = &dm270_dma_ops2;
}
