/*
 * linux/arch/armnommu/mach-dm270/time.c
 *   Timer functions for the DM270.
 *         
 *    Copyright (C) 2003 Cadenux.com <http://www.cadenux.com>
 *    Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived from linux/arch/armnommu/mach-dsc25/time.c
 * Derived from linux/arch/armnommu/mach-dsc21/setup.c
 *
 *    Copyright (C) 2001 RidgeRun, Inc.
 *    Author: RidgeRun, Inc.  stevej@ridgerun.com
 *         
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/config.h>
#include <linux/time.h>
#include <linux/timex.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <asm/hardware.h>

#define TIMER_COUNTS_UP 1

/* This is the high precision clock used for interrupt latency
 * calculations.
 */

#if defined(CONFIG_ILATENCY) || defined(CONFIG_PREEMPT_TIMES)
#include <asm/irq.h> /* For IRQ_TIMER */

#ifndef IRQ_TIMER
# error "IRQ_TIMER is not defined"
#endif

unsigned long ilatency_jiffies       = 0;
static unsigned long dm270_last_time = 0;

static inline int dm270_timer_pending(void)
{
  return ((inw(DM270_IRQ1_STATUS) & 1) == 0);
}

unsigned long dm270_readtimer(void)
{
  unsigned long tmp;
  unsigned long elapsed;
  unsigned long working_jiffies;
  int pending;
  int flags;

  /* Keep time stable until we are satisfied with our time
   * calculation.
   */

  __save_flags(flags); __intr_cli();
  working_jiffies = ilatency_jiffies;

  /* Resample the time count register until we are are certain
   * that no timer interrupt expired between them.
   */

#ifdef TIMER_COUNTS_UP
  /* Compute the elapsed count. The current count tells us how
   * many counts have elapsed since the last interrupt (COUNT UP)
   */

  do
    {
      tmp     = inw(DM270_TIMER0_CNT);
      pending = dm270_timer_pending();
      elapsed = inw(DM270_TIMER0_CNT);
    }
  while (elapsed < tmp);

#else
  /* Compute the elapsed count. The current count tells us how
   * many counts remain until the next interrupt (COUNTS DOWN).
   * CLOCK_DIVISOR tells us how many counts total occur between
   * interrupts. Subtract to get the counts since the last interrupt.
   */

  do
    {
      tmp     = inw(DM270_TIMER0_CNT);
      pending = dm270_timer_pending();
      elapsed = inw(DM270_TIMER0_CNT);
    }
  while (elapsed > tmp);

  elapsed = CLOCK_DIVISOR - elapsed;

#endif

  /* If there is a pending timer interrupt, then bump the working
   * jiffie counter.  When interrupts are re-enabled, this value will
   * be added to ilatency_jiffie counter.
   */

  if (pending)
    {
      working_jiffies++;
    }

  /* Add the ticks for the elapsed jiffy count.  Note, we
   * have lots of issues here:  missed jiffy counts, race
   * conditions with timer expiration and the count sample
   * time, etc.  We are going to gloss over these for now
   * and assume everything will work out in the long run.
   */

  elapsed += working_jiffies * CLOCK_DIVISOR;

  /* Just one sanity check:  Make sure that time is ever
   * incrementing. The timer interrupt is often lost -- during
   * boot and when the system load is heavy.  The ilatency_jiffies
   * will assure that the time returned by this function is monotonic
   * increasing.  The effective jiffies here can differ considerably
   * from system jiffies!
   */

  if (dm270_last_time <= (0xffffffff - CLOCK_DIVISOR))
    {
      while (elapsed < dm270_last_time)
	{
	  elapsed += CLOCK_DIVISOR;
	  ilatency_jiffies++;
	}
    }

  /* We are out of here */

  dm270_last_time = elapsed;
  __intr_restore_flags(flags);
  return elapsed;
}
#endif

unsigned long dm270_gettimeoffset(void)
{
  unsigned long elapsed;

#ifdef TIMER_COUNTS_UP
  /* Compute the elapsed count. The current count tells us how
   * many counts have elapsed since the last interrupt (COUNT UP)
   */

  elapsed = inw(DM270_TIMER0_CNT);

#else
  /* Compute the elapsed count. The current count tells us how
   * many counts remain until the next interrupt (COUNTS DOWN).
   * CLOCK_DIVISOR tells us how many counts total occur between
   * interrupts. Subtract to get the counts since the last interrupt.
   */

  elapsed = CLOCK_DIVISOR - inw(DM270_TIMER0_CNT);

#endif

  /* Convert the elapsed count to usecs. I guess there are 'tick' usecs
   * between every interrupt. 
   */

  return (unsigned long)((elapsed * tick) / CLOCK_DIVISOR);
}

void dm270_timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  do_timer(regs);
}

