/************************************************************************
 * linux/drivers/char/dsp_dm270.c
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived from linux/drivers/char/dsp_dsc25.c
 *
 *   Copyright(c) 2002, Cadenux, LLC
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

/* Functional differences between the ARM / DSP Communication (HPIB)
 *
 * DM270                            DSC21
 * o Only HPI 16-bit mode available o Both HPI 16-bit and 8-bit modes can
 *   for communication.               be used for communication
 * o DSP uses bit 0 of 0xffff (the  o DSP interrupts ARM using the HINT
 *   memory mapped port) to           pin of the HPIBCTL register
 *   interrupt the ARM.
 * o No handshaking possible        o Handshaking possible in HPI-i mode
 * o The HPIB connection width on   o ARM - HPIB connection is 16-bit
 *   both the ARM and DSP side is     DSP - DSP connection is 8 or 16-bit
 *   16-bit
 */

/* DSP Boot Sequence
 *
 * The ARM should reset the DSP via the DSP Controller register bit. When
 * done, the ARM should take the DSP out of reset. At this stage the DSP
 * begins execution at 0xFF80, a hard coded address in ROM, which places
 * the DSP into IDLE 1 mode (a power down mode). This gives the ARM the
 * opportunity to download the DSP code by using the HPI-16 interface.
 * After it completes downloading the DSP image, the ARM can send an
 * interrupt to the DSP, which wakes it up from IDLE 1 mode and jump 
 * to address 0x7F8 0 where it starts running its application code loaded
 * by the ARM.  
 *
 * On the `C54x DSP, the reset vector is located at address 0x0FF80.
 * Within the DM270, this memory location has been ROM'ed with a small
 * bootloader. It is achieved in 3 steps:
 *
 * 1. ARM RELEASES RESET:
 *    a. The DSP goes out of reset and load its program counter (PC)
 *       register with 0x0FF80.
 *    b. The ROM code found at this location branches the DSP to address
 *       0x0F800 where the initialization routine resides. 
 *    c. At address 0x0F800, the DSP status register PMST is initialized
 *       to mode the vector table to 0x07F80, the interrupts are disabled
 *       except for INT0 and finally, the DSP is sent to a power down mode
 *       (Idle 1).
 * 2. ARM LOADS CODE:  
 *    a. While in this mode, the ARM loads the DSP internal 32k words of
 *       DARAM memory with the DSP code.
 * 3. ARM RE-STARTS DSP
 *    a. When the ARM is done downloading the DSP code, it wakes it up
 *       from IDLE mode by asserting INT0. 
 *    b. The DSP then branches to address 0x7F80 where the new interrupt
 *       vector table is located. The ARM should have loaded this location
 *       with at least a branch to the start of your code. 
 */

/************************************************************************
 * Included Files
 ************************************************************************/

/* Use versioning if needed */

#include <linux/autoconf.h>

#ifdef CONFIG_MODVERSIONS
# undef MODVERSIONS
# define MODVERSIONS
#endif

#ifdef MODVERSIONS
#  include <linux/modversions.h>
#endif

#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/ioport.h>

#include <asm/types.h>
#include <asm/semaphore.h>
#include <asm/segment.h>

#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <asm/arch/dsp.h>
#include <asm/arch/dsp_ioctl.h>
#include <asm/arch/sysdep.h>

/************************************************************************
 * Definitions
 ************************************************************************/

/* #undef DSP_DEBUG */
/* #define DSP_DEBUG */

#ifdef DSP_DEBUG
# define dbg(format, arg...) \
  printk(KERN_DEBUG __FUNCTION__ ": " format "\n" , ##arg)
#else
# define dbg(x...)
#endif

#define info(format, arg...) \
  printk(KERN_INFO __FUNCTION__ ": " format "\n" , ##arg)
#define warn(format, arg...) \
  printk(KERN_WARNING __FUNCTION__ ": " format "\n" , ##arg)
#define err(format, arg...) \
  printk(KERN_ERR __FUNCTION__ ": " format "\n" , ##arg)

/* DSP Signals Corresponding to Bits in HPIBCTL and/or HPISTAT 
 * BIO     Branch control. A branch can  be conditionally executed
 *         when BIO is active. If low, the processor executes the
 *         conditional instruction. For the XC instruction, the BIO
 *         condition is sampled during the decode phase of the 
 *         pipeline; all other instructions sample BIO during the read phase
 *         of the pipeline. 
 * HOLDN   HOLD is asserted to request control of the address, data, and
 *         control lines. When acknowledged by the 'C54x, these lines go
 *         into the high-impedance state. 
 * HOLDA   Hold acknowledge. HOLDA indicates that the '5409 is in a hold
 *         state and that the address, data,  and control lines are in the
 *         high-impedance state, allowing the external memory interface to be  
 *         accessed by other devices. HOLDA also goes into the high-impedance
 *         state when OFF is low.  
 * RSN     Reset.  RS causes the DSP to terminate  execution and causes a 
 *         reinitialization of the CPU and peripherals. When RS is brought
 *         to a high level, execution begins at location 0FF80h of program  
 *         memory. RS affects various registers and status bits. 
 * INT0    External user interrupt. INT0 is prioritized and maskable by the
 *         interrupt mask register and the interrupt mode bit. INT0 can be
 *         polled and reset by way of the interrupt flag register. 
 * NMI     Nonmaskable interrupt. NMI is an external interrupt that cannot be
 *         masked by way of the INTM or the IMR. When  NMI is activated, the
 *         processor traps to the appropriate vector location.  
 * HPIENA  HPIENA must be driven high during reset to enable the HPI. An
 *         internal pulldown resistor is always active and the HPIENA pin
 *         is sampled on the rising edge of RS. If HPIENA is left open or
 *         is driven low during reset, the HPI module is disabled. Once the
 *         HPI is disabled, the HPIENA pin has no effect until the '5409 is
 *         reset.
 * XF      External flag output (latched software -programmable signal). XF
 *         is set high by the SSBX XF instructio n, set low by the RSBX XF
 *         instruction or by loading ST1. XF is used for signaling other
 *         processors in multiprocessor configurations or used as a
 *         general-purpose output pin. XF goes into the high-impedance state
 *         when OFF is low, and is set high at reset.  
 * HSTINT  Host Port Interrupt.  Status bit that shows the state of the DSP
 *         I/O port corresponding to bit zero at address 0xFFFF.  This I/O
 *         port is internally tied to the ARM's interrupt controller, and
 *         is used to allow the DSP to interrupt the ARM.
 */
 

/* HPIBCTL: HPIB Control Register ***************************************
 *
 * HPI Control Register Bit Summary 
 *
 * Bit    Name    Reset     R/W    Function
 *                Value
 * 15:11  RSV                      Reserved
 * 10     DBIO    1         R/W    DSP BIO Control
 *                                   0: BIO pin asserted
 *                                   1: Normal
 *                                 DBIO controls the state of the DSP's BIOn
 *                                 pin.  When the DSP is the BIO state, it
 *                                 can branch and control the program.
 *  9     DHOLD   1         R/W    DSP Hold Request Control
 *                                   0: Hold
 *                                   1: Normal
 *                                 DHOLD controls the state of the DSP's HOLD
 *                                 pin. 
 *  8     DRST    1         R/W    DSP Reset
 *                                   0: DSP reset
 *                                   1: Normal
 *                                 DRST controls the state of the DSP's RSN
 *                                 pin, and is used to reset the DSP.  
 *  7     DINT0   1         R/W    DSP INT0 Control (ARM to DSP interrupt)
 *                                   0: INT0
 *                                   1: Normal
 *                                 DINT0 controls the state of the DSP's INT0
 *                                 pin, and is used interrupt the DSP from
 *                                 ARM.  INT0 is generated in the DSP when
 *                                 ARM sets this bit to "0".  After the
 *                                 interrupt is generated, ARM must return
 *                                 this bit to "1".
 *  6     RSV                      Reserved
 *  5     EXCHG   0         R/W    Byte Exchange
 *                                   0: No exchange
 *                                   1: Exchange upder 8-bits of the HPI16
 *                                      data bus with the 
 *                                      lower 8-bits and exchange the lower
 *                                      8-bits of the HPI16 data bus with
 *                                      the uper 8-bits.  
 *                                 The exchange is done when data is
 *                                 transfered in both directions ARM to DSP
 *                                 and DSP to ARM (write and read). 
 *  4     RSV                      Reserved
 *  3     HPNMI   1         R/W    HPI NMI Control (NMI Interrupt)
 *                                   0: NMI
 *                                   1: Normal
 *                                 HPNMI controls the state of the DSP's
 *                                 NMIn pin, and is used interrupt the DSP
 *                                 from the ARM.  NMI is generated in the DSP 
 *                                 when ARM sets this bit to "0".  After the
 *                                 interrupt is generated, ARM must return
 *                                 this bit to "1".
 *  2:1   RSV                      Reserved
 *  0     HPIEN   1         R/W    Enables HPI Control
 *                                   0: Disabled
 *                                  1: Enabled
 *                                 HPIEN controls the DSP's HPIEN pin.  When
 *                                 HPIEN is disabled, all HPI functions are
 *                                 invalid.
 *
 * After writing a '0' to bits 3, 7, or 8 of HPIBCTL (0x00), it is very
 * important to write a '1' back into the bits.  Otherwise, the DSP will
 * keep getting interrupted or will stay in a reset mode. 
 */

#define DM270_HPIB_DBIO    0x0400
#define DM270_HPIB_DHOLD   0x0200
#define DM270_HPIB_DRST    0x0100
#define DM270_HPIB_DINT0   0x0080
#define DM270_HPIB_EXCHG   0x0020
#define DM270_HPIB_HPNMI   0x0008
#define DM270_HPIB_HPIEN   0x0001

/* This is the initial value of the DSC24_HPIBCTL register: */

#define DM270_HPIBCTL_INIT 0x0789

/* HPIBSTAT:  HPIB Status Register
 *
 * HPI Status Register Bit Summary 
 *
 * Bit    Name    Reset     R/W    Function
 *                Value
 * 15:13  RSV                      Reserved
 * 12     DXF     1         R      DSP XF Status
 *                                 Indicates status of DSP XF signal output
 * 11:9   RSV                      Reserved
 *  8     HOLDA   1         R      DSP Hold Acknowledge
 *                                   0: DSP hold status
 *                                   1: Normal
 *                                 Indicates status of the DSP's HOLDA
 *                                 signal output. After a hold request is
 *                                 sent to the DSP by DHOLD in HPIBCLT, 
 *                                 the acknowledgement can be checked by
 *                                 this bit.
 * 7:1    RSV                      Reserved
 * 0      HINT    1          R     Host Port Interrupt Status
 *                                   0: DSP sent interrupt request to host
 *                                      (INT11)
 *                                   1: Normal
 * Indicates status of DSP HINT signal output from DSP to host.       
 */

#define DM270_HPIB_DXF   0x1000
#define DM270_HPIB_HOLDA 0x0100
#define DM270_HPIB_HINT  0x0001

/* Macros to manage access to dsp_drvr_s */

#define IS_OPEN(d)              (((d)->opn_count)>=0)
#define IS_INBOX_WAITING(d)     waitqueue_active(&(d)->rd_waitq)
#define IS_OUTBOX_WAITING(d)    waitqueue_active(&(d)->wr_waitq)
#define IS_READ_ERROR(d)        _MB_ERROR((d)->inbox)
#define IS_WRITE_COMPLETE(d)    _MB_COMPLETE((d)->outbox)
#define IS_INBOX_NONEMPTY(d)    _MB_INBOX_NONEMPTY((d)->inbox)
#define IS_OUTBOX_NONFULL(d)    _MB_OUTBOX_NONFULL((d)->outbox)

/* Delay values in jiffies */

#define RST_JIFFY_DELAY1         5
#define RST_JIFFY_DELAY2        10

#define INT0_JIFFY_DELAY1        5
#define INT0_JIFFY_DELAY2       10

/************************************************************************
 * Private Type Declarations
 ************************************************************************/

/* The current state of the DSP interface.  This state information is
 * shared between all instances of the DSP driver. Interrupts must be
 * disabled whenever this structure is accesses.
 */

enum dsp_state_e
{
  DSPSTATE_RESET = 0,           /* State of the DSP during a RESET */
  DSPSTATE_IDLE,                /* State of the DSP after a RESET */
  DSPSTATE_RUNNING              /* State of the DSP after started */
};

struct dsp_state_s
{
  unsigned short     hpibctl;   /* Current state of the HPIBCTL register */
  enum dsp_state_e   state;     /* (See dsp_state_e) */
};

/* Driver state information. */

struct dsp_drvr_s
{
#ifdef DECLARE_WAITQUEUE
  wait_queue_head_t  rd_waitq;  /* Used to wake up blocked reads */
  wait_queue_head_t  wr_waitq;  /* Used to wake up blocked writes */
#else
  struct wait_queue *rd_waitq;  /* Used to wake up blocked reads */
  struct wait_queue *wr_waitq;  /* Used to wake up blocked writes */
#endif
  int                opn_count; /* Number of open's on this minor. */
  int                rd_error;  /* Non-zero if read error reported */
  struct dsp_mb_s   *inbox;     /* Incoming (read) mailbox */
  struct dsp_mb_s   *outbox;    /* Outgoing (write) mailbox */
};

/************************************************************************
 * Local Function Prototypes
 ************************************************************************/

static void
dsp_delay(int jiffy_wait);

static int
dsp_load_code(struct dsp_load_s *load);

static int
dsp_memclr(struct dsp_memclr_s *memclr);

static int
dsp_mbox(struct dsp_mbspecifier_s *mbspec, int minor);

static int
dsp_ioctl(struct inode *inode, struct file *filep, unsigned int cmd,
	  unsigned long arg);

static int
dsp_open(struct inode *inode, struct file *filep);

#ifdef LINUX_20
static void
dsp_release (struct inode *inode, struct file *filep);
#else
static int
dsp_release (struct inode *inode, struct file *filep);
#endif /* LINUX_20 */

#ifdef LINUX_20
static int
dsp_llseek(struct inode *inode, struct file *file, off_t offset, int org);
#endif /* LINUX_20 */

#ifdef LINUX_20
static int
dsp_read(struct inode *inode, struct file *filep, char *buf, int count);
#else
static ssize_t
dsp_read(struct file *filep, char *buf, size_t count, loff_t *f_pos);
#endif /* LINUX_20 */
        
#ifdef LINUX_20
static int
dsp_write(struct inode *inode, struct file *file, const char *buf, int count);
#else /* LINUX_20 */
static ssize_t
dsp_write(struct file *filep, const char *buf, size_t count, loff_t *ppos);
#endif /* LINUX_20 */

static void
dsp_interrupt(int irq, void *dev_id, struct pt_regs *regs);

#ifdef MODULE
struct file_operations *dsp_get_fileops(void);
#endif

/************************************************************************
 * Private Constant Data
 ************************************************************************/

static const char myname[] = "dsp";

/************************************************************************
 * Private Variables
 ************************************************************************/

/* The current state of the DSP interface.  This state information is
 * shared between all instances of the DSP driver. Interrupts must be
 * disabled whenever this structure is accesses.
 */

static struct dsp_state_s dsp_state;

/* Current state of all of the DSP devices */

static struct dsp_drvr_s dsp_drvr[DSP_MINOR_MAX];

/* Device File Operations */

static struct file_operations dsp_fileops =
{
#ifdef LINUX_20
  lseek:     dsp_llseek,        /* lseek */
#else /* LINUX_20 */
  owner:     THIS_MODULE,
  llseek:    no_llseek,         /* llseek */
#endif /* LINUX_20 */
  ioctl:     dsp_ioctl,         /* ioctl */
  open:      dsp_open,          /* open code */
  release:   dsp_release,       /* close code */
  read:      dsp_read,          /* read */
  write:     dsp_write,         /* write */
};

/* Give modules access to the DSP ioctl calls without having to
 * search for the DSP module.
 */

#ifdef MODULE
#ifdef LINUX_20
static struct symbol_table dsp_symbol_table =
{
# include <linux/symtab_begin.h>
  X(dsp_get_fileops),
# include <linux/symtab_end.h>
};
#else
EXPORT_SYMBOL(dsp_get_fileops);
#endif
#endif

/************************************************************************
 * Inline functions
 ************************************************************************/

static inline void
dsp_memcpy(unsigned short *to, unsigned short *from, int nwords)
{
  for (; nwords > 0; nwords--, to++, from++)
    {
      outw(inw((unsigned long)from), (unsigned long)to);
    }
}

static inline void
dsp_memset(unsigned short *to, unsigned short value, int nwords)
{
  for (; nwords > 0; nwords--, to++)
    {
      outw(value, (unsigned long)to);
    }
}

static inline int
dsp_verify(unsigned short *addr1, unsigned short *addr2, int nwords)
{
  for (; nwords > 0; nwords--, addr1++, addr2++)
    {
      if (inw((unsigned long)addr1) != inw((unsigned long)addr2))
	return -EREMOTEIO;
    }
  return 0;
}



static inline int
dsp_force_idle(int cpu)
{
  dsp_state.hpibctl &= ~DM270_HPIB_DRST;
  dsp_state.state    = DSPSTATE_RESET;
  outw(dsp_state.hpibctl, DM270_HPIBCTL);
  dsp_delay(RST_JIFFY_DELAY1);

  dsp_state.hpibctl |= DM270_HPIB_DRST;
  outw(dsp_state.hpibctl, DM270_HPIBCTL);
  dsp_delay(RST_JIFFY_DELAY2);
  dsp_state.state    = DSPSTATE_IDLE;

  return 0;
}

static inline int
dsp_interrupt_dsp(int cpu)
{
  dsp_state.hpibctl &= ~DM270_HPIB_DINT0;
  outw(dsp_state.hpibctl, DM270_HPIBCTL);
  dsp_delay(INT0_JIFFY_DELAY1);

  dsp_state.hpibctl |= DM270_HPIB_DINT0;
  outw(dsp_state.hpibctl, DM270_HPIBCTL);
  dsp_delay(INT0_JIFFY_DELAY2);
  dsp_state.state    = DSPSTATE_RUNNING;

  return 0;
}

#define dsp_start(cpu) dsp_interrupt_dsp(cpu)

/************************************************************************
 * Private Functions
 ************************************************************************/

/************************************************************************
 * dsp_delay
 ************************************************************************/

static void
dsp_delay(int jiffy_wait)
{
  unsigned long start_time = jiffies;

  /* Wait until the elapsed time exceeds the requested delay */

  while ((jiffies - start_time) < jiffy_wait)
    {
      /* sys_sched_yield? */
    }
}

/************************************************************************
 * dsp_load_code
 ************************************************************************/

static int
dsp_load_code(struct dsp_load_s *load)
{
  unsigned long to = (unsigned long)load->to;

  /* Verify that the source address is properly 16-bit aligned. */

  if (((((unsigned long)load->from) & 1) != 0) || ((load->count & 1) != 0))
    {
      dbg("Source buffer not aligned");
      return -EINVAL;
    }

  /* Verify the the "to" address does, indeed, like within the DSP
   * address space.
   */

  else if (!DSPMEM_DSP_ADDR(to, load->cpu))
    {
      warn("Address 0x%lx doe not like in the DSP address space", to); 
      return -EINVAL;
    }

  /* Verify that the DSP is idle and ready for program load. */

  else if (dsp_state.state != DSPSTATE_IDLE)
    {
      dbg("DSP state is %d", dsp_state.state);
      return -EBUSY;
    }

  else
    {
      /* Convert the DSP address to an ARM address */

      to = DSPMEM_DSP_TO_ARM(to, load->cpu);

      /* And copy the data to the DSP */

      dsp_memcpy((unsigned short*)to, load->from, load->count >> 1);

      /* Verify that the copy was successful */

      return dsp_verify((unsigned short*)to, load->from, load->count >> 1);
    }
  return 0;
}

/************************************************************************
 * dsp_memclr
 ************************************************************************/

static int
dsp_memclr(struct dsp_memclr_s *memclr)
{
  unsigned long addr = (unsigned long)memclr->addr;

  /* Verify that the count is properly 16-bit aligned. */

  if ((memclr->count & 1) != 0)
    {
      dbg("Byte count not aligned");
      return -EINVAL;
    }

  /* Verify the the address does, indeed, like within the DSP
   * address space.
   */

  else if (!DSPMEM_DSP_ADDR(addr, memclr->cpu))
    {
      warn("Address 0x%lx doe not like in the DSP address space", addr); 
      return -EINVAL;
    }

  /* Verify that the DSP is idle and ready for program load. */

  else if (dsp_state.state != DSPSTATE_IDLE)
    {
      dbg("DSP state is %d", dsp_state.state);
      return -EBUSY;
    }

  else
    {
      /* Convert the DSP address to an ARM address */

      addr = DSPMEM_DSP_TO_ARM(addr, memclr->cpu);

      /* And memset the DSP memory to zero */

      dsp_memset((unsigned short*)addr, 0, memclr->count >> 1);
    }
  return 0;
}

/************************************************************************
 * dsp_mbox
 ************************************************************************/

static int
dsp_mbox(struct dsp_mbspecifier_s *mbspec, int minor)
{
  unsigned long      addr = (unsigned long)mbspec->addr;
  struct dsp_drvr_s *drvr = &dsp_drvr[minor];
  struct dsp_mb_s   *mbox;
  int                flags;

  /* Verify the the address does, indeed, like within the DSP
   * address space.
   */

  if (!DSPMEM_DSP_ADDR(addr, mbspec->cpu))
    {
      warn("Address 0x%lx doe not like in the DSP address space", addr); 
      return -EINVAL;
    }

  else if ((unsigned)minor >= DSP_MINOR_MAX)
    {
      dbg("Bad minor (minor) %d", minor);
      return -EACCES;
    }

  else
    {
      /* Convert the DSP address to an ARM address */

      mbox = (struct dsp_mb_s*)DSPMEM_DSP_TO_ARM(addr, mbspec->cpu);

      /* Keep things stationary while we set this up */

      save_flags(flags); cli();

      /* Add the mailbox reference to the the driver state info. Note
       * that if a mailbox was previously declared, that old mailbox
       * is abandoned with no special actions.
       */

      if (mbspec->wr_dir)
	{
	  /* Its a new outbox. Verify that no user is waiting the
	   * current outbox.
	   */

	  if (IS_OUTBOX_WAITING(drvr))
	    {
	      dbg("Outbox is busy");
	      restore_flags(flags);
	      return -EBUSY;
	    }
	  else
	    {
	      drvr->outbox = mbox;
	    }
	}
      else
	{
	  /* Its a new inbox. Verify that no user is waiting the
	   * current outbox.
	   */

	  if (IS_INBOX_WAITING(drvr))
	    {
	      dbg("Inbox is busy");
	      restore_flags(flags);
	      return -EBUSY;
	    }
	  else
	    {
	      drvr->inbox  = mbox;
	    }
	}

      /* Initialize the mailbox */

      memset(mbox, 0, sizeof(struct dsp_mb_s));
      mbox->state =  DSPMB_READY;
      mbox->size  =  mbspec->size;
      restore_flags(flags);
    }
  return 0;
}

/************************************************************************
 * dsp_ioctl
 ************************************************************************/

static int
dsp_ioctl (struct inode *inode, struct file *filep,
	   unsigned int cmd, unsigned long arg)
{
  int            minor    = MINOR(inode->i_rdev);
  signed int     size     = _IOC_SIZE(cmd);
  int            err      = 0;

  dbg("cmd=0x%x minor=%d arg=0x%lx", cmd, minor, arg);
    
  /* Range check the IOCTL cmd */

  if ((_IOC_TYPE(cmd) != DSP_MAGIC_NUMBER) ||
      (_IOC_NR(cmd) > DSP_IOC_MAXNR)) 
    {
      dbg("Bad ioctl value");
      return -EINVAL;
    }

  /* Check address validity */

  if (_IOC_DIR(cmd) & _IOC_READ)
    {
      err = verify_area(VERIFY_READ, (void*)arg, size);

      if (err)
	{
	  dbg("verify_area(VERIFY_READ) error:%d",err);
	}
    } 
  else if (_IOC_DIR(cmd) & _IOC_WRITE) 
    {
      err = verify_area(VERIFY_WRITE, (void*)arg, size);
          
      if (err)
	{
	  dbg("verify_area(VERIFY_WRITE) error:%d",err);
	}
    }

  if (err) 
    {
      return err;
    }

  /* Command processing */

  switch (cmd)
    {
      /* IOCTLs for loading DSP code ************************************
       *
       * After the ARM takes DSP out of reset, the DSP will begin execution
       * at 0xFF80, a hard coded address in DSP ROM, which places the DSP into
       * IDLE 1 mode (a power down mode). This gives the ARM the opportunity
       * to download the DSP code. After it completes downloading the DSP
       * image, the ARM can send an interrupt to the DSP, which wakes it up
       * from IDLE 1 mode and jump to address 0x7F80 where it starts running
       * its application code loaded by the ARM.
       */

      /* COMMAND: DSP_RESET.  This command resets the DSP.  At the conclusion
       * of the reset, the DSP will be in IDLE more ready for code download.
       *
       * ARGUMENT: None
       */

    case DSP_RESET:
       dbg("cmd=DSP_RESET");
       err = dsp_force_idle((int)arg);
       break;

       /* COMMAND: DSP_LOAD.  This command will load program data into DSP
	* memory while the DSP is in IDLE.  The DSP_RESET command should
	* have been issued prior to loading.
	*
	* ARGUMENTS: A reference to an instance of 'struct dsp_load_s'
	*/

    case DSP_LOAD:
       dbg("cmd=DSP_LOAD");
       err = dsp_load_code((struct dsp_load_s*)arg);
       break;

       /* COMMAND: DSP_MEMCLR.  This command will clear program data in the
	* DSP memory while the DSP is in IDLE.  The DSP_RESET command should
	* have been issued prior to loading.  This command may be used to
	* handle initializing of program BSS sections.
	*
	* ARGUMENTS: A reference to an instance of 'struct dsp_memclr_s'
	*/

    case DSP_MEMCLR:
       dbg("cmd=DSP_MEMCLR");
       err = dsp_memclr((struct dsp_memclr_s*)arg);
       break;

       /* COMMAND: DSP_START.  After the program has been loaded (via DSP_LOAD,
	* this command can be used to start the DSP program.  This command will
	* interrupt the DSP and the DSP will branch to address 0x7F80 where the
	* new interrupt vector table is located. The ARM should have loaded this
	* location with at least a branch to the start of the new program.
	*
	* ARGUMENT: A cpu number (usually zero).
	*/

    case DSP_START:
       dbg("cmd=DSP_START");
       err = dsp_start((int)arg);
       break;

       /* COMMAND: DSP_MBOX.  Sets the address of a mailbox to be used by the
	* driver for reading or writing from the DSP.
	*
	* ARGUMENT:  A reference to an instance of a struct dsp_mbspecifier_s
	* that describes the DSP mailbox to use.  The struct dsp_mbspecifier_s
	* contains (among other things):
	*
	*  addr:   The address of the mailbox in the DSP address space.  At this
	*          DSP address, the driver will assume the existence of a
	*          struct dsp_mb_s and will use this address for sending or
	*          receviing messages as determine by the wr_dir setting.
	*  wr_dir: If zero, the mailbox will be used for reading data sent from
	*          the DSP (only); if non-zero, the mailbox will be used for
	*          writing data to the DSP (only).
	*
	* The driver read operation will fail unless a read mailbox is specified
	* (wr_dir == 0); the write operation will fail if no write mailbox is
	* specified (wr_dir != 0).  Note that two calls with the DSP_MBOX are
	* required to support full duplex communications.
	*
	* Each minor device specifies a DSP/ARM "channel."  These MBOX settings
	* apply to all users of a given "channel."
	*/

    case DSP_MBOX:
       dbg("cmd=DSP_MBOX");
       err = dsp_mbox((struct dsp_mbspecifier_s*)arg, minor);
       break;

     default:
       warn("Unrecognized IOCTL command %d", cmd);
       err = -EINVAL;
       break;
    }

  dbg("returning err=%d", err);
  return err;
}

/************************************************************************
 * dsp_open
 ************************************************************************/

static int
dsp_open(struct inode *inode, struct file *filep)
{
  int minor = MINOR(inode->i_rdev);
  struct dsp_drvr_s *drvr = &dsp_drvr[minor];

  dbg("minor=%d", minor);

  /* Verify that the DSP minor number is within the range of those supported
   * by this driver.
   */

  if ((unsigned)minor >= DSP_MINOR_MAX)
    {
      dbg("Bad minor (minor) %d", minor);
      return -EACCES;
    }

  /* Initialized the file structure */

  filep->private_data = 0;
  filep->f_pos        = 0;

  /* Mark the driver as open */

  drvr->opn_count++;

  MOD_INC_USE_COUNT;
  return 0;
}

/************************************************************************
 * dsp_release
 ************************************************************************/

#ifdef LINUX_20
static void
dsp_release (struct inode *inode, struct file *filep)
#else
static int
dsp_release (struct inode *inode, struct file *filep)
#endif /* LINUX_20 */
{
  int minor = MINOR(inode->i_rdev);
  struct dsp_drvr_s *drvr = &dsp_drvr[minor];

  /* Decrement the open count. */

  drvr->opn_count--;
  if (drvr->opn_count <= 0)
    {
      /* Restore default settings */

      memset(drvr, 0, sizeof(struct dsp_drvr_s));

      /* Re-initialize wait queues */

#ifdef DECLARE_WAITQUEUE
      init_waitqueue_head(&drvr->rd_waitq);
      init_waitqueue_head(&drvr->wr_waitq);
#endif /* DECLARE_WAITQUEUE */

      MOD_DEC_USE_COUNT;
    }

#ifndef LINUX_20
  return 0;
#endif /* LINUX_20 */
}

/**************************************************************************
 * dsp_llseek
 **************************************************************************/

#ifdef LINUX_20
static int dsp_llseek(struct inode *inode, struct file *file,
		      off_t offset, int org)
{
  return -ESPIPE;
}
#endif /* LINUX_20 */

/************************************************************************
 * dsp_read
 ************************************************************************/

#ifdef LINUX_20
static int
dsp_read(struct inode *inode, struct file *filep, char *buf, int count)
{
  unsigned int       minor = MINOR(inode->i_rdev);

#else /* LINUX_20 */

static ssize_t
dsp_read(struct file *filep, char *buf, size_t count, loff_t *f_pos)
{
  unsigned int       minor = MINOR(filep->f_dentry->d_inode->i_rdev);
#endif /* LINUX_20 */
  struct dsp_drvr_s *drvr = &dsp_drvr[minor];
  struct dsp_mb_s   *mbox;
  int                flags;

  dbg("Read call");

  /* Verify the the minor number is within range */

  if ((unsigned)minor >= DSP_MINOR_MAX)
    {
      dbg("Bad minor (minor) %d", minor);
      return -EACCES;
    }

  /* Verify that an inbox has been associated with this driver
   * instance.
   */

  else if (!drvr->inbox)
    {
      dbg("Inbox has not been assigned");
      return -EACCES;
    }

  /* The supplied buffer must be at least as large as the mailbox
   * buffer.  We will not handle truncated messages!
   */

  else if (count < (drvr->inbox->size << 1))
    {
      dbg("Receive buffer (%d) is smaller than inbox (%d)",
	  count, (drvr->inbox->size << 1));
      return -EINVAL;
    }

  /* Verify that the destination address is word aligned */

  else if (((((unsigned long)buf) & 1) != 0) || ((count & 1) != 0))
    {
      dbg("Destination buffer not aligned");
      return -EINVAL;
    }

  /* Verify that the DSP is running and can send messages */

  else if (dsp_state.state != DSPSTATE_RUNNING)
    {
      dbg("DSP state is %d", dsp_state.state);
      return -EBUSY;
    }

  /* Verify that we have actually been asked to read something */

  else if (count > 0)
    {
      /* Keep things stationary while we set this up */

      save_flags(flags); cli();

      /* Loop waiting until an inbox to be non-empty */

      while (!IS_INBOX_NONEMPTY(drvr))
	{
	  /* We will have to block.  UNLESS the open flag was
	   * O_NONBLOCK.  In this case, return -EAGAIN.
	   */

	  if (filep->f_flags & O_NONBLOCK)
	    {
	      restore_flags(flags);
	      dbg("Returing -EAGAIN");
	      return -EAGAIN;
	    }
	  else
	    {
	      dbg("Waiting for inbox (%d)", drvr->inbox->state);
	      interruptible_sleep_on(&drvr->rd_waitq);
	      dbg("Awakened (%d)", drvr->inbox->state);

	      /* Check if we were awakened by a signal */
#ifdef LINUX_20
	      if (current->signal & ~current->blocked)
#else
	      if (signal_pending(current))
#endif /* LINUX_20 */
		{
		  restore_flags(flags);
		  dbg("Returing -EINTR");
		  return -EINTR;
		}
	    }
	  cli();
	}
      /* Sample the mailbox before restoring flags.  We do this in case
       * the mbox gets changed by an ioctl call.
       */

      mbox = drvr->inbox;
      restore_flags(flags);

      /* Copy the message from the inbox */

      count = mbox->used << 1;
      if (count > 0)
	{
	  dsp_memcpy((unsigned short*)buf,
		     &mbox->mb[mbox->offset],
		     mbox->used);
	}
      mbox->used = 0;
      mbox->offset = 0;

      /* And signal the DSP */

      mbox->state = DSPMB_COMPLETE;
      dsp_interrupt_dsp(0);

      return count;
    }
  return 0;
}

/************************************************************************
 * dsp_write
 ************************************************************************/

#ifdef LINUX_20

static int
dsp_write(struct inode *inode, struct file *filep, const char *buf, int count)
{
  unsigned int       minor = MINOR(inode->i_rdev);
  struct dsp_drvr_s *drvr = &dsp_drvr[minor];
  struct dsp_mb_s   *mbox;
  int                wcount;
  int                flags;

  dbg("Write call");

#else /* LINUX_20 */

static ssize_t
dsp_write(struct file *filep, const char *buf, size_t count, loff_t *ppos)
{
  unsigned int       minor = MINOR(filep->f_dentry->d_inode->i_rdev);
  struct dsp_drvr_s *drvr = &dsp_drvr[minor];
  struct dsp_mb_s   *mbox;
  int                wcount;
  int                flags;

  dbg("Write call");

  /*  Cannot seek (pwrite) on this device  */

  if (ppos != &filep->f_pos)
    {
      dbg("Cannot seek on this device");
      return -ESPIPE;
    }
  else
#endif /* LINUX_20 */

  /* Verify the the minor number is within range */

  if ((unsigned)minor >= DSP_MINOR_MAX)
    {
      dbg("Bad minor (minor) %d", minor);
      return -EACCES;
    }

  /* Verify that an outbox has been associated with this driver
   * instance.
   */

  else if (!drvr->outbox)
    {
      dbg("Outbox has not been assigned");
      return -EACCES;
    }

  /* Verify that the outbound message will actually fit in the outbox */

  else if (count > (drvr->outbox->size << 1))
    {
      dbg("Message size (%d) is bigger than outbox (%d)",
	  count, (drvr->outbox->size << 1));
      return -EINVAL;
    }

  /* Verify that the source address is word aligned */

  else if (((((unsigned long)buf) & 1) != 0) || ((count & 1) != 0))
    {
      dbg("Source buffer not aligned");
      return -EINVAL;
    }

  /* Verify that the DSP is running and can receive messages */

  else if (dsp_state.state != DSPSTATE_RUNNING)
    {
      dbg("DSP state is %d", dsp_state.state);
      return -EBUSY;
    }

  /* Verify that we have actually been asked to write something */

  else if (count > 0)
    {
      /* Keep things stationary while we set this up */

      save_flags(flags); cli();

      /* Loop waiting until an outbox is available */

      while (!IS_OUTBOX_NONFULL(drvr))
	{
	  /* We will have to block.  UNLESS the open flag was
	   * O_NONBLOCK.  In this case, return -EAGAIN.
	   */

	  if (filep->f_flags & O_NONBLOCK)
	    {
	      restore_flags(flags);
	      dbg("Returing -EAGAIN");
	      return -EAGAIN;
	    }
	  else
	    {
	      dbg("Waiting for outbox (%d)", drvr->outbox->state);
	      interruptible_sleep_on(&drvr->wr_waitq);
	      dbg("Awakened (%d)", drvr->outbox->state);

	      /* Check if we were awakened by a signal */
#ifdef LINUX_20
	      if (current->signal & ~current->blocked)
#else
	      if (signal_pending(current))
#endif /* LINUX_20 */
		{
		  restore_flags(flags);
		  dbg("Returing -EINTR");
		  return -EINTR;
		}
	    }
	  cli();
	}

      /* Sample the mailbox before restoring flags.  We do this in case
       * the mbox gets changed by an ioctl call.
       */

      mbox = drvr->outbox;
      restore_flags(flags);

      /* Copy the message into the outbox */

      wcount = count >> 1;
      dsp_memcpy(mbox->mb, (unsigned short*)buf, wcount);
      mbox->used = wcount;
      mbox->offset = 0;

      /* Check for read errors */

      if (drvr->rd_error)
	{
	  count = drvr->rd_error;
	}

      /* And signal the DSP */

      drvr->outbox->state = DSPMB_AVAILABLE;
      dsp_interrupt_dsp(0);

      return count;
    }
  return 0;
}

/************************************************************************
 * dsp_interrupt
 ************************************************************************/

static void
dsp_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  struct dsp_drvr_s *drvr;
  int i;

  dbg("irq%d", irq);

  if (irq == DM270_IRQ_DSP)
    {
      /* Wake up all instances that are waiting on an interupt */

      for (i = 0; i < DSP_MINOR_MAX; i ++)
	{
	  /* Is this instance open? */

	  drvr = &dsp_drvr[i];
	  if (IS_OPEN(drvr))
	    {
	      /* Yes, is it waiting for an incoming message?
	       * If so, has a message been received?
	       */

	      if (drvr->inbox && IS_INBOX_WAITING(drvr) &&
		  IS_INBOX_NONEMPTY(drvr))
		{
		  /* Let the world know that we have the inbox */

		  drvr->inbox->state = DSPMB_WAIT;

		  /* Wake up dsp_read */

		  wake_up_interruptible(&drvr->rd_waitq);
		}

	      /* Or, is it waiting to send and outbound message?
	       * If so, is there room in the outbox to send it now?
	       */

	      else if (drvr->outbox && IS_OUTBOX_WAITING(drvr) &&
		       IS_OUTBOX_NONFULL(drvr))
		{
		  /* Check if an error was reported */

		  if (IS_READ_ERROR(drvr))
		    {
		      drvr->rd_error = (int)drvr->outbox->state;
		    }
		  else
		    {
		      drvr->rd_error = 0;
		    }

		  /* Let the world know that we have the outbox */

		  drvr->outbox->state = DSPMB_WAIT;

		  /* Wake up dsp_write */

		  wake_up_interruptible(&drvr->wr_waitq);
		}

	      /* Check for the (uninteresting) read completion event */

	      else if (drvr->outbox && IS_WRITE_COMPLETE(drvr))
		{
		  /* Mark the inbox ready for the next message */

		  drvr->outbox->state = DSPMB_READY;
		}
	    }
	}
    }
}

/************************************************************************
 * Global Functions
 ************************************************************************/

/************************************************************************
 * dsp_init
 ************************************************************************/

#ifdef LINUX_20
int dsp_init(void)
#else
int __init dsp_init(void)
#endif
{
  int retval = 0;
  int i;

  /* Announce that we are installing */

  printk("DM270 DSP Control Driver\n");

  /* Basic initialization of device nodes. */

  retval = register_chrdev(DSP_CONTROL_MAJOR, myname, &dsp_fileops);
  if (retval != 0)
    {
      err("Cannot get major number %d", DSP_CONTROL_MAJOR);
    }
  else
    {
      dbg("DSP major number=%d", DSP_CONTROL_MAJOR);

      /* Reserve the DSP register block */

      request_region(HPIB_REGISTER_BASE, HPIB_BLOCK_SIZE, myname);

      /* Reset the DSP */

      dsp_force_idle(0);

      /* Set the DSP interface into a known, initial state. */

      memset(&dsp_state, 0, sizeof(struct dsp_state_s));
      dsp_state.hpibctl = DM270_HPIBCTL_INIT;
      dsp_state.state   = DSPSTATE_IDLE;
      outw(dsp_state.hpibctl, DM270_HPIBCTL);

      /* Set the DSP driver into a known, initial state. */

      for (i = 0; i < DSP_MINOR_MAX; i ++)
	{
	  memset(&dsp_drvr[i], 0, sizeof(struct dsp_drvr_s));

	  /* Initialize wait queues */

#ifdef DECLARE_WAITQUEUE
	  init_waitqueue_head(&dsp_drvr[i].rd_waitq);
	  init_waitqueue_head(&dsp_drvr[i].wr_waitq);
#endif /* DECLARE_WAITQUEUE */
	}

      /* Get and enable the IRQ */

      retval = request_irq(DM270_IRQ_DSP, dsp_interrupt, SA_INTERRUPT,
			   myname, dsp_interrupt);
      if (retval != 0)
	{
	  err("Unable to get IRQ for the DSP driver");
	  unregister_chrdev(DSP_CONTROL_MAJOR, myname);
	}
#ifdef MODULE
      else
	{
	  REGISTER_SYMTAB(&dsp_symbol_table);
	}
#endif
    }

  return retval;
}

/************************************************************************
 * dsp_exit
 ************************************************************************/

#ifdef MODULE
#ifdef LINUX_20
void dsp_exit(void)
#else
void __exit dsp_exit(void)
#endif
{
  release_region(HPIB_REGISTER_BASE, HPIB_BLOCK_SIZE);
  unregister_chrdev(DSP_CONTROL_MAJOR, myname);
  free_irq(DM270_IRQ_DSP, dsp_interrupt);
}
#endif

/************************************************************************
 * Module stuff
 ************************************************************************/

/************************************************************************
 * dsp_get_fileops
 ************************************************************************/

struct file_operations *dsp_get_fileops(void)
{
  return &dsp_fileops;
}

#ifdef MODULE
module_init(dsp_init);
module_exit(dsp_exit);
#endif
