/************************************************************************
 * linux/drivers/char/gio_dm270.c
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived from linux/drivers/char/gio_dsc25.c
 *
 *   Copyright(c) 2002, Cadenux, LLC
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

/************************************************************************
 * Included Files
 ************************************************************************/

/* Use versioning if needed */

#include <linux/autoconf.h>

#ifdef CONFIG_MODVERSIONS
# undef MODVERSIONS
# define MODVERSIONS
#endif

#ifdef MODVERSIONS
#  include <linux/modversions.h>
#endif

#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>

#include <asm/types.h>
#include <asm/semaphore.h>
#include <asm/segment.h>

#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <asm/arch/gio_ioctl.h>
#include <asm/arch/sysdep.h>
#include <asm/arch/memconfig.h>
#include <linux/poll.h>

/************************************************************************
 * Definitions
 ************************************************************************/

/* #undef GIO_DEBUG */
/* #define GIO_DEBUG */

#ifdef GIO_DEBUG
# define dbg(format, arg...) \
  printk(KERN_DEBUG __FUNCTION__ ": " format "\n" , ##arg)

#else
# define dbg(x...)
#endif

#define info(format, arg...) \
  printk(KERN_INFO __FUNCTION__ ": " format "\n" , ##arg)
#define err(format, arg...) \
  printk(KERN_ERR __FUNCTION__ ": " format "\n" , ##arg)

/* The DM270 has 32 GIO pins for supporting different customer peripheral
 * controllers.
 */

#define NUM_GIOS 32

/* For the DM270, GIO0-GIO15 can be configured as external interrupt ports */

#define NUM_GIO_INTERRUPTS  16

/* For the DM270, a GIO pin may be used as ethernet interrupt.
 * GIO initialization for this pin was performed very early
 * in the low-level setup.  We only have to avoid mucking with those
 * settings.
 */

#if defined(CONFIG_DM270_NET)
# define ETH_GIO_BIT GIO_4_BIT
#endif

#if defined (CONFIG_BLK_DEV_IDE_DM270_CFC)
# if defined(BSPCONF_DM270_TI)
#  define CF_PWR_GIO_BIT GIO_14_BIT
# endif
#endif

#if defined(BSPCONF_DM270_TI)
# define UART1_RXD1_GIO_PIN 23
# define UART1_TXD1_GIO_PIN 24
#endif

#if 0
# define CCD_WEN_GIO_PIN    26
# define CCD_SHUT_GIO_PIN   27
#endif


/************************************************************************
 * DIR0 Register Bit Definitions:
 * Bits 0-15: Port Direction Bit.  Sets the IO direction for each
 *            port.  Bit15 sets the GIO15 pin; Bit 0 sets the GIO0 pin.
 *            0=output; 1=input
 */

/************************************************************************
 * DIR1 Register Bit Definitions:
 * Bits 0-15: Port Direction Bit.  Sets the IO direction for each
 *            port.  Bit15 sets the GIO31 pin; Bit 0 sets the GIO16 pin.
 *            0=output; 1=input
 */

/************************************************************************
 * INV0 Register Bit Definitions:
 * Bits 0-15: Port Inversion Bit.  The polarity of the IO port is
 *            reversed.  Bit15 sets the GIO15 pin; Bit 0 sets the GIO0 pin.
 *            0=Unreversed; 1=reversed
 */

/************************************************************************
 * INV1 Register Bit Definitions:
 * Bits 0-15: Port Inversion Bit.  The polarity of the IO port is
 *            reversed.  Bit15 sets the GIO31 pin; Bit 0 sets the GIO16 pin.
 *            0=Unreversed; 1=reversed
 */

/************************************************************************
 * BITSET0 Register Bit Definitions:
 * Bits 0-15: GIO bit set 0.  If "1" is written in the bit when the GIO pin
 *            is set in the output, the corresponding GIO pin becomes "1."
 *            Bit 15 corresponds to GIO15 and bit 0 corresponds to GIO0.
 *            When the GIO pin i sset in the input, it is possible to read
 *            from the bit to which the logical level of the GIO pin
 *            corresponds.  If it is set in the reverse/inverse in INV0
 *            or INV1 register and if GIO is set to input then the
 *            reversed/inversed value can be read.  When GIO is set to
 *            output, the INV bit is ignoredl.
 */

/************************************************************************
 * BITSET1 Register Bit Definitions:
 * Bits 0-15: GIO bit set 1.  It is the same function as GIO BITSET0.
 *            Bit 15 corresponds to GIO31 and bit 0 corresponds to GIO16.
 */

/************************************************************************
 * BITCLR0 Register Bit Definitions:
 * Bits 0-15: GIO bit clear 0. When GIO is set to output and "1" is written
 *            to the corresponding bit, the GIO bit is cleared to "0."
 *            Bit 15 corresponds to GIO15 and bit 0 corresponds to GIO0.
 */

/************************************************************************
 * BITCLR1 Register Bit Definitions:
 * Bits 0-15: GIO bit clear 1.  It is the same function as GIO BITCLR0.
 *            Bit 15 corresponds to GIO31 and bit 0 corresponds to GIO16.
 */

/************************************************************************
 * IRQPORT Register Bit Definitions:
 * Bits 0-7:  Interrupt Port. GIO07-GIO0 is used as an external interrupt
 *            port.  Bit 7 sets GIO7 and bit 0 sets GIO0. 0=usual GIO,
 *            1=external interrupt.  When used as an external interrupt,
 *            the IO setting of DIR0 register is set to input.  Interrupt
 *            detects the pulse more than once cycle of the ARM clock
 *            when the reversing input is "0" and when the nonreversing
 *            input is "1" by the INV register.
 * Bits 8-15: Reserved.
 */

/* For the DM270 GIO8-GIO33 can be used as dedicated function ports
 * rather than general purpose IO ports.  GIO8-GIO33 functions are
 * selected in FSEL0 and FSEL1 registers. See asm/arch-dm270/gio_ioctl.h
 * for possible settings.
 */

#define FSEL0_FIRST_GIO              8
#define FSEL0_LAST_GIO              23
#define FSEL1_FIRST_GIO             24
#define FSEL1_LAST_GIO              33

/************************************************************************
 * Private Type Declarations
 ************************************************************************/

/* Driver state information. */

struct gio_state_s
{
#ifdef DECLARE_WAITQUEUE
  wait_queue_head_t  gio_waitq[NUM_GIO_INTERRUPTS];
#else
  struct wait_queue *gio_waitq[NUM_GIO_INTERRUPTS];
#endif

  char               interrupt_pending[NUM_GIO_INTERRUPTS];
  unsigned long      opened_pins;	
  unsigned long      direction_pins;
  unsigned long      interrupt_pins;
  unsigned long      edge_detect;
  unsigned long      function_pins;
};

/* FSEL decoding structure */

struct fsel_decode
{
  unsigned char mask  : 4;
  unsigned char shift : 4;
};

/************************************************************************
 * Local Function Prototypes
 ************************************************************************/

static int
gio_ioctl(struct inode *inode, struct file *filep, unsigned int cmd,
	  unsigned long arg);

static int
gio_open(struct inode * inode, struct file * filep);

#ifdef LINUX_20
static void
gio_release (struct inode *inode, struct file *filep);
#else
static int
gio_release (struct inode *inode, struct file *filep);
#endif /* LINUX_20 */

#ifdef LINUX_20
static int
gio_read(struct inode *inode, struct file *filep, char *buf, int count);
#else
static ssize_t
gio_read(struct file *filep, char *buf, size_t count, loff_t *f_pos);
#endif /* LINUX_20 */

static unsigned int 
gio_poll(struct file *filep, poll_table *wait);  

static void
gio_interrupt(int irq, void *dev_id, struct pt_regs *regs);

/************************************************************************
 * Private Constant Data
 ************************************************************************/

/* FSEL decoding structure */

static const struct fsel_decode fsel0_decode[] =
{
  {1,  0}, /* GIO8 */
  {1,  1}, /* GIO9 */
  {1,  2}, /* GIO10 */
  {1,  3}, /* GIO11 */
  {1,  4}, /* GIO12 */
  {1,  5}, /* GIO13 */
  {1,  6}, /* GIO14 */
  {1,  7}, /* GIO15 */
  {1,  8}, /* GIO16 */
  {1,  9}, /* GIO17 */
  {1, 10}, /* GIO18 */
  {1, 11}, /* GIO19 */
  {1, 12}, /* GIO20 */
  {1, 13}, /* GIO21 */
  {1, 14}, /* GIO22 */
  {1, 15}, /* GIO23 */

};

static const struct fsel_decode fsel1_decode[] =
{
  {1,  0}, /* GIO24 */
  {3,  1}, /* GIO25 */
  {1,  3}, /* GIO26 */
  {1,  4}, /* GIO27 */
  {1,  5}, /* GIO28 */
  {1,  6}, /* GIO29 */
  {1,  7}, /* GIO30 */
  {1,  8}, /* GIO31 */
  {1,  9}, /* GIO32 */
  {1, 10}, /* GIO33 */
};

const char *gio_names[] =
{
  "gio0", "gio1",  "gio2",  "gio3",  "gio4",  "gio5",  "gio6",  "gio7",
  "gio8", "gio9", "gio10", "gio11", "gio12", "gio13", "gio14", "gio15"

};

/************************************************************************
 * Private Variables
 ************************************************************************/

/* Current state of all of the GIO devices */

static struct gio_state_s gio_state;

/* Device File Operations */

static struct file_operations gio_fileops =
{
  ioctl:     gio_ioctl,         /* ioctl */
  open:      gio_open,          /* open code */
  release:   gio_release,       /* close code */
  read:      gio_read,          /* Wait-for-interrupt code */
  poll:      gio_poll,          /* poll / select code */
};

/* Give other modules access to the GIO ioctl calls without having to
 * search for the GIO module.
 */

#ifdef MODULE
#if 1
static struct symbol_table gio_symbol_table =
{
# include <linux/symtab_begin.h>
  X(gio_get_fileops),
# include <linux/symtab_end.h>
};
#else
EXPORT_SYMBOL(gio_get_fileops);
#endif
#endif

/************************************************************************
 * Private Functions
 ************************************************************************/

/************************************************************************
 * gio_ioctl
 ************************************************************************/

static int
gio_ioctl (struct inode *inode, struct file *filep,
	   unsigned int cmd, unsigned long arg)
{
  unsigned long  reg;
  unsigned short value;
  int            err      = 0;
  int            pin      = MINOR(inode->i_rdev);
  signed int     size     = _IOC_SIZE(cmd);

  dbg("cmd=0x%x pin=%d arg=0x%lx", cmd, pin, arg);
  dbg("opened_pins=0x%08lx interrupt_pins=0x%04lx edge_detect=0x%04lx"
      "direction_pins=0x%08lx function_pins=0x%08lx",
      gio_state.opened_pins, gio_state.interrupt_pins, gio_state.edge_detect,
      gio_state.direction_pins, gio_state.function_pins);
    
  /* Range check the IOCTL cmd */

  if ((_IOC_TYPE(cmd) != GIO_MAGIC_NUMBER) || (_IOC_NR(cmd) > GIO_IOC_MAXNR)) 
    {
      dbg("Bad ioctl value");
      return -EINVAL;
    }

  /* Check address validity */

  if (_IOC_DIR(cmd) & _IOC_READ)
    {
      err = verify_area(VERIFY_READ, (void*)arg, size);

      if (err)
	{
	  dbg("verify_area(VERIFY_READ) error:%d",err);
	}
    } 
  else if (_IOC_DIR(cmd) & _IOC_WRITE) 
    {
      err = verify_area(VERIFY_WRITE, (void*)arg, size);
          
      if (err)
	{
	  dbg("verify_area(VERIFY_WRITE) error:%d",err);
	}
    }

  if (err) 
    {
      return err;
    }

  /* Command processing */

  switch (cmd) 
    {
    case GIO_DIRECTION:

      /******************************************************************
       * GIO_DIRECTION: The GIO_DIRECTION can be either input (GIO_INPUT)
       * or output (GIO_OUTPUT).  If unspecified, the GIO will default
       * to input.
       */

      {
	dbg("GIO_DIRECTION: pin=%d arg=%ld", pin, arg);

	/* Pins 0-15 are controlled by bits in DIR0; pins 16-31 are
	 * controlled by bits in DIR1.
	 */

	if (pin < 16)
	  {
	    reg = DM270_GIO_DIR0;
	    value = pin;
	  }
	else
	  {
	    reg = DM270_GIO_DIR1;
	    value = pin - 16;
	  }

	/* Set or clear the direction bit associated with the pin */

	if (arg)
	  {
	    outw((inw(reg) | (1 << value)), reg);
	    gio_state.direction_pins |= (1 << pin);
	  }
	else
	  {
	    outw((inw(reg) & ~(1 << value)), reg);
	    gio_state.direction_pins &= ~(1 << pin);
	  }
      }
      break;

      /******************************************************************
       * GIO_INVERSION: The GIO polarity of the GIO pin can be inverted.
       * If unspecified, the default behavior will be NOT inverted.
       */

    case GIO_INVERSION:
      {
	dbg("GIO_INVERSION: pin=%d arg=%ld", pin, arg);

	/* Pins 0-15 are controlled by bits in INV0; pins 16-31 are
	 * controlled by bits in INV1.
	 */

	if (pin < 16)
	  {
	    reg = DM270_GIO_INV0;
	    value = pin;
	  }
	else
	  {
	    reg = DM270_GIO_INV1;
	    value = pin - 16;
	  }

	/* Set or clear the inversion bit associated with the pin */

	if (arg)
	  {
	    outw((inw(reg) | (1 << value)), reg);
	  }
	else
	  {
	    outw((inw(reg) & ~(1 << value)), reg);
	  }
      }
      break;

      /******************************************************************
       * GIO_BITSET: If the direction is set to output, this ioctl will set
       * the output.  If the direction is input, this ioctl will return
       * the value of the GIO pin (0, 1, or -1 on error).
       */

    case GIO_BITSET:
      {
	dbg("GIO_BITSET: pin=%d arg=%ld", pin, arg);

	if ((gio_state.direction_pins & (1 << pin)) != 0)
	  {
	    /* Pin is programmed for input */

	    /* Just like in read(), reading value of pin indicates
	       that the interrupt has been received and processed.
	       This allows application to block with select then
	       do read via GIO_BITSET without a useless call to read() */

	    gio_state.interrupt_pending[pin] = 0;


	    /* Input state for pins 0-15 is provided in BITSET1; pin
	     * 16-31 state is provided in BITSET2.
	     */

	    if (pin < 16)
	      {
		/* If the pin is set, return "1" otherwise the
		 * default behavior for this ioctl is to return
		 * zero.
		 */

		if (inw(DM270_GIO_BITSET0) & (1 << pin))
		  {
		    return 1;
		  }
	      }
	    else
	      {
		/* If the pin is set, return "1" otherwise the
		 * default behavior for this ioctl is to return
		 * zero.
		 */

		if (inw(DM270_GIO_BITSET1) & (1 << (pin - 16)))
		  {
		    return 1;
		  }
	      }
	  }
	else
	  {
	    /* Pin is programmed for output */

	    /* Check if the pin has been programmed as an interrupt:  We
	     * will not set a bit on an interrupt pin.
	     */

	    if ((gio_state.interrupt_pins & (1 << pin)) != 0)
	      {
		dbg("pin %d is programmed for interrupt", pin);
		return -EINVAL;
	      }

	    /* Output state for pins 0-15 is controlled in BITSET0; pin
	     * 16-31 state is controlled in BITSET1.
	     */

	    if (pin < 16)
	      {
		outw(1 << pin, DM270_GIO_BITSET0);
	      }
	    else
	      {
		outw(1 << (pin - 16), DM270_GIO_BITSET1);
	      }
	  }
      }
      break;

      /******************************************************************
       * GIO_BITCLR: If the direction is set to output, this ioctl will
       * clear the output.  If the direction is input, this ioctl will
       * return the value of the GIO pin (0, 1, or -1 on error).
       */

    case GIO_BITCLR:
      {
	dbg("GIO_BITCLR: pin=%d arg=%ld", pin, arg);

	if ((gio_state.direction_pins & (1 << pin)) != 0)
	  {
	    /* Pin is programmed for input */

	    /* Just like in read(), reading value of pin indicates
	       that the interrupt has been received and processed.
	       This allows application to block with select then
	       do a read with GIO_BITCLR without a useless call to read() */

	    gio_state.interrupt_pending[pin] = 0;

	    /* Input state for pins 0-15 is provided in BITCLR1; pin
	     * 16-31 state is provided in BITCLR2.
	     */

	    if (pin < 16)
	      {
		/* If the pin is set, return "1" otherwise the
		 * default behavior for this ioctl is to return
		 * zero.
		 */

		if (inw(DM270_GIO_BITCLR0) & (1 << pin))
		  {
		    return 1;
		  }
	      }
	    else
	      {
		/* If the pin is set, return "1" otherwise the
		 * default behavior for this ioctl is to return
		 * zero.
		 */

		if (inw(DM270_GIO_BITCLR1) & (1 << (pin - 16)))
		  {
		    return 1;
		  }
	      }
	  }
	else
	  {
	    /* Pin is programmed for output */

	    /* Check if the pin has been programmed as an interrupt:  We
	     * will not clear a bit on an interrupt pin.
	     */

	    if ((gio_state.interrupt_pins & (1 << pin)) != 0)
	      {
		dbg("pin %d is programmed for interrupt", pin);
		return -EINVAL;
	      }


	    /* Output state for pins 0-15 is controlled in BITCLR0; pin
	     * 16-31 state is controlled in BITCLR1.
	     */

	    if (pin < 16)
	      {
		outw(1 << pin, DM270_GIO_BITCLR0);
	      }
	    else
	      {
		outw(1 << (pin - 16), DM270_GIO_BITCLR1);
	      }
	  }
      }
      break;

      /******************************************************************
       * GIO_IRQPORT: For the DM270, GIO0-GIO15 can be configured as
       * external interrupt ports
       */

    case GIO_IRQPORT:
      {
	dbg("GIO_IRQPORT: pin=%d arg=%ld", pin, arg);

	/* Verify that pin corresponds to one of GIO0-GIO15. */

	if (pin >= NUM_GIO_INTERRUPTS)
	  {
	    dbg("pin=%d will not support interrupts", pin);
	    return -EINVAL;
	  }

	/* 0 = detach from interrupt;
	 * 1 = attach to interrupt with single edge detection
	 * 2 = attach to interrupt with double edge detection
	 */

	if (/* (arg < 0) && */ (arg > 2)) /* arg is unsigned */
	  {
	    dbg("Invalid argument %ld", arg);
	    return -EINVAL;
	  }

	/* Is this a request to attach to the interrupt? */

	if (arg != 0)
	  {
	    /* Yes, attach to the interrupt */

	    /* Have we already attached to this interrupt? */

	    if ((gio_state.interrupt_pins & (1 << pin)) == 0)
	      {
		/* No, ... then do so now. */

		if (request_irq(pin + DM270_IRQ_EXB0, &gio_interrupt, 
				SA_INTERRUPT, gio_names[pin], 
				&gio_interrupt))
		  {
		    err("Could not attach to IRQ%d", pin);
		    return -EBUSY;
		  }
		else
		  {
		    dbg("Attached to IRQ%d", pin);
		    gio_state.interrupt_pins |= (1 << pin);
		  }
	      }
	    else
	      {
		dbg("IRQ%d already attached", pin);
	      }
	  }
	else
	  {
	    /* Detach from the interrupt */

	    if ((gio_state.interrupt_pins & (1 << pin)) != 0)
	      {
		free_irq(pin + DM270_IRQ_EXB0, &gio_interrupt);
		gio_state.interrupt_pins &= ~(1 << pin);
	      }
	    else
	      {
		dbg("IRQ%d already detached", pin);
	      }
	  }

	/* In any event, is this a request to enable the
	 * interrupt with double edge detection?
	 */

	if (arg == 2)
	  {
	    /* Yes, set the edge detection bit. */

	    gio_state.edge_detect |= (1 << pin);
	  }
	else
	  {
	    /* No, clear the edge detection bit. */

	    gio_state.edge_detect &= ~(1 << pin);
	  }

	/* Then make it so */

	outw(gio_state.edge_detect, DM270_GIO_IRQEDGE);
	outw(gio_state.interrupt_pins, DM270_GIO_IRQPORT);
      }
      break;

      /******************************************************************
       * GIO_FSEL: For the DM270 GIO8-GIO28 can be used as dedicated
       * function ports rather than general purpose IO ports.
       */

    case GIO_FSEL:
      {
	const struct fsel_decode *decode;
	unsigned short mask;

	dbg("GIO_FSEL: pin=%d arg=%ld", pin, arg);

	/* For the DM270 GIO8-GIO28 can be used as dedicated function ports
	 * rather than general purpose IO ports.  GIO8-GIO28 functions are
	 * selected in FSEL0 and FSEL1 registers.
	 */

	/* Verify that the selected GIO is within raange */

	if (pin < FSEL0_FIRST_GIO)
	  {
	    dbg("pin=%d cannot be used as function port", pin);
	    return -EINVAL;
	  }
	else if (pin <= FSEL0_LAST_GIO)
	  {
	    dbg("Using FSEL0 for pin=%d", pin);
	    decode = &fsel0_decode[pin - FSEL0_FIRST_GIO];
	    reg = DM270_GIO_FSEL0;
	  }
	else if (pin <= FSEL1_LAST_GIO)
	  {
	    dbg("Using FSE10 for pin=%d", pin);
	    decode = &fsel1_decode[pin - FSEL1_FIRST_GIO];
	    reg = DM270_GIO_FSEL1;
	  }
	else
	  {
	    dbg("pin=%d cannot be used as function port", pin);
	    return -EINVAL;
	  }

	dbg("For pin=%d decode mask=0x%02x shift=%d",
	    pin, decode->mask, decode->shift);

	/* Check if the argument is within range. */

	if (arg > decode->mask)
	  {
	    dbg("arg=%ld is out for range for pin=%d", arg, pin);
	    return -EINVAL;
	  }

	/* We can handle one and two bit settings */

	mask = decode->mask << decode->shift;
	dbg("mask=0x%04x", mask);

	/* Zero always means "normal" GIO mode */

	if (arg == 0)
	  {
	    dbg("clearing under mask=0x%04x", mask);
	    outw((inw(reg) & ~mask), reg);
	    gio_state.function_pins &= ~(1 << pin);
	  }
	else
	  {
	    dbg("Setting %ld under mask=0x%04x", arg, mask);
	    outw(((inw(reg) & ~mask) | (arg << decode->shift)), reg);
	    gio_state.function_pins |= (1 << pin);
	  }
      }
      break;

      /* Unlike the DSC25 and DSC21, the DM270 has no bitrate registers */

    case GIO_BITRATE:
      {
	err("DM270 GIO_BITRATE not supported (pin=%d arg=%ld)", pin, arg);
	return -ENOSYS;
      }
      break;

    default:
      dbg("Error, falling through switch");
      return -EINVAL;
      break;
    }
  return 0;
}

/************************************************************************
 * gio_open
 ************************************************************************/

static int
gio_open(struct inode * inode, struct file * filep)
{
  int pin = MINOR(inode->i_rdev);

  dbg("pin=%d", pin);

  /* Verify that the GIO pin number is within the range of those supported
   * by this driver.
   */

  if (pin >= NUM_GIOS)
    {
      dbg("Bad minor (pin) %d", pin);
      return -EACCES;
    }

  /* Initialized the file structure */

  filep->private_data = 0;
  filep->f_pos        = 0;

  /* Check if this GIO has already been opened.  Sorry only one per
   * customer.
   */

  if (gio_state.opened_pins & (1 << pin))
    {
      err("GIO%d already opened", pin);
      return -EACCES;
    }

  /* Mark the driver as open */

  gio_state.opened_pins |= (1 << pin);
  MOD_INC_USE_COUNT;
  return 0;
}

/************************************************************************
 * gio_release
 ************************************************************************/

#ifdef LINUX_20
static void
gio_release (struct inode *inode, struct file *filep)
#else
static int
gio_release (struct inode *inode, struct file *filep)
#endif /* LINUX_20 */
{
  int pin = MINOR(inode->i_rdev);

  /* Restore default pin settings */

  (void)gio_ioctl(inode, filep, GIO_DIRECTION, GIO_INPUT);
  (void)gio_ioctl(inode, filep, GIO_INVERSION, GIO_UNINVERTED);
  (void)gio_ioctl(inode, filep, GIO_IRQPORT,   GIO_NOINTERRUPT);
  (void)gio_ioctl(inode, filep, GIO_FSEL,      GIO_NORMAL);

  /* And mark the GIO un-opened */

  gio_state.opened_pins &= ~(1 << pin);

  MOD_DEC_USE_COUNT;

#ifndef LINUX_20
  return 0;
#endif /* LINUX_20 */
}

/************************************************************************
 * gio_read
 ************************************************************************/

#ifdef LINUX_20
static int
gio_read(struct inode *inode, struct file *filep, char *buf, int count)
#else
static ssize_t
gio_read(struct file *filep, char *buf, size_t count, loff_t *f_pos)
#endif /* LINUX_20 */
{
#ifdef LINUX_20
  int pin = MINOR(inode->i_rdev);
#else
  int pin = MINOR(filep->f_dentry->d_inode->i_rdev);
#endif /* LINUX_20 */
  int flags;

  /* Make sure that this pin has been programmed for interrupts.
   * If not, we'll return success immediately.
   */

  if ((gio_state.interrupt_pins & (1 << pin)) != 0)
    {
      /* Loop until we receive an interrupt on this pin. */

      save_flags(flags); cli();
      while (gio_state.interrupt_pending[pin] == 0)
	{
	  /* If the open flag was O_NONBLOCK and and no interrupt has been
	   * received, return -EAGAIN.
	   */

	  if (filep->f_flags & O_NONBLOCK)
	    {
	      restore_flags(flags);
	      dbg("Returing -EAGAIN");
	      return -EAGAIN;
	    }

	  dbg("%s waiting on timer queue for GIO%d", current->comm, pin);
	  interruptible_sleep_on(&gio_state.gio_waitq[pin]);
	  dbg("%s waiting from timer queue for GIO%d", current->comm, pin);

	  /* Check signals */

#ifdef LINUX_20
	  if (current->signal & ~current->blocked)
#else
	  if (signal_pending(current))
#endif /* LINUX_20 */
	    {
	      dbg("Returing -EINTER");
	      return -EINTR;
	    }
	  cli();
	}
      restore_flags(flags);

      /* Indicate that the interrupt has been received and processed. */

      gio_state.interrupt_pending[pin] = 0;
    }
  return 0;
}
/************************************************************************
 * gio_poll
 ************************************************************************/

/* Called without the kernel lock - fine */

static unsigned int 
gio_poll(struct file *filep, poll_table *wait)
{
  int pin = MINOR(filep->f_dentry->d_inode->i_rdev);

  unsigned int mask = 0;

  /* Make sure that this pin has been programmed for interrupts.
   * If not, we'll return immediately.
   */

  if ((gio_state.interrupt_pins & (1 << pin)) == 0)
    {
      return mask;
    }

  poll_wait(filep, &gio_state.gio_waitq[pin], wait);

  if (gio_state.interrupt_pending[pin] != 0)
    {
      mask |= POLLIN | POLLRDNORM;
    }

  return(mask);
}

/************************************************************************
 * gio_interrupt
 ************************************************************************/

static void
gio_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  dbg("irq%d", irq);

  if ((irq >= DM270_IRQ_EXB0) && (irq <= DM270_IRQ_EXB15))
    {

      /* GIO interrupts are received on DM270_IRQ_EXB0-15.  We assume here
       * that these are numbered contiguously.  Note that there is not
       * check to see if the pin is programmed to receive interrupts.
       */

      int pin = irq - DM270_IRQ_EXB0;

      /* Indicate that the interrupt has been received on this pin and
       * wake up gio_read if it is waiting for the interrupt.
       */

      gio_state.interrupt_pending[pin] = 1;
      wake_up_interruptible(&gio_state.gio_waitq[pin]);
    }
}

/************************************************************************
 * Global Functions
 ************************************************************************/

/************************************************************************
 * gio_dm270_init
 ************************************************************************/

int gio_dm270_init(void)
{
  /* Announce that we are installing */

  printk("DM270 GIO Driver\n");

  /* Basic initialization of device nodes. */

  if (register_chrdev(GIO_MAJOR, "gio", &gio_fileops) != 0)
    {
      err("Cannot get major number %d\n", GIO_MAJOR);
      return -1;
    }
  else
    {
      dbg("GIO major number=%d\n", GIO_MAJOR);
    }

  /* Set gio ports AND gio_state into known state.
   * The pins default to inputs, so don't initialize BITSET/BITCLR.
   */

  memset(&gio_state, 0, sizeof(struct gio_state_s));

  /* Initialize wait queues for the pins that might be used as
   * interrupts.
   */

#ifdef DECLARE_WAITQUEUE
  {
    int i;
    for (i = 0; i < NUM_GIO_INTERRUPTS; i++)
      {
	init_waitqueue_head(&gio_state.gio_waitq[i]);
      }
  }
#endif /* DECLARE_WAITQUEUE */

  /* Set aside a GIO pin if it might be used for ethernet. */

#ifdef CONFIG_DM270_NET
  gio_state.opened_pins |= ETH_GIO_BIT;
#endif

#if defined (CONFIG_BLK_DEV_IDE_DM270_CFC)
# if defined(BSPCONF_DM270_TI)
   /* Set aside a GIO pin if it might be used for Compact Flash. */
   gio_state.opened_pins |= CF_PWR_GIO_BIT;
# endif
#endif

#if defined(BSPCONF_DM270_TI)
   /* Set aside GIO pins for UART 1 */
   gio_state.opened_pins |= (1 << UART1_RXD1_GIO_PIN);
   gio_state.opened_pins |= (1 << UART1_TXD1_GIO_PIN);
#endif

#if 0
   /* Set aside GIO pins for CCD controller */
   gio_state.opened_pins |= (1 << CCD_WEN_GIO_PIN);
   gio_state.opened_pins |= (1 << CCD_SHUT_GIO_PIN);
#endif

  /* Make all pins controlled by the GIO driver default to input pins */
  {
    u16 dir0;

    dir0=0xffff; /* bit set - input; bit clear - output */

#ifdef CONFIG_DM270_NET
    /* Leave ETH_GIO_BIT unchanged */
    if ( (inw(DM270_GIO_DIR0) & ETH_GIO_BIT) == 0 )
      {
        dir0 &= ~ETH_GIO_BIT;
      }
#endif

#if defined(CONFIG_BLK_DEV_IDE_DM270_CFC)
# if defined(BSPCONF_DM270_TI)
    /* Leave CF_PWR_GIO_BIT unchanged */
    if ( (inw(DM270_GIO_DIR0) & CF_PWR_GIO_BIT) == 0 )
      {
        dir0 &= ~CF_PWR_GIO_BIT;
      }
# endif
#endif

    gio_state.direction_pins = 0xffff0000 | dir0;
    outw(   (gio_state.direction_pins        & 0xffff), DM270_GIO_DIR0);
    outw( ( (gio_state.direction_pins >> 16) & 0xffff), DM270_GIO_DIR1);
  }

  /* Make all pins controlled by the GIO driver default to non-inverted */
  {
    u16 invert0;

    invert0 = 0x0000; /* bit clear - noninverted; bit set - inverted */

#ifdef CONFIG_DM270_NET
    /* Leave ETH_GIO_BIT unchanged */
    if ( (inw(DM270_GIO_INV0) & ETH_GIO_BIT) != 0 )
      {
        invert0 |= ETH_GIO_BIT;
      }
#endif

#if defined (CONFIG_BLK_DEV_IDE_DM270_CFC)
# if defined(BSPCONF_DM270_TI)
 /* Leave CF_PWR_GIO_BIT unchanged */
    if ( (inw(DM270_GIO_INV0) & CF_PWR_GIO_BIT) != 0 )
      {
        invert0 |= CF_PWR_GIO_BIT;
      }
# endif
#endif

    outw(invert0, DM270_GIO_INV0);
    outw(0x0000, DM270_GIO_INV1);
  }

  /* Make all pins controlled by the GIO driver default to non-interrupt */
  {
    u16 irq_value;

    irq_value =  0x0000; /* clear - normal; set - interrupt */

#ifdef CONFIG_DM270_NET
    /* Leave ETH_GIO_BIT unchanged */
    if ( (inw(DM270_GIO_IRQPORT) & ETH_GIO_BIT) != 0 )
      {
        irq_value |= ETH_GIO_BIT;
      }
#endif

#if defined (CONFIG_BLK_DEV_IDE_DM270_CFC)
# if defined(BSPCONF_DM270_TI)
    if ( (inw(DM270_GIO_IRQPORT) & CF_PWR_GIO_BIT) != 0 )
      {
        irq_value |= CF_PWR_GIO_BIT;
      }
# endif
#endif

    gio_state.interrupt_pins = irq_value;

    outw(irq_value, DM270_GIO_IRQPORT);
  }

  /* By default, all of the ports are programmed as normal GIO pins */
  {
    unsigned short fsel0 = 0;
    unsigned short fsel1 = 0;

#if defined(BSPCONF_DM270_TI)
    /* GIO23/24 pins used for UART 1 */
    fsel0 |= GIO_FSEL0_RXD1;
    fsel1 |= GIO_FSEL1_TXD1;
#endif

#if 0
    /* GIO26/27 pins used for CCD controller */
    fsel1 |= GIO_FSEL1_C_WEN;
    fsel1 |= GIO_FSEL1_CCD_SHUT;
#endif

    gio_state.function_pins = DM270_GIO_FSEL1 << 16 | DM270_GIO_FSEL0;

    outw(fsel0, DM270_GIO_FSEL0);
    outw(fsel1, DM270_GIO_FSEL1);
  }

#if defined (CONFIG_BLK_DEV_IDE_DM270_CFC)
# if defined(BSPCONF_DM270_TI)
    /* Drive GIO 14 low to power CompactFlash */
    outw(CF_PWR_GIO_BIT, DM270_GIO_BITCLR0);
# endif
#endif

#ifdef MODULE
  REGISTER_SYMTAB(&gio_symbol_table);
#endif

  return 0;
}

/************************************************************************
 * gio_dm270_exit
 ************************************************************************/

#ifdef MODULE
void gio_dm270_exit(void)
{
  unregister_chrdev(GIO_MAJOR, "gio");
}
#endif

/************************************************************************
 * Module stuff
 ************************************************************************/

/************************************************************************
 * gio_get_fileops
 ************************************************************************/

struct file_operations *gio_get_fileops(void)
{
  return &gio_fileops;
}

#ifdef MODULE
module_init(gio_dm270_init);
module_exit(gio_dm270_exit);
#endif
