/**************************************************************************
 * linux/drivers/char/wdt_dm270.c
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived from linux/drivers/char/wdt_dsc25.c
 *
 *   Copyright (C) 2002 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * (c) Copyright 2001 RidgeRun, Inc <www.ridgerun.com>
 * (c) Copyright 1996-1997 Alan Cox <alan@redhat.com>, All Rights Reserved.
 *				http://www.redhat.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *	
 * Neither Alan Cox nor CymruNet Ltd. admit liability nor provide 
 * warranty for any of this software. This material is provided 
 * "AS-IS" and at no charge.	
 *
 * (c) Copyright 1995    Alan Cox <alan@lxorguk.ukuu.org.uk>
 *
 * Release 0.08.
 *
 * Fixes
 *	Dave Gregorich	:	Modularisation and minor bugs
 *	Alan Cox	:	Added the watchdog ioctl() stuff
 *	Alan Cox	:	Fixed the reboot problem (as noted by
 *				Matt Crocker).
 *	Alan Cox	:	Added wdt= boot option
 *	Alan Cox	:	Cleaned up copy/user stuff
 *	Tim Hockin	:	Added insmod parameters, comment cleanup
 *				Parameterized timeout
 *	Tigran Aivazian	:	Restructured wdt_init() to handle failures
 *      Gordon McNutt   :       Leveraged for TI watchdog timers
 *      Gregory Nutt    :	Leveraged to 2.4.x Linux; Made watchdog
 *				independent of timer module.
 *	Gregory Nutt	:	Leveraged to the DSC25 watchdog from the 
 *				Industrial Computer Source WDT500/501 driver
 *	Gregory Nutt	:	Leveraged to the DM270 watchdog.
 **************************************************************************/

/* TMS320DM270 supports a watch dog timer in order to detect when the
 * system is operating out of control.  If the watch dog timer is reset
 * within the set time, interrupt or reset is generated in ARM.  There is
 * no functional difference between the DSC21 and the DM270 watchdog timer
 * modules.
 */

/**************************************************************************
 * Conditional Compilation
 **************************************************************************/

/* Enables debug output via printk */

/* #define WDT_DEBUG 1 */

/* Determines default behavior of watchdog:  Interrupt or reset */

#define CONFIG_WATCHDOG_RESET

/**************************************************************************
 * Included Files
 **************************************************************************/


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/notifier.h>

#include <asm/arch/sysdep.h> /* LINUX_20 */

#ifndef LINUX_20
#include <linux/slab.h>
#include <linux/reboot.h>
#include <linux/init.h>
#else
#include <linux/malloc.h>
#endif /* LINUX_20 */

#ifndef LINUX_20
#include <asm/uaccess.h>
#endif /* LINUX_20 */
#include <asm/io.h>
#include <asm/system.h>

#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <asm/arch/wdt_ioctl.h>

/* Use versioning if needed */

#if defined(CONFIG_MODVERSIONS)
# undef MODVERSIONS
# define MODVERSIONS
#endif

#if defined(MODVERSIONS)
# include <linux/modversions.h>
#endif

/**************************************************************************
 * Definitions
 **************************************************************************/

/* Debug definitions */

#if defined(WDT_DEBUG)

 /* define MY_KERN_DEBUG KERN_DEBUG */
# define MY_KERN_DEBUG

# define dbg(format, arg...) \
    printk(MY_KERN_DEBUG __FUNCTION__ ": " format "\n" , ## arg)

# if defined(VERBOSE)
#  define verbose(format, arg...) \
    printk(MY_KERN_DEBUG __FUNCTION__ ": " format "\n" , ## arg)
# else /* VERBOSE */
#  define verbose(format, arg...) do {} while (0)
# endif /* VERBOSE */

#else /* WDT_DEBUG */

# undef  VERBOSE
# define dbg(format, arg...) do {} while (0)
# define verbose(format, arg...) do {} while (0)

#endif /* WDT_DEBUG */

#define err(format, arg...) \
  printk(KERN_ERR __FUNCTION__ ": " format "\n" , ## arg)
#define info(format, arg...) \
  printk(KERN_INFO __FUNCTION__ ": " format "\n" , ## arg)
#define warn(format, arg...) \
  printk(KERN_WARNING __FUNCTION__ ": " format "\n" , ## arg)
#define critical(format, arg...) \
  printk(KERN_CRIT __FUNCTION__ ": " format "\n" , ## arg)

#ifdef LINUX_20
  /* register_chrdev() will not except character device major
   * numbers greater than 127 in Linux 2.0.  Use the first number
   * in the range 60-64 of Local/Experimental numbers.
   */

# define WATCHDOG_MAJOR 60
#else
  /* Use the first number in the range 240-254 of Local/Experimental
   * numbers.
   */

# define WATCHDOG_MAJOR 240
#endif /* LINUX_20 */

/**************************************************************************
 * WDTMD: Watchdog Timer Mode Register
 * Bit   Name         Description 
 * 15-3  --Reserved--
 * 2     OS           Output selection. 
 *                    0: Interrupt to ARM. 
 *                    1: Reset ARM. 
 * 1     --Reserved--
 * 0     WDTEN        Watch dog timer enable bit. 
 *                    0:     Disable 
 *                    1:     Enable
 */

#define WDTMD_RESET     0x0000
#define WDTMD_INTERRUPT 0x0004
#define WDTMD_DISABLE   0x0000
#define WDTMD_ENABLE    0x0001

/**************************************************************************
 * WDTRST: Watchdog Timer Reset Register
 * Bit   Name         Description 
 * 15-1  --Reserved--
 * 0     WDTRST       Watch Dog timer reset. When "1" is written,
 *                    the watch dog timer is reset. 
 */

#define WDTRST_RESET  0x0001

/**************************************************************************
 * WDTPRSCL: Watchdog Timer Prescaler Register
 * Bit   Name         Description 
 * 15-10 --Reserved--
 * 9-0   PRSCL        Timer performs count operation with 27MHz clock
 *                    divided by (RSCL+1)
 */

#define WDTPRSCL_MAX_PRESCALER 0x3ff

#define WDTPRSCL_MAX_USEC      353200
#define WDTPRSCL_MIN_DIVISOR   0x10000

/* Base the timer calculations on a 27MHz clock, since it is generally less 
 * than the ARM clock and is a constant to multiply/divide by.
 */

#define WDTPRSCL_CLOCK_MHZ     27
#define WDTPRSCL_CLOCK_MHZx2   2*WDTPRSCL_CLOCK_MHZ

/**************************************************************************
 * WDTDIV: Watchdog Timer Divisor Register
 * Bit   Name         Description 
 * 15-0  DIV          Timer's count is (PRSCL+1)x(DIV)
 */

/**************************************************************************
 * Private Types
 **************************************************************************/

/**************************************************************************
 * Global Function Prototypes
 **************************************************************************/

/**************************************************************************
 * Private Function Prototypes
 **************************************************************************/

#if !defined(MODULE) && !defined(LINUX_20)
static int
wdt_setup(char *str) __init;
#endif

static void
wdt_interrupt(int timer, void *dev_id, struct pt_regs *regs);

#ifdef LINUX_20
static int
wdt_llseek(struct inode *inode, struct file *file, off_t offset, int org);
#endif /* LINUX_20 */

#if defined(WDT_DEBUG)
#ifdef LINUX_20
static int
wdt_read (struct inode *inode, struct file *filep, char *buf, int count);
#else
static ssize_t
wdt_read(struct file *filep, char *buf, size_t count, loff_t *f_pos);
#endif /* LINUX_20 */
#endif /* WDT_DEBUG */

#ifdef LINUX_20
static int
wdt_write(struct inode *inode, struct file *file, const char *buf, int count);
#else /* LINUX_20 */
static ssize_t
wdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos);
#endif /* LINUX_20 */

static int
wdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	  unsigned long arg);

static int
wdt_open(struct inode *inode, struct file *file);

#ifdef LINUX_20
static void
wdt_release(struct inode *inode, struct file *file);
#else /* LINUX_20 */
static int
wdt_release(struct inode *inode, struct file *file);
#endif /* LINUX_20 */

#ifndef LINUX_20
static int
wdt_notify_sys(struct notifier_block *this, unsigned long code, void *unused);
#endif /* LINUX_20 */

/**************************************************************************
 * Private Data
 **************************************************************************/

static int wdt_is_open           = 0;
static unsigned long wdt_timeout = WDTPRSCL_MAX_USEC;

#ifdef CONFIG_WATCHDOG_RESET
static unsigned short wdtmd_mode = WDTMD_RESET;
#else
static unsigned short wdtmd_mode = WDTMD_INTERRUPT;
#endif

static const char myname[]       = "watchdog";

/* Kernel Interfaces */

static struct file_operations wdt_fops =
{
#ifdef LINUX_20
  lseek:    wdt_llseek,
#else /* LINUX_20 */
  owner:    THIS_MODULE,
  llseek:   no_llseek,
#endif /* LINUX_20 */
#if defined(WDT_DEBUG)
  read:     wdt_read,
#endif /* WDT_DEBUG  */
  write:    wdt_write,
  ioctl:    wdt_ioctl,
  open:     wdt_open,
  release:  wdt_release,
};

/* The WDT card needs to learn about soft shutdowns in order to
 * turn the timebomb registers off.
 */

#ifndef LINUX_20
static struct notifier_block wdt_notifier =
{
  wdt_notify_sys,
  NULL,
  0
};
#endif /* LINUX_20 */

/**************************************************************************
 * Inline Timer Logic
 **************************************************************************/

static inline void
wdt_disable_watchdog(void)
{
  outw(wdtmd_mode|WDTMD_DISABLE, DM270_WDTMD);
  dbg("Disabled watchdog");
}

static inline void
wdt_enable_watchdog(void)
{
  dbg("Enabling watchdog mode");
  outw(wdtmd_mode|WDTMD_ENABLE, DM270_WDTMD);
}

static inline void
wdt_reset(void)
{
  /* Reset the DM270 timer.  This will automatically reload the
   * timer to the values specified by prescalar and count registers.
   */

  outw(WDTRST_RESET, DM270_WDTRST);
}

static inline int
prescale_to_ptv(int in)
{
  if (in <= 2)
    return 0;
  if (in <= 4)
    return 1;
  if (in <= 8)
    return 2;
  if (in <= 16)
    return 3;
  if (in <= 32)
    return 4;
  if (in <= 64)
    return 5;
  if (in <= 128)
    return 6;
  /* > 128 to <= 256 */
  return 7;
}

static int
wdt_set_usec(unsigned long usec_value)
{
  /* The prescaler value will be set such that clock/prescaler is the
   * number of clock ticks per counter in WDTDIV.
   */

  unsigned long prescaler  = WDTPRSCL_MAX_PRESCALER;

  /* divisor is the number of counts until the interrupt comes.
   */

  unsigned long divisor    = 1;

  dbg("Watchdog timer usec_value=%ld", usec_value);
     
  /* Calculate a value of prescaler and divisor that will be able
   * to count to the usec_value.  It may not be exact or the best
   * possible set, but it's a quick and simple algorithm.
   *
   *    divisor max = WDTPRSCL_MIN_DIVISOR
   *    prescaler max = WDTPRSCL_MAX_PRESCALER
   */

  do
    {
      divisor = (WDTPRSCL_CLOCK_MHZx2 * usec_value) / (prescaler * 2);
      dbg("divisor=0x%lx prescaler=0x%lx", divisor, prescaler);
      if (divisor >= WDTPRSCL_MIN_DIVISOR)
	{
	  if (prescaler == WDTPRSCL_MAX_PRESCALER)
	    { 
	      /* this is the max possible ~2.5 seconds. */

	      dbg("prescaler=0x%lx too big!", prescaler);
	      return -EINVAL;
	    }
	  prescaler *= 2;
	  if (prescaler > WDTPRSCL_MAX_PRESCALER)
	    {
	      prescaler = WDTPRSCL_MAX_PRESCALER;
	    }
	}
    }
  while (divisor >= WDTPRSCL_MIN_DIVISOR);

  wdt_disable_watchdog();
  wdt_reset();

  dbg("WDTDIV=0x%04lx\n", divisor);
  outw(divisor, DM270_WDTDIV);

  dbg("WDTPRSCL=0x%04lx\n", prescaler - 1);
  outw(prescaler - 1, DM270_WDTPRSCL);
         
  wdt_enable_watchdog();
  return 0;
}

/**************************************************************************
 * Private Functions
 **************************************************************************/

/**************************************************************************
 * wdt_setup:
 * @str: command line string
 *
 * Setup options. N/A.
 **************************************************************************/
 
#if !defined(MODULE) && !defined(LINUX_20)
static int __init
wdt_setup(char *str)
{
  return 1;
}

__setup("wdt=", wdt_setup);

#endif /* !MODULE */

/**************************************************************************
 * wdt_interrupt:
 * @timer:		Timer number
 * @dev_id:	Unused as we don't allow multiple devices.
 *
 * Handle an interrupt from the timer. This only happens when the timer
 *      is configured in 'interrupt' instead of 'reset' mode (which is never
 *      by this driver). I include this here as an example of how we could
 *      do something else besides h/w reset (Dan mentioned some realtime
 *      applications that might want to shutdown equipment or something 
 *      instead of just blindly resetting the MPU).
 **************************************************************************/
 
static void
wdt_interrupt(int timer, void *dev_id, struct pt_regs *regs)
{
  critical("WDT expired!");
	
#if defined(SOFTWARE_REBOOT)
#if defined(ONLY_TESTING)
  critical("Would Reboot");
#else		
  critical("Initiating system reboot");
  machine_restart(NULL);
#endif		
#else
  critical("But since the WDT is configued in interrupt "\
	   "mode -- without software reboot -- nothing will happen");
#endif		
}

/**************************************************************************
 * wdt_llseek
 *
 * Handle an lseek request.
 **************************************************************************/

#ifdef LINUX_20
static int wdt_llseek(struct inode *inode, struct file *file,
		      off_t offset, int org)
{
  return -ESPIPE;
}
#endif /* LINUX_20 */

/**************************************************************************
 * wdt_read:
 * @file: file handle to the watchdog board
 * @buf: buffer to write 1 byte into
 * @count: length of buffer
 * @ptr: offset (no seek allowed)
 *
 * Read reports the temperature in degrees Fahrenheit. The API is in
 * farenheit. It was designed by an imperial measurement luddite.
 **************************************************************************/
 
#if defined(WDT_DEBUG)
#ifdef LINUX_20
static int
wdt_read (struct inode *inode, struct file *filep, char *buf, int count)
#else
static ssize_t
wdt_read(struct file *filep, char *buf, size_t count, loff_t *f_pos)
#endif /* LINUX_20 */
{
  static int not_end = 0;
  char buffer[256];
  int len;

  dbg("filep=0x%p size=%d", filep, count);

  /* Have we returned the end of file yet? */

  if (not_end) 
    {
      /* No... next time we will. */

      not_end = 0;

      /* Return debug information. */

      dbg("WDTMD(0x%08x)=0x%04x WDTPRSCL(0x%08x)=0x%04x WDTDIV(0x%08x)=0x%04x\n",
	  DM270_WDTMD,    inw(DM270_WDTMD),
	  DM270_WDTPRSCL, inw(DM270_WDTPRSCL),
	  DM270_WDTDIV,   inw(DM270_WDTDIV));

      sprintf(buffer,
	      "WDTMD(0x%08x)=0x%04x WDTPRSCL(0x%08x)=0x%04x WDTDIV(0x%08x)=0x%04x\n",
	      DM270_WDTMD,    inw(DM270_WDTMD),
	      DM270_WDTPRSCL, inw(DM270_WDTPRSCL),
	      DM270_WDTDIV,   inw(DM270_WDTDIV));

      len = strlen(buffer);
      if (len > count) len = count;

      copy_to_user(buf, buffer, len);
    }
  else
    {
      /* Return end of file */

      not_end = 1;
      len     = 0;
    }
  return len;
}
#endif /* WDT_DEBUG */

/**************************************************************************
 * wdt_write:
 * @file: file handle to the watchdog
 * @buf: buffer to write (unused as data does not matter here 
 * @count: count of bytes
 * @ppos: pointer to the position to write. No seeks allowed
 *
 * A write to a watchdog device is defined as a keepalive signal. Any
 * write of data will do, as we we don't define content meaning.
 **************************************************************************/

#ifdef LINUX_20
static int
wdt_write(struct inode *inode, struct file *file, const char *buf, int count)
{
#else /* LINUX_20 */
static ssize_t
wdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
  /*  Can't seek (pwrite) on this device  */

  if (ppos != &file->f_pos)
    return -ESPIPE;
#endif /* LINUX_20 */

  dbg("Write call");
  if (count)
    {
      wdt_reset();
      return 1;
    }
  return 0;
}

/**************************************************************************
 * wdt_ioctl:
 * @inode: inode of the device
 * @file: file handle to the device
 * @cmd: watchdog command
 * @arg: argument pointer
 *
 * The watchdog API defines a common set of functions for all watchdogs
 * according to their available features. We only actually usefully support
 * querying capabilities and current status. 
 **************************************************************************/
 
static int
wdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	  unsigned long arg)
{
  int retval = 0;

  dbg("ioctl Call: inode=%p file=%p cmd=0x%x arg=0x%lx",
      inode, file, cmd, arg);

  /* Range check the IOCTL cmd */

  if (_IOC_TYPE(cmd) != WATCHDOG_IOCTL_BASE)
    {
      err("Invalid ioctl base value");
      return -EINVAL;
    }

  /* Process the the IOCTL command (see include/linux/watchdog.h) */

  switch(cmd)
    {
      /* The WDIOC_KEEPALIVE command will reset the watchdog timer to its
       * timeout value.  The timer must be reset periodically to keep the
       * watchdog timer from expiring. Any data accompanying the
       * WDIOC_KEEPALIVE command is ignored.
       */

    case WDIOC_KEEPALIVE:
      {
	dbg("WDIOC_KEEPALIVE");
	wdt_reset();
      }
      break;

      /* The WDIOC_SETTIMEOUT ioctl command will set the current timeout to
       * the specified timeout value.  The watchdog timer will be reset set to
       * this new timeout value immediately. This timeout value will also be
       * used on each time the watchdog is subsequently reset. The accompanyin
       *  data argument is interpreted as the new, unsigned timeout value in
       * microseconds.  The maximum,valid value for watchdog timeout value is
       * 353200 microseconds.
       */

    case WDIOC_SETTIMEOUT:
      {
	dbg("WDIOC_SETTIMEOUT, arg=%ld", arg);
	if (arg <= WDTPRSCL_MAX_USEC)
	  {
	    wdt_timeout = arg;
	    wdt_set_usec(wdt_timeout);
	  }
	else
	  {
	    dbg("Timeout is out of range");
	    return -EINVAL;
	  }
      }
      break;

      /* The WDIOC_SETMODE  ioctl command will set the current watchdog mode
       * of operations to either interrupt the ARM or to RESET the ARM. The
       * accompanying data value must be either WDT_RESET_MODE or
       * WDT_INTERRUPT_MODE.
       */

    case WDIOC_SETMODE:
      {
	dbg("WDIOC_SETMODE, arg=%ld", arg);

	if ((arg == WDT_RESET_MODE) || (arg == WDT_INTERRUPT_MODE))
	  {
	    if (wdtmd_mode == WDTMD_RESET)
	      {
		dbg("Selecting reset mode");
		wdtmd_mode = WDTMD_RESET;
	      }
	    else
	      {
		dbg("Selecting interrupt mode");
		wdtmd_mode = WDTMD_INTERRUPT;
	      }

	    /* Disable the watchdog */

	    wdt_disable_watchdog();
	    wdt_reset();

	    /* Re-enabling the watchdog will instantiate the new mode */

	    wdt_enable_watchdog();
	  }
	else
	  {
	    dbg("Invalid mode");
	    return -EINVAL;
	  }
      }
      break;

    default:
      {
	err("Invalid ioctl command");
	retval = -ENOIOCTLCMD;
      }
      break;
    }
  return retval;
}

/**************************************************************************
 * wdt_open:
 * @inode: inode of device
 * @file: file handle to device
 *
 * Our misc device has been opened. It's single open and on open we load
 *      the counter and start the countdown.
 **************************************************************************/
 
static int
wdt_open(struct inode *inode, struct file *file)
{
  dbg("Open Module");

  if(wdt_is_open)
    return -EBUSY;

  /* Activate */

  wdt_is_open=1;

  /* Load the timer with its max count and start it running. */

  wdt_set_usec(WDTPRSCL_MAX_USEC);
  MOD_INC_USE_COUNT;
  return 0;
}

/**************************************************************************
 * wdt_release:
 * @inode: inode to board
 * @file: file handle to board
 *
 * The watchdog has a configurable API. There is a religious dispute 
 * between people who want their watchdog to be able to shut down and 
 * those who want to be sure if the watchdog manager dies the machine
 * reboots. In the former case we disable the counters, in the latter
 * case you have to open it again very soon.
 **************************************************************************/
 
#ifdef LINUX_20
static void
wdt_release(struct inode *inode, struct file *file)
#else /* LINUX_20 */
static int
wdt_release(struct inode *inode, struct file *file)
#endif /* LINUX_20 */
{
  dbg("Release Module");

  wdt_disable_watchdog();
  wdt_is_open=0;

  MOD_INC_USE_COUNT;
#ifndef LINUX_20
  return 0;
#endif /* LINUX_20 */
}

/**************************************************************************
 * wdt_notify_sys:
 * @this: our notifier block
 * @code: the event being reported
 * @unused: unused
 *
 * Our notifier is called on system shutdowns. We want to turn the card
 * off at reboot otherwise the machine will reboot again during memory
 * test or worse yet during the following fsck. This would suck, in fact
 * trust me - if it happens it does suck.
 **************************************************************************/

#ifndef LINUX_20
static int
wdt_notify_sys(struct notifier_block *this, unsigned long code, void *unused)
{
  if(code==SYS_DOWN || code==SYS_HALT)
    {
      /* Turn the card off */
    }
  return NOTIFY_DONE;
}
#endif /* LINUX_20 */

/**************************************************************************
 * Public Functions
 **************************************************************************/

/**************************************************************************
 * wdt_exit
 *
 * Unload the watchdog. You cannot do this with any file handles open.
 * If your watchdog is set to continue ticking on close and you unload
 * it, well it keeps ticking. We won't get the interrupt but the board
 * will not touch PC memory so all is fine. You just have to load a new
 * module in 19 seconds or reboot.
 **************************************************************************/
 
#ifdef LINUX_20
void
wdt_exit(void)
#else /* LINUX_20 */
void __exit
wdt_exit(void)
#endif /* LINUX_20 */
{
  dbg("Unloading watchdog");

  release_region(WDT_REGISTER_BASE, WDT_BLOCK_SIZE);
#ifndef LINUX_20
  unregister_reboot_notifier(&wdt_notifier);
#endif /* LINUX_20 */
  unregister_chrdev(WATCHDOG_MAJOR, myname);
  wdt_disable_watchdog();
  free_irq(DM270_IRQ_WDT, wdt_interrupt);
}

/**************************************************************************
 * wdt_init:
 *
 * Set up the WDT watchdog board. All we have to do is grab the
 * resources we require and bitch if anyone beat us to them.
 * The open() function will actually kick the board off.
 **************************************************************************/
 
#ifdef LINUX_20
int
wdt_init(void)
#else /* LINUX_20 */
int __init
wdt_init(void)
#endif /* LINUX_20 */
{
  int ret;

  /* Announce that we are installing */

  printk("DM270 Watchdog Driver\n");

  /* Register ourselves as the watchdog device driver */

  ret = register_chrdev(WATCHDOG_MAJOR, myname, &wdt_fops);
  if (ret)
    {
      err("Cannot register on major=%d ret=%d\n", 
	  WATCHDOG_MAJOR, ret);
      goto out;
    }

  /* Initialize the watchdog timer */

  wdt_disable_watchdog();
  request_region(WDT_REGISTER_BASE, WDT_BLOCK_SIZE, myname);

  /* Register for an interrupt level callback through wdt_interrupt */

  dbg("Attaching to Watchdog Timer IRQ(%d)", DM270_IRQ_WDT);

  /* Get the IRQ */

  if (request_irq(DM270_IRQ_WDT, wdt_interrupt, SA_INTERRUPT,
		  myname, wdt_interrupt)) 
    {
      err("Unable to get IRQ for the Watchdog driver\n");
      goto outmisc;
    }

#ifndef LINUX_20
  ret = register_reboot_notifier(&wdt_notifier);
  if(ret)
    {
      err("Can't register reboot notifier (err=%d)", ret);
      goto outtimer;
    }
#endif /* LINUX_20 */

  ret = 0;

 out:
  return ret;

#ifndef LINUX_20
 outtimer:
  free_irq(DM270_IRQ_WDT, wdt_interrupt);
#endif /* LINUX_20 */

 outmisc:
  unregister_chrdev(WATCHDOG_MAJOR, myname);
  goto out;
}

module_init(wdt_init);
module_exit(wdt_exit);

#ifndef LINUX_20
EXPORT_NO_SYMBOLS;
#endif /* LINUX_20 */

