/**************************************************************************
 *
 * linux/drivers/ide/ide-dm270-cfc.c
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * linux/drivers/ide/ide-dsc25.c
 *
 *   Copyright (C) 2002 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived loosely from ide-swarm.c, so:
 *  
 *   Copyright (C) 1998 Paul Mackerras.
 *   Copyright (C) 1995-1998 Mark Lord
 *   Copyright (C) 2001 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 **************************************************************************/

/**************************************************************************
 * Included Files
 **************************************************************************/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/ide.h>

#include <asm/irq.h>
#include <asm/io.h>

#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <asm/arch/extmem.h>

/**************************************************************************
 * Conditional Compilation
 **************************************************************************/

/* Define the following to use the CF in IO mode.  Undefine the following
 * to use the CF in memory mapped mode.  IO mode will be necessary if you
 * need to receive -STSCHG interrupts.
 */

#define CFC_IO_MODE

/* To enable DEBUG mode for the DM270, define the following: */

/* #define DEBUG 1 */

/**************************************************************************
 * Definitions
 **************************************************************************/

#ifdef DEBUG
# define dbg(format, arg...) printk(format, ##arg)
#else
# define dbg(x...)
#endif

/* CFC hardware resides in external memory interface #1 (EM1). 
 * EM1 is programed to reside at DM270_CFC_BASE_ADDR.
 * Offsets 00:0000-00:07ff are data and attribute memory;
 * offsets 40:0000-40:07ff are IO space.
 */

#ifdef CFC_IO_MODE
# define CF_REGISTER_BASE       DM270_CFC_IOSPACE_ADDR
#else
# define CF_REGISTER_BASE       DM270_CFC_BASE_ADDR
#endif

/* CFC registers in attribute memory.  That registers begin at an offset
 * of 0x0200 into card memory.
 */

#define CFC_CONFIG_OPTION_REG   (DM270_CFC_BASE_ADDR + 0x00000200)
#define CFC_CONFIG_STATUS_REG   (DM270_CFC_BASE_ADDR + 0x00000202)
#define CFC_PIN_REPLACEMENT_REG (DM270_CFC_BASE_ADDR + 0x00000204)
#define CFC_SOCKET_AND_COPY_REG (DM270_CFC_BASE_ADDR + 0x00000206)

/**************************************************************************
 * Private Inline Functions
 **************************************************************************/

static inline int dm270_ide_ack_intr(ide_hwif_t *hwif)
{
  /* Set all interrupt status bits to ones (acknowledging all) */

  //outw(DM270_EXBC_BUSSTS_ALL, DM270_EXBC_BUSSTS);
  return 0;
}

/**************************************************************************
 * Public Functions
 **************************************************************************/

void __init dm270_cfc_probe(void)
{
  ide_hwif_t *hwif;
  int i;

  /* Find the first untaken slot in hwifs */

  for (i = 0; i < MAX_HWIFS; i++)
    {
      if (!ide_hwifs[i].hw.io_ports[IDE_DATA_OFFSET])
	{
	  break;
	}
    }

  if (i == MAX_HWIFS)
    {
      printk("No space for DM270 CompactFlash driver in ide_hwifs[].  Not enabled.\n");
      return;
    }

  /* Set up the DM270 register addresses */

  hwif = &ide_hwifs[i];
  hwif->hw.io_ports[IDE_DATA_OFFSET]    = (CF_REGISTER_BASE + 0x00);
  hwif->hw.io_ports[IDE_ERROR_OFFSET]   = (CF_REGISTER_BASE + 0x01);
  hwif->hw.io_ports[IDE_NSECTOR_OFFSET] = (CF_REGISTER_BASE + 0x02);
  hwif->hw.io_ports[IDE_SECTOR_OFFSET]  = (CF_REGISTER_BASE + 0x03);
  hwif->hw.io_ports[IDE_LCYL_OFFSET]    = (CF_REGISTER_BASE + 0x04);
  hwif->hw.io_ports[IDE_HCYL_OFFSET]    = (CF_REGISTER_BASE + 0x05);
  hwif->hw.io_ports[IDE_SELECT_OFFSET]  = (CF_REGISTER_BASE + 0x06);
  hwif->hw.io_ports[IDE_STATUS_OFFSET]  = (CF_REGISTER_BASE + 0x07);
  hwif->hw.io_ports[IDE_CONTROL_OFFSET] = (CF_REGISTER_BASE + 0x0e);
  hwif->hw.io_ports[IDE_IRQ_OFFSET]     = (CF_REGISTER_BASE + 0x0f);

  /* The external memory interface includes three interrupts, DM270_IRQ_EM_IF0,
   * DM270_IRQ_EM_IF1 and DM270_IRQ_EM_IF2.   DM270_IRQ_EM_IF0 is generated
   * when DMA completes. DM270_IRQ_EM_IF1 is generated when there is a
   * BUS RELEASE request from an external device. DM270_IRQ_EM_IF2 can be
   * triggered by any one of the card interrupts.
   */

  hwif->hw.ack_intr                     = dm270_ide_ack_intr;
  hwif->hw.irq                          = DM270_CFC_CFRDY_IRQ;

  /* At present, we have no special function for handling polled
   * IDE transfers.
   */

  /* hwif->ideproc                      = dm270_ideproc; */

  /* We ask the IDE subsystem to probe the CF card to determine the
   * drive characteristics.
   */

  hwif->noprobe                         = 0;

  /* Dump the initial state of CF-related registers */

  dbg("CS1CTRL1A(0x%08x)=0x%04x\n",
      DM270_EXBC_CS1CTRL1A, inw(DM270_EXBC_CS1CTRL1A));
  dbg("CS1CTRL1B(0x%08x)=0x%04x\n",
      DM270_EXBC_CS1CTRL1B, inw(DM270_EXBC_CS1CTRL1B));
  dbg("CS1CTRL2(0x%08x)=0x%04x\n",
      DM270_EXBC_CS1CTRL2, inw(DM270_EXBC_CS1CTRL2));

  dbg("DSPSTR1(0x%08x)=0x%04x\n",
      DM270_EXBC_DPSTR1, inw(DM270_EXBC_DPSTR1));
  dbg("DSPSTR2(0x%08x)=0x%04x\n",
      DM270_EXBC_DPSTR2, inw(DM270_EXBC_DPSTR2));

  dbg("BUSCTRL(0x%08x)=0x%04x\n",
      DM270_EXBC_BUSCTRL, inw(DM270_EXBC_BUSCTRL));
  dbg("BUSRLS(0x%08x)=0x%04x\n",
      DM270_EXBC_BUSRLS, inw(DM270_EXBC_BUSRLS));
  dbg("CFCTRL1(0x%08x)=0x%04x\n",
      DM270_EXBC_CFCTRL1, inw(DM270_EXBC_CFCTRL1));
  dbg("CRCTRL2(0x%08x)=0x%04x\n",
      DM270_EXBC_CFCTRL2, inw(DM270_EXBC_CFCTRL2));
  dbg("BUSINTEN(0x%08x)=0x%04x\n",
      DM270_EXBC_BUSINTEN, inw(DM270_EXBC_BUSINTEN));
  dbg("BUSSTS(0x%08x)=0x%04x\n",
      DM270_EXBC_BUSSTS, inw(DM270_EXBC_BUSSTS));
  dbg("BUSWAITMD(0x%08x)=0x%04x\n",
      DM270_EXBC_BUSWAITMD, inw(DM270_EXBC_BUSWAITMD));

  /* Select the attribute memory space (The configuration register
   * lies in attribute memory).
   */

  outw(DM270_CFC_ATTRIBUTE_MEM, DM270_EXBC_CFCTRL2);
  dbg("CFCTRL2(0x%08x)=0x%04x\n",
      DM270_EXBC_CFCTRL2, inw(DM270_EXBC_CFCTRL2));

  /* Dump the initial state of CFC registers */

  dbg("CFC Config Option Reg(0x%08x)=0x%02x\n",
      CFC_CONFIG_OPTION_REG, inw(CFC_CONFIG_OPTION_REG));
  dbg("CFC Config/Status Reg(0x%08x)=0x%02x\n",
      CFC_CONFIG_STATUS_REG, inw(CFC_CONFIG_STATUS_REG));

#ifdef CFC_IO_MODE
  /* Put the CF card in I/O interface mode by writing 0x0001 to
   * Configuration Option Register.
   */

  outw(0x0001, CFC_CONFIG_OPTION_REG);

  /* Enabled interrupts on the state change signal (STSCHG).
   * (Only works in I/O interface mode).
   */

  outw(0x0040, CFC_CONFIG_STATUS_REG);
#else
  /* Make sure that the CF card in is memory mapped mode by
   * writing 0x0000 to the Configuration Option Register.
   */

  outw(0x0000, CFC_CONFIG_OPTION_REG);
#endif

  /* Dump the current state of CFC registers */

  dbg("CFC Config Option Reg(0x%08x)=0x%02x\n",
      CFC_CONFIG_OPTION_REG, inw(CFC_CONFIG_OPTION_REG));
  dbg("CFC Config/Status Reg(0x%08x)=0x%02x\n",
      CFC_CONFIG_STATUS_REG, inw(CFC_CONFIG_STATUS_REG));
  dbg("CFC Pin Replacement Reg(0x%08x)=0x%02x\n",
      CFC_PIN_REPLACEMENT_REG, inw(CFC_PIN_REPLACEMENT_REG));
  dbg("CFC Socket and Copy Reg(0x%08x)=0x%02x\n",
      CFC_SOCKET_AND_COPY_REG, inw(CFC_SOCKET_AND_COPY_REG));

  /* Select common memory space so that the IDE registers will
   * be found (either in common memory or in IO space, depending
   * upon the CFC_IO_MODE setting).
   */

  outw(DM270_CFC_COMMON_MEM, DM270_EXBC_CFCTRL2);

  /* Set all interrupt status bits to ones (acknowledging all) */

  // outw(DM270_EXBC_BUSSTS_ALL, DM270_EXBC_BUSSTS);

  /* Interrupt enable bits are in the BUSINTEN register (0003:0a20)
   * Tell DM270 extern bus controller to enable CFRDY interrupts from
   * the IDE card.
   */

  outw(DM270_BUSINTEN_CBIEN, DM270_EXBC_BUSINTEN);

  dbg("BUSINTEN(0x%08x)=0x%04x\n",
      DM270_EXBC_BUSINTEN, inw(DM270_EXBC_BUSINTEN));
  dbg("BUSSTS(0x%08x)=0x%04x\n",
      DM270_EXBC_BUSSTS, inw(DM270_EXBC_BUSSTS));
  dbg("CF DATA(0x%08lx)=0x%04x\n",
      hwif->hw.io_ports[IDE_DATA_OFFSET],
      inw(hwif->hw.io_ports[IDE_DATA_OFFSET]));
  dbg("CF ERROR(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_ERROR_OFFSET],
      inb(hwif->hw.io_ports[IDE_ERROR_OFFSET]));
  dbg("CF NSECTOR(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_NSECTOR_OFFSET],
      inb(hwif->hw.io_ports[IDE_NSECTOR_OFFSET]));
  dbg("CF SECTOR(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_SECTOR_OFFSET],
      inb(hwif->hw.io_ports[IDE_SECTOR_OFFSET]));
  dbg("CF LCYL(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_LCYL_OFFSET],
      inb(hwif->hw.io_ports[IDE_LCYL_OFFSET]));
  dbg("CF HCYL(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_HCYL_OFFSET],
      inb(hwif->hw.io_ports[IDE_HCYL_OFFSET]));
  dbg("CF SELECT(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_SELECT_OFFSET],
      inb(hwif->hw.io_ports[IDE_SELECT_OFFSET]));
  dbg("CF STATUS(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_STATUS_OFFSET],
      inb(hwif->hw.io_ports[IDE_STATUS_OFFSET]));
  dbg("CF CONTROL(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_CONTROL_OFFSET],
      inb(hwif->hw.io_ports[IDE_CONTROL_OFFSET]));
  dbg("CF IRQ(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_IRQ_OFFSET],
      inb(hwif->hw.io_ports[IDE_IRQ_OFFSET]));

  printk("DM270 CompactFlash configured as device %i\n", i);
}

