/**************************************************************************
 *
 * linux/drivers/ide/ide-dm270-drive.c
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived from linux/drivers/ide/ide-dm270-cfc.c
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * linux/drivers/ide/ide-dsc25.c
 *
 *   Copyright (C) 2002 Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived loosely from ide-swarm.c, so:
 *  
 *   Copyright (C) 1998 Paul Mackerras.
 *   Copyright (C) 1995-1998 Mark Lord
 *   Copyright (C) 2001 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 **************************************************************************/

/**************************************************************************
 * Included Files
 **************************************************************************/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/ide.h>

#include <asm/irq.h>
#include <asm/io.h>

#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#include <asm/arch/extmem.h>

/**************************************************************************
 * Conditional Compilation
 **************************************************************************/

/* To enable DEBUG mode for the DM270, define the following: */

/* #define DEBUG 1 */

/**************************************************************************
 * Definitions
 **************************************************************************/

#ifdef DEBUG
# define dbg(format, arg...) printk(format, ##arg)
#else
# define dbg(x...)
#endif

/* IDE hardware resides in external memory interface #3 (EM3). 
 * EM3 is programed to reside at DM270_CS3_START_ADDR.  However,
 * because of logic to physical address issues, the following
 * offset values are used:
 */

#define DM270_COMMON_OFFSET   (0x00100000)
#define DM270_REGISTER_OFFSET (0x00400000 + DM270_COMMON_OFFSET)
#define DM270_CONTROL_OFFSET  (0x00800000 + DM270_COMMON_OFFSET)
#define DM270_REGISTER_BASE   (DM270_CS3_START_ADDR + DM270_REGISTER_OFFSET)
#define DM270_CONTROL_BASE    (DM270_CS3_START_ADDR + DM270_CONTROL_OFFSET)

/**************************************************************************
 * Private Inline Functions
 **************************************************************************/

static inline int dm270_ide_ack_intr(ide_hwif_t *hwif)
{
  return 0;
}
        
/**************************************************************************
 * Public Functions
 **************************************************************************/

void __init dm270_ide_probe(void)
{
  ide_hwif_t *hwif;
  int i;

  /* Find the first untaken slot in hwifs */

  for (i = 0; i < MAX_HWIFS; i++)
    {
      if (!ide_hwifs[i].hw.io_ports[IDE_DATA_OFFSET])
	{
	  break;
	}
    }

  if (i == MAX_HWIFS)
    {
      printk("No space for DM270 IDE driver in ide_hwifs[].  Not enabled.\n");
      return;
    }

  /* Set up the DM270 register addresses */

  hwif = &ide_hwifs[i];
  hwif->hw.io_ports[IDE_DATA_OFFSET]    = (DM270_REGISTER_BASE + 0x00);
  hwif->hw.io_ports[IDE_ERROR_OFFSET]   = (DM270_REGISTER_BASE + 0x02);
  hwif->hw.io_ports[IDE_NSECTOR_OFFSET] = (DM270_REGISTER_BASE + 0x04);
  hwif->hw.io_ports[IDE_SECTOR_OFFSET]  = (DM270_REGISTER_BASE + 0x06);
  hwif->hw.io_ports[IDE_LCYL_OFFSET]    = (DM270_REGISTER_BASE + 0x08);
  hwif->hw.io_ports[IDE_HCYL_OFFSET]    = (DM270_REGISTER_BASE + 0x0a);
  hwif->hw.io_ports[IDE_SELECT_OFFSET]  = (DM270_REGISTER_BASE + 0x0c);
  hwif->hw.io_ports[IDE_STATUS_OFFSET]  = (DM270_REGISTER_BASE + 0x0e);
  hwif->hw.io_ports[IDE_CONTROL_OFFSET] = (DM270_CONTROL_BASE  + 0x0c);
  hwif->hw.io_ports[IDE_IRQ_OFFSET]     = (DM270_CONTROL_BASE  + 0x0e);

  /* The IDE device will interrupt on the line associated with GIO11 */

  hwif->hw.ack_intr                     = dm270_ide_ack_intr;
  hwif->hw.irq                          = DM270_IRQ_EXB11;

  /* At present, we have no special function for handling polled
   * IDE transfers.
   */

  /* hwif->ideproc                      = dm270_ideproc; */

  /* We ask the IDE subsystem to probe the IDE interface to determine the
   * drive characteristics.
   */

  hwif->noprobe                         = 0;

#if defined(BSPCONF_DM270_INGENIENT_TOKRA)
  /* Tokra has a CF card as the master, but an IDE hard drive as
   * the slave.  Skip probe for the CF card during boot up.
   */
  #ifdef CONFIG_BLK_DEV_IDE_CF_DETECT
    hwif->drives[0].noprobe   = 1;
  #endif
  /* In any event, mark the master CF driver as removable */
  hwif->drives[0].removable = 1;
#endif

  /* Enable interrupts here... */

  dbg("IDE DATA(0x%08lx)=0x%04x\n",
      hwif->hw.io_ports[IDE_DATA_OFFSET],
      inw(hwif->hw.io_ports[IDE_DATA_OFFSET]));
  dbg("IDE ERROR(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_ERROR_OFFSET],
      inb(hwif->hw.io_ports[IDE_ERROR_OFFSET]));
  dbg("IDE NSECTOR(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_NSECTOR_OFFSET],
      inb(hwif->hw.io_ports[IDE_NSECTOR_OFFSET]));
  dbg("IDE SECTOR(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_SECTOR_OFFSET],
      inb(hwif->hw.io_ports[IDE_SECTOR_OFFSET]));
  dbg("IDE LCYL(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_LCYL_OFFSET],
      inb(hwif->hw.io_ports[IDE_LCYL_OFFSET]));
  dbg("IDE HCYL(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_HCYL_OFFSET],
      inb(hwif->hw.io_ports[IDE_HCYL_OFFSET]));
  dbg("IDE SELECT(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_SELECT_OFFSET],
      inb(hwif->hw.io_ports[IDE_SELECT_OFFSET]));
  dbg("IDE STATUS(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_STATUS_OFFSET],
      inb(hwif->hw.io_ports[IDE_STATUS_OFFSET]));
  dbg("IDE CONTROL(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_CONTROL_OFFSET],
      inb(hwif->hw.io_ports[IDE_CONTROL_OFFSET]));
  dbg("IDE IRQ(0x%08lx)=0x%02x\n",
      hwif->hw.io_ports[IDE_IRQ_OFFSET],
      inb(hwif->hw.io_ports[IDE_IRQ_OFFSET]));

  printk("DM270 IDE configured as device %i\n", i);
}

