/************************************************************************
 * linux/drivers/mtd/maps/dm270-cfi.c
 *
 *   Copyright(c) 2003, Cadenux, LLC
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived from linux/drivers/mtd/maps/dsc25-cfi.c
 *
 *   Copyright(c) 2002, Cadenux, LLC
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

/************************************************************************
 * Included Files
 ************************************************************************/

/* Use versioning if needed */

#include <linux/config.h>

#ifdef CONFIG_MODVERSIONS
# undef MODVERSIONS
# define MODVERSIONS
#endif

#ifdef MODVERSIONS
# include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/cfi.h>
#include <asm/io.h>
#include <asm/arch/memconfig.h>

/************************************************************************
 * Conditional Compilation
 ************************************************************************/

#define USE_IOREMAP 0

/* #undef DM270_DEBUG */
/* #define DM270_DEBUG */

/************************************************************************
 * Definitions
 ************************************************************************/

#ifdef DM270_DEBUG
# define dbg(format, arg...) \
  printk(KERN_DEBUG __FUNCTION__ ": " format "\n" , ##arg)
#else
# define dbg(x...)
#endif

#define info(format, arg...) \
  printk(KERN_INFO __FUNCTION__ ": " format "\n" , ##arg)
#define err(format, arg...) \
  printk(KERN_ERR __FUNCTION__ ": " format "\n" , ##arg)

#if LINUX_VERSION_CODE < 0x20212 && defined(MODULE)
# define dm270_init init_module
# define dm270_cleanup cleanup_module
#endif

/************************************************************************
 * Private Function Prototypes
 ************************************************************************/

static __u8 dm270_read8(struct map_info *map, unsigned long offset);
static __u16 dm270_read16(struct map_info *map, unsigned long offset);
static __u32 dm270_read32(struct map_info *map, unsigned long offset);
static void dm270_copy_from(struct map_info *map, void *to,
			    unsigned long from, ssize_t length);
static void dm270_write8(struct map_info *map, __u8 d,
			 unsigned long addr);
static void dm270_write16(struct map_info *map, __u16 d,
			  unsigned long addr);
static void dm270_write32(struct map_info *map, __u32 d,
			  unsigned long addr);
static void dm270_copy_to(struct map_info *map, unsigned long to,
			  const void *from, ssize_t length);

/************************************************************************
 * Private Data
 ************************************************************************/

static struct mtd_info *cfi_mtd = NULL;

static struct map_info dm270_cfi_map =
{
  name:       "DM270 CFI",
  size:       0,
  buswidth:   2,
  read8:      dm270_read8,
  read16:     dm270_read16,
  read32:     dm270_read32,
  copy_from:  dm270_copy_from,
  write8:     dm270_write8,
  write16:    dm270_write16,
  write32:    dm270_write32,
  copy_to:    dm270_copy_to,
#if !USE_IO_REMAP
  map_priv_1: BSPCONF_FLASH_BASE,
#else
  map_priv_1: 0 /* Not used */
#endif
  map_priv_2: 0 /* Not used */

};

#define NUM_CFI_PARTITIONS 4
static struct mtd_partition dm270_cfi_partitions[NUM_CFI_PARTITIONS] =
{
  {
    name:       "DM270 Bootloader",
    offset:     BSPCONF_BTLDR_FLASH_OFFSET,
    size:       BSPCONF_BTLDR_FLASH_SIZE,
    mask_flags: MTD_WRITEABLE,  /* force read-only */
  },
  {
    name:       "DM270 Bootloader Parameters",
    offset:     BSPCONF_PARAMS_FLASH_OFFSET,
    size:       BSPCONF_PARAMS_FLASH_SIZE,
    mask_flags: MTD_WRITEABLE,  /* force read-only */
  },
  {
    name:       "DM270 Kernel Image",
    offset:     BSPCONF_KERNEL_FLASH_OFFSET,
    size:       BSPCONF_KERNEL_FLASH_SIZE,
    mask_flags: MTD_WRITEABLE,  /* force read-only */
  },
  {
    name:       "DM270 File System",
    offset:     BSPCONF_FS_FLASH_OFFSET,
    size:       BSPCONF_FS_FLASH_SIZE,
  }
};

#define ROOT_FS_PARTITION 3

/************************************************************************
 * Private Functions
 ************************************************************************/

/************************************************************************
 * dm270_read8
 ************************************************************************/

static __u8 dm270_read8(struct map_info *map, unsigned long offset)
{
  register __u32 byaddr = map->map_priv_1 + offset;

  /* No special alignment requirements for bytes */

  return inb(byaddr);
}

/************************************************************************
 * dm270_read16
 ************************************************************************/

static __u16 dm270_read16(struct map_info *map, unsigned long offset)
{
  /* Check for alignment (we assume that map_priv_1 is aligned). */

  if ((offset & 1) == 0)
    {
      register __u32 waddr = map->map_priv_1 + offset;

      /* outw is guaranteed to perform an atomic 16-bit write */

      return inw(waddr);
    }
  else
    {
      __u16 retval;
      register __u8 *src  = (__u8*)(map->map_priv_1 + offset);
      register __u8 *dest = (__u8*)&retval;

      /* This should be true regardless of endian-ness.  This
       * code is awkward becasue outb is probably a macro. */

      *dest++ = inb((unsigned long)src);
      src++;
      *dest++ = inb((unsigned long)src);

      return retval;
    }
}

/************************************************************************
 * dm270_read32
 ************************************************************************/

static __u32 dm270_read32(struct map_info *map, unsigned long offset)
{
  /* Check for alignment (we assume that map_priv_1 is aligned). */

  register __u32 alignment = offset & 3;

  if (alignment == 0)
    {
      register __u32 waddr = map->map_priv_1 + offset;

      /* outw is guaranteed to perform an atomic 16-bit write */

      return inl(waddr);
    }
  /* Do we have 16-bit alignment */

  else if (alignment == 2)
    {
      __u32 retval;
      register __u16 *src  = (__u16*)(map->map_priv_1 + offset);
      register __u16 *dest = (__u16*)&retval;

      /* This should be true regardless of endian-ness.  This
       * code is awkward becasue outb is probably a macro. */

      *dest++ = inw((unsigned long)src);
      src++;
      *dest   = inw((unsigned long)src);

      return retval;
    }

  /* No, we have only byte alignment */

  else
    {
      __u32 retval;
      register __u8 *src  = (__u8*)(map->map_priv_1 + offset);
      register __u8 *dest = (__u8*)&retval;

      /* This should be true regardless of endian-ness.  This
       * code is awkward becasue outb is probably a macro. */

      *dest++ = inb((unsigned long)src);
      src++;
      *dest++ = inb((unsigned long)src);
      src++;
      *dest++ = inb((unsigned long)src);
      src++;
      *dest   = inb((unsigned long)src);

      return retval;
    }
}

/************************************************************************
 * dm270_write8
 ************************************************************************/

static void dm270_write8(struct map_info *map, __u8 d,
			 unsigned long offset)
{
  register __u32 byaddr = map->map_priv_1 + offset;

  /* No special alignment requirements for bytes */

  outb(d, byaddr);
}

/************************************************************************
 * dm270_write16
 ************************************************************************/

static void dm270_write16(struct map_info *map, __u16 d,
			  unsigned long offset)
{
  /* Check for alignment (we assume that map_priv_1 is aligned). */

  if ((offset & 1) == 0)
    {
      register __u32 waddr = map->map_priv_1 + offset;

      /* outw is guaranteed to perform an atomic 16-bit write */

      outw(d, waddr);
    }
  else
    {
      register __u8 *src  = (__u8*)&d;
      register __u8 *dest = (__u8*)(map->map_priv_1 + offset);
      register __u8 val;

      /* This should be true regardless of endian-ness.  This
       * code is awkward becasue outb is probably a macro. */

      val = *src++;
      outb(val, (unsigned long)dest);
      dest++;
      val = *src; 
      outb(val, (unsigned long)dest);
    }
}

/************************************************************************
 * dm270_write32
 ************************************************************************/

static void dm270_write32(struct map_info *map, __u32 d,
			  unsigned long offset)
{
  /* Check for alignment (we assume that map_priv_1 is aligned). */

  register __u32 alignment = offset & 3;

  if (alignment == 0)
    {
      register __u32 laddr = map->map_priv_1 + offset;

      /* outl is guaranteed to perform an atomic 16-bit write */

      outl(d, laddr);
    }

  /* Do we have 16-bit alignment */

  else if (alignment == 2)
    {
      register __u16 *src  = (__u16*)&d;
      register __u16 *dest = (__u16*)(map->map_priv_1 + offset);
      register __u16 val;

      /* This should be true regardless of endian-ness.  This
       * code is awkward becasue outw is probably a macro. */

      val = *src++;
      outw(val, (unsigned long)dest);
      dest++;
      val = *src; 
      outw(val, (unsigned long)dest);
    }

  /* No, we have only byte alignment */

  else
    {
      register __u8 *src  = (__u8*)&d;
      register __u8 *dest = (__u8*)(map->map_priv_1 + offset);
      register __u8 val;

      /* This should be true regardless of endian-ness.  This
       * code is awkward becasue outb is probably a macro. */

      val = *src++;
      outb(val, (unsigned long)dest);
      dest++;
      val = *src++;
      outb(val, (unsigned long)dest);
      dest++;
      val = *src++;
      outb(val, (unsigned long)dest);
      dest++;
      val = *src;
      outb(val, (unsigned long)dest);
    }
}

/************************************************************************
 * dm270_copy_from
 ************************************************************************/

static void dm270_copy_from(struct map_info *map, void *to,
			    unsigned long from, ssize_t length)
{
  memcpy(to, (void*)(map->map_priv_1 + from), length);
}

/************************************************************************
 * dm270_copy_to
 ************************************************************************/

static void dm270_copy_to(struct map_info *map, unsigned long to,
			  const void *from, ssize_t length)
{
  memcpy((void*)(map->map_priv_1 + to), from, length);
}

/************************************************************************
 * Public Functions
 ************************************************************************/

/************************************************************************
 * dm270_init
 ************************************************************************/

int __init dm270_init(void)
{
  int retval = 0;

#if USE_IO_REMAP
  dm270_cfi_map.map_priv_1 =
    (unsigned long)ioremap(BSPCONF_FLASH_BASE, BSPCONF_FLASH_SIZE);

  if (!dm270_cfi_map.map_priv_1)
    {
      err("Failed to ioremap() CFI flash");
      return(-EIO);
    }
#endif

  if ((cfi_mtd = do_map_probe("cfi_probe", &dm270_cfi_map)))
    {

      info("CFI FLASH device size: %dK", cfi_mtd->size >> 10);
      cfi_mtd->module = THIS_MODULE;
      retval = add_mtd_partitions(cfi_mtd, dm270_cfi_partitions,
				  NUM_CFI_PARTITIONS);
#ifndef CONFIG_BLK_DEV_INITRD
      if (!retval)
	{
	  ROOT_DEV = MKDEV(MTD_BLOCK_MAJOR, ROOT_FS_PARTITION);
	}
#endif
    }
  else
    {
      /* CFI Flash Device not found */

      dbg("CFI FLASH probe failed");
#if USE_IO_REMAP
      iounmap((void *) dm270_cfi_map.map_priv_1);
#endif
      retval = -ENXIO;
    }

  dbg("dm270_init returning %d", retval);
  return retval ;
}

/************************************************************************
 * dm270_cleanup
 ************************************************************************/

void __exit dm270_cleanup(void)
{
  if (cfi_mtd)
    {
      del_mtd_partitions(cfi_mtd);
      map_destroy(cfi_mtd);
    }

#if USE_IO_REMAP
  if (dm270_cfi_map.map_priv_1)
    {
      iounmap((void *)dm270_cfi_map.map_priv_1);
      dm270_cfi_map.map_priv_1 = 0;
    }
#endif
}

/************************************************************************
 * Module stuff
 ************************************************************************/

module_init(dm270_init);
module_exit(dm270_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Gregory Nutt <greg.nutt@cadenux.com>");
MODULE_DESCRIPTION("Cadenux DM270 CFI FLASH support");
