/*
 * FILE NAME kernel/ilatcurve.c
 *
 * BRIEF MODULE DESCRIPTION
 * This file implements a sample based interrupt latency curve for Linux.
 *
 * Author: David Singleton
 *	MontaVista Software, Inc.
 *      support@mvista.com
 *
 * Copyright 2001 MontaVista Software Inc.
 *
 * Adapted for this context by:
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <asm/system.h>
#include <asm/current.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/ilatency.h>
#ifdef CONFIG_SMP
#include <linux/smp.h>
#endif

unsigned long bucketlog[BUCKETS];
unsigned long maximum_off;
unsigned int  ilatency_start;
unsigned int  ilatency_stop;
unsigned long total_ilat_samples; /* total number of samples collected */

static int read_lines_completed = 0;

unsigned int clock_diff(unsigned int start, unsigned int stop)
{
#ifdef CONFIG_CLOCK_COUNTS_DOWN
#ifdef CONFIG_24BIT_COUNTER
  if (stop < start)
    {
      stop = 0xffffff - stop; 
    }
#endif /* CONFIG_24BIT_COUNTER */
  return (start - stop);
#else	/* CONFIG_CLOCK_COUNTS_DOWN */
#ifdef CONFIG_24BIT_COUNTER
  if (start < stop)
    {
      start = 0xffffff - start; 
    }
#endif /* CONFIG_24BIT_COUNTER */
  return (stop - start);
#endif /* CLOCK_COUNTS_DOWN */
}

void inline interrupt_overhead_start()
{
  readclock(ilatency_start);
}

void inline interrupt_overhead_stop()
{
  readclock(ilatency_stop);
}

asmlinkage void inthoff_logentry(unsigned diff)
{
  unsigned sampletime = diff;
  sampletime /= BUCKET_SIZE;
  if (maximum_off < diff)
    {
      maximum_off = diff;
    }

  if (sampletime < 0)
    {
      bucketlog[0]++;
    }
  else if (sampletime > LAST_BUCKET)
    {
      bucketlog[LAST_BUCKET]++;
    }
  else
    {
      bucketlog[sampletime]++;
    }
  total_ilat_samples++;
  return;
}

/* Writing to /proc/maxoff resets the counters */

static ssize_t
ilatcurve_write_proc(struct file * file, const char * buf,
		     size_t count, loff_t *ppos)
{
  extern interrupt_latency_log_t latency;

  total_ilat_samples = 0;
  memset(&bucketlog, 0, sizeof(unsigned long) * BUCKETS);
  memset(&latency, 0, sizeof(interrupt_latency_log_t));
  latency.logging = ON;
  return count;
}

static int ilatcurve_read_proc(char *page, char **start, off_t off,
			       int count, int *eof, void *data)
{
  int   len   = 0;
  int   i;

  /* page  = Beginning of buffer [in]
   * start = Where new data starts [out]
   * off   = File offset to obtain data [in (ignored)]
   * count = Number of bytes needed [in]
   * eof   = EOF indication [in/out]
   * data  = (not used)
   */

  if (off == 0)
    {
      /* Just been opened so display the header information 
       * also stop logging  BUGBUG: stop logging while we are 
       * reading, initialize the index numbers in the log array
       */

      unsigned int usecs;

      /* Note:  read_lines_completed is a global, we can really
       * only support one reader of the /proc "file" at a time.
       * We do nothing here to support multiple readers.
       */

      read_lines_completed = 0;
      *eof = 0;

      /* Get the same interrupt overhead time. */

      usecs = clock_to_usecs(clock_diff(ilatency_start, ilatency_stop));

      /* Show some intial information + headers */

      len += sprintf(&page[len],
		     "%lu samples logged with timer of %u ticks/us\n",
		     total_ilat_samples, (unsigned)ticks_per_usec);
      len += sprintf(&page[len],
		     "Interrupt overhead: %u us\n",
		     usecs);
      len += sprintf(&page[len],
		     "Maximum interrupt off time: %lu us\n",
		     maximum_off);
      len += sprintf(&page[len], "\nBUCKET COUNT\n");
    }

  /* Dump the sample log on the screen */

  for (i = read_lines_completed; i < BUCKETS; i++)
    {
      /* Make sure that there is space for the next entry */

      if (len + 17 > count)
	break;

      /* That last entry is a special case */

      if (i == (BUCKETS-1))
	{
	  /* Add the last entry */

	  len +=
	    sprintf(&page[len],
		    "%5u >%8lu\n",
		    (i + 1) * BUCKET_SIZE, bucketlog[LAST_BUCKET]);
	}
      else
	{
	  /* Add the next entry */

	  len +=
	    sprintf(&page[len],
		    "%5u  %8lu\n",
		    (i + 1) * BUCKET_SIZE, bucketlog[i]);
	}

      read_lines_completed++;
    }

  /* Handle end of file */

  if (i >= BUCKETS)
    {
      *eof = 1;
    }

  *start = page;
  return len;
}

int __init ilatcurve_init(void)
{
  struct proc_dir_entry *entry;

  readclock_init();
#ifndef CONFIG_PPC
  ticks_per_usec = TICKS_PER_USEC;
#endif
#ifdef CONFIG_SMP
  printk("Interrupt holdoff times are not supported on SMP systems\n");
#else
  printk("Interrupt holdoff times measurement enabled at %u ticks per usec\n",
	 ticks_per_usec);

  /* this is where each platform must turn the
   * delta 'ticks' into microseconds.  Each architecture
   * does it differently.
   */
#ifdef CONFIG_PROC_FS
  entry = create_proc_read_entry("ilatcurve", S_IWUSR | S_IRUGO, NULL,
				 ilatcurve_read_proc, 0);
  if (entry)
    {
      entry->write_proc = (write_proc_t *)ilatcurve_write_proc;
    }
#endif
#endif
  return 0;
}

__initcall(ilatcurve_init);


