/*
 * Preemption Latency Measurement Tool
 * Copyright (C) 2001 MontaVista Software Inc.
 *
 * Measures the duration of locks held in the kernel.
 * Useful for pinpointing long-held locks, which we
 * can not preempt during.
 *
 * * 20011220 Robert M. Love
 * 	- add SH arch support
 * 	- tidy things to better support new arches
 * 	- add include/asm/preem_latency.h
 * 	- now synced with 2.4.17
 * 
 * * 20011115 Robert M. Love
 * 	- resync with 2.4.15-pre4 and the latest preempt
 * 	  kernel patch
 * 
 * * 20010923 Robert M. Love
 * 	- remove causes "lowlat" and "check"
 * 	- update for revised preemption patch
 * 	- remove defines for preempt_lock_check and window
 * 
 * * 20010919 Robert M. Love
 * 	whitespace cleanup, remove unneeded code, remove
 * 	ifdefs around latency_cause, etc
 * 
 * * 20010918 Robert M. Love
 * 	update for 2.4.10-pre11 and 2.4.9-ac13
 *
 * Adapted for this context by:
 *
 *   Copyright (C) 2003 Cadenux, LLC (http://www.cadenux.com)
 *   author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 */

#include <asm/system.h>
#include <asm/current.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
//#include <linux/irq.h>
#include <asm/hw_irq.h>
#include <linux/interrupt.h>

/* We need a method of timing.  On i386 this is easy
 * with the TSC.  On other arches we can use their
 * existing timing systems or hack something together.
 *
 * The defines are:
 * 	readclock(x)      put timing value in x (unsigned int)
 * 	readclock_init()  initialize timing
 *
 * readclock must be a monotonic incremental counter.
 */

#include <asm/preem_latency.h>

#define NUM_LOG_ENTRY  20
#define NUM_COUNT       2
#define COMMAND_LENGTH 16
#define NUM_INTERRUPT  16

#define HEAD_LINES 2
#define HEAD1 (1 - HEAD_LINES)
#define HEAD2 (2 - HEAD_LINES)

#define SfS 0 /* waiting for a preempt off start1 call */
#define SfC 1 /* waiting for a preempt off end call */
#define SfM 2 /* waiting for either of the above */

extern void latency_check(const char *, unsigned, int);
extern void latency_end(const char *, unsigned);

struct point_id
{
  const char *FileName;
  unsigned Address;    
  unsigned FileLine;
  unsigned ProcId;
  char ProcName[COMMAND_LENGTH];
};

/* Worst recorded latencies */

struct log_entry
{
  unsigned        latency;
  int             cause;
  int             mask;
  struct point_id start;
  struct point_id end;
};

struct latency_data
{
  unsigned sync_flag;    /* For synchro between start and end */
  unsigned num_intrs;    /* Total number interrupts & intrs in range */
  unsigned num_count;    /* Layout */
  unsigned num_entry;
  int      least;        /* Lowest latency we care about */
  unsigned count[NUM_COUNT];
  struct log_entry  entry[NUM_LOG_ENTRY]; /* Worst recorded latencies */
  int      cause;
  int      mask;
  struct point_id latency_start;
  unsigned latency_start_count;
};

static char *cause_name[] =
{
  "Unknown",
  "spin_lock",
  "reacqBKL",
  "BKL",
  "Softirq"
};

#define CAUSE_MAX (sizeof(cause_name)/sizeof(cause_name[0]) - 1)

static unsigned log_flag   = 0; /* 0 - no logging; 1 - logging */
static unsigned panic_flag = 0; /* Set to 1 if something is really wrong */

struct latency_data latency_data[NR_CPUS] = { {0} };

static int  g_read_completed = 0;
static int  g_cpu            = 0;
static char g_buff[84];
static int  g_buffidx        = 0;
static int  g_buf_end        = 0;

/* LOCAL */
#ifdef CLOCK_COUNTS_DOWN
#define CLKDIFF(start,end) ((start) - (end))
#else
#define CLKDIFF(start,end) ((end) - (start))
#endif
/* LOCAL */

/* Strategy : 
 * 
 * Four points are identified
 *   S : the starting point right after system call, user -> kernel
 *   S1: alternative starting points 
 *      . faults (including device_not_available which does go through 
 *              error_code : label)
 *      . interrupts (all go through do_IRQ)
 *   C : need_resched is checked and schedule() is invoked if flag is set
 *   B : before RESTORE ALL and if control is returnning to user land
 *
 * For S points :
 *      it MUST follow a break point
 *      it initializes timing values for the starting point
 *
 * For S1 points :
 *      if it follows a breakpoint, treat it as an S point
 *      otherwise, ignore it.
 *
 * For C points :
 *      it must follow a C point or a S point
 *      calculate the interval and reset the starting point
 *
 * For B points :
 *      It must follow a C point or an S point
 *      calculate the internval and invalidate the starting point
 *
 */

static void reset_latency_data(void)
{
  int i;
  int cpu;

  for (cpu = 0; cpu < smp_num_cpus; cpu++)
    {
      latency_data[cpu].num_count  = NUM_COUNT;
      latency_data[cpu].num_entry  = NUM_LOG_ENTRY;
      latency_data[cpu].sync_flag  = SfS;
      latency_data[cpu].num_intrs  = 0;
      latency_data[cpu].least      = 0;

      for (i = 0; i < latency_data[cpu].num_count; 
	   latency_data[cpu].count[i++] = 0);
    }
  return;
}

asmlinkage void latency_start(const char *fname,unsigned lineno, int cause)
{
  struct latency_data *l = &latency_data[smp_processor_id()];

  /* If we are not logging or we have an error, do nothing */

  if ((log_flag == 0) || (panic_flag != 0))
    return;

  if (l->sync_flag != SfC)
    {
      readclock(l->latency_start_count);
      l->mask                   = 0;
      l->cause                  = cause;
      l->sync_flag              = SfC;
      l->latency_start.Address  = (int)__builtin_return_address(0);
      l->latency_start.FileName = fname;
      l->latency_start.FileLine = lineno;
      l->latency_start.ProcId   = current->pid;
      memcpy(l->latency_start.ProcName, current->comm, 16);
    }
}

void latency_cause(int was, int tobe)
{
  struct latency_data *l = &latency_data[smp_processor_id()];

  if (was == -100)
    {
      l->mask |= tobe;
      return;
    }

  if (l->cause == was)
    l->cause = tobe;
}

void latency_logentry(unsigned diff, const char *fname,
		unsigned lineno, unsigned Address, int mask)
{
  struct latency_data *l = &latency_data[smp_processor_id()];
  unsigned lowest = 0xffffffff;
  unsigned lowestIndex = 0;
  unsigned sameIndex = 0xffffffff;
  int i;

  if (diff < l->least) 
    return;

  /* Check if we need to log this event */

  for (i = 0; i < NUM_LOG_ENTRY; i++)
    {
      if (lowest > l->entry[i].latency)
	{
	  lowest = l->entry[i].latency;
	  lowestIndex = i;
	}

      /* If the call addresses match, it is the same path */

      if ((Address == l->entry[i].end.Address) &&
	  (l->latency_start.Address == l->entry[i].start.Address))
	{
	  sameIndex = i;
	  break;
	}
    }

  if (sameIndex == 0xffffffff)
    {
      i = lowestIndex;

      /* We would like "least" to be the smallest latency in the table
       * but we are pressed for time so we settle for it being what 
       * may have been the lowest, as this new entry may replace 
       * the found lowest with a higher value, in which case we don't 
       * have the true lowest.  Oh well, we  get closer as time 
       * passes.  The real reason for this is to speed things up, 
       * so we don't worry too much if this isn't exact.
       */

      l->least = l->entry[i].latency;
    }
  else
    {
      i = sameIndex;
    }

  if (diff > l->entry[i].latency)
    {
      l->entry[i].latency        = diff;
      l->entry[i].cause          = l->cause; 
      l->entry[i].mask           = l->mask; 

      l->entry[i].end.FileName   = fname;
      l->entry[i].end.FileLine   = lineno;
      l->entry[i].end.Address    = Address;
      l->entry[i].end.ProcId     = current->pid;

      memcpy(l->entry[i].end.ProcName,
	     current->comm, COMMAND_LENGTH);

      l->entry[i].start.FileName = l->latency_start.FileName;
      l->entry[i].start.FileLine = l->latency_start.FileLine;
      l->entry[i].start.Address  = l->latency_start.Address; 
      l->entry[i].start.ProcId   = l->latency_start.ProcId;

      memcpy(l->entry[i].start.ProcName, 
	     l->latency_start.ProcName, COMMAND_LENGTH);
    }

  l->num_intrs++;
}

/* Called at end of preemption time */

asmlinkage void latency_end(const char *fname, unsigned lineno)
{
  struct latency_data *l = &latency_data[smp_processor_id()];
  unsigned endCount;

  /* if we are not logging or we have an error, do nothing */

  if ((log_flag == 0) || (panic_flag != 0))
    {
      return;
    }

  /* read entry again */

  readclock(endCount);

  if (l->sync_flag == SfS)
    {
      l->count[SfS] ++;
      return;
    }

  /* LOCAL */

  latency_logentry(CLKDIFF(l->latency_start_count, endCount), fname, lineno,
		   (int)__builtin_return_address(0), l->mask);

  /* LOCAL */

  l->sync_flag = SfS;
  return;
}

/* Latency_check is for the end of one period and the start of another */

asmlinkage void latency_check(const char *fname, unsigned lineno, int cause)
{
  struct latency_data *l = &latency_data[smp_processor_id()];
  unsigned endCount;

  /* If we are not logging or we have an error, do nothing */

  if ((log_flag == 0) || (panic_flag != 0))
    {
      return;
    }

  /* Read entry again */

  readclock(endCount);

  if (l->sync_flag != SfS)
    {
      /* LOCAL */

      latency_logentry(CLKDIFF(l->latency_start_count, endCount), fname,
		       lineno, (int)__builtin_return_address(0), l->mask);
      /* LOCAL */
    }

  /* Re-set the starting point */

  l->sync_flag              = SfM;
  l->mask                   = 0;
  l->cause                  = cause;
  l->latency_start.Address  = (int) __builtin_return_address(0);
  l->latency_start.FileName = fname;
  l->latency_start.FileLine = lineno;
  l->latency_start.ProcId   = current->pid;

  memcpy(l->latency_start.ProcName, current->comm, 16);
  readclock(l->latency_start_count);
}

/* Note following code to read /proc file is not SMP-safe. */

static int latencytimes_read_proc(char *page, char **start, off_t off,
				  int count, int *eof, void *data)
{
  char * const base = page;
  struct latency_data *l;
  int i, j, max;

  /* page  = Beginning of buffer [in]
   * start = Where new data starts [out]
   * off   = File offset to obtain data [in (ignored)]
   * count = Number of bytes needed [in]
   * eof   = EOF indication [in/out]
   * data  = (not used)
   */

  *start = page;

  if (off == 0)
    {
      /* Just been opened */

      log_flag         = 0;  /* stop logging */
      g_read_completed = HEAD1;
      g_cpu            = 0;
      g_buffidx        = 0;
      g_buf_end        = 0;
    }
  else if ((i = g_buf_end - g_buffidx) > 0)
    {
      if (i > count)
	i = count;

      memcpy(base, &g_buff[g_buffidx], i);
      g_buffidx += i;
      return i;
    }
  else if (g_read_completed == NUM_LOG_ENTRY)
    {
      if (++g_cpu >= smp_num_cpus)
	{
	  *eof = 1;
	  reset_latency_data();
	  log_flag = 1;  /* start logging */
	  return 0;
	}

      g_read_completed = HEAD1;
    }

  g_buffidx = 0;
  l = &latency_data[g_cpu];

  switch (g_read_completed)
    {
    case HEAD1:
      g_buf_end =
	sprintf(&g_buff[0],
		"CPU %d WORST %d LATENCY TIMES OF %d MEASURED IN THIS PERIOD.\n",
		g_cpu,NUM_LOG_ENTRY,l->num_intrs);
      break;
    case HEAD2:
      g_buf_end =
	sprintf(&g_buff[0],
		"  USEC      CAUSE     MASK      START FILE:LINE   ADDRESS        END FILE:LINE\n");
      break;
    default:
      for (j = max = i = 0; j < NUM_LOG_ENTRY; j++)
	{
	  if( l->entry[j].latency > max )
	    {
	      max = l->entry[j].latency;
	      i = j;
	    }
	}
      g_buf_end =
	sprintf(&g_buff[0],
		"%6d %10s %8x %15s:%-5d %8x %15s:%d\n",
		(int)clock_to_usecs(l->entry[i].latency),
		(l->entry[i].cause == -99) ? "unknown" :
		(l->entry[i].cause < 0) ?
		irq_desc[~l->entry[i].cause].action->name :
		(l->entry[i].cause > CAUSE_MAX) ? "Unknown" :
		cause_name[l->entry[i].cause],
		l->entry[i].mask,
		l->entry[i].start.FileName == (char *)0 ?
		"entry.S" : 
		l->entry[i].start.FileName,
		l->entry[i].start.FileLine,
		l->entry[i].start.Address,
		/*l->entry[i].start.ProcId,
		  l->entry[i].start.ProcName,*/
		l->entry[i].end.FileName == (char *)0 ?
		"entry.S" : 
		l->entry[i].end.FileName,
		l->entry[i].end.FileLine
		/*,
		  l->entry[i].end.ProcId,
		  l->entry[i].end.ProcName*/);

      /* Clearing these two effectivly clears the entry */

      l->entry[i].latency = 0;
      l->entry[i].end.Address = 0;
    }

  g_read_completed++;
  if ((i = g_buf_end) > count)
    i = count;

  memcpy(base, &g_buff[0], i);

  g_buffidx += i;
  return i;
}

int __init latencytimes_init(void)
{
  readclock_init();
#ifdef CONFIG_PROC_FS
  create_proc_read_entry("latencytimes", 0, 0, latencytimes_read_proc, 0);
#endif
  reset_latency_data();
  return 0;
}

__initcall(latencytimes_init);
