/*
 * File: btldr_pi.c
 *
 * An implementation of a bootloader command processor which exposes the
 * btldr_pi.h interface. Please see btldr_pi.h for more info.
 *
 * SEE ALSO:
 *   btldr_pi.h
 *
 * Copyright (C) 2002 RidgeRun, Inc.
 * Author: RidgeRun, Inc  <skranz@ridgerun.com>
 *  - Support for DSC25 added, 9-6-02, Gregory Nutt
 *  - Support for DM270 added, 2-19-03, Gregory Nutt
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Please report all bugs/problems to the author or <support@dsplinux.net>
 *
 * key: RRGPLCR (do not remove)
 *
 */

#include "types.h"
#include "btldr_pi.h"
#include "flash.h"
#include "io.h"
#include "memconfig.h"
#include "util.h"
#if ( BSPCONF_KERNEL_COMPRESSED == 1 ) ||  ( BSPCONF_FS_COMPRESSED == 1 )
#include "decompress.h"
#endif

extern int __KernCommandLineMagicStr; // defined by linker script.
extern int __KernCommandLineOverride; // defined by linker script.
extern int __EtherMACMagicStr;        // defined by linker script.
extern int __EtherMAC;                // defined by linker script.

extern int btldr_start; // These values are determined by the
extern int btldr_end;   // the linker script when the btldr is
                        // built. This allows the bootloader to
                        // know at run time what address range
                        // it occupies. This is important for the
                        // bootloader's ability to store itself in
                        // flash.

extern unsigned int FlashFileSysStart; // These values are determined by the
extern unsigned int FlashFileSysEnd ;  // linker script and allow the user to
                                       // override the start and end boundaries
                                       // that define where the filesystem will
                                       // be stored in FLASH. This lets users,
                                       // that know exactly how big there kernel
                                       // is, to get rid of the some padding that
                                       // might normally exists between the kernel
                                       // and filesystem component bins as defined
                                       // by the comp_flash_info* arrays defined
                                       // later in this file. This is talked about
                                       // in the memory map portion of the BSP
                                       // documentation.

unsigned long FlashLockMode; // This is a global variable that will
                             // be tested by the flash*.c module before
                             // erases or writes flash. If the special
                             // code 0xBEEFFEED is not present here then
                             // the all flash mod operations will fail
                             // and report a flash locked state.

char *free_memory_start;     // used by gunzip uncompressor
int free_memory_size;        // in bytes


extern int main(int argc, char *argv[]); // Signature of the bootloader
                                         // entry point. This entry_addr
                                         // will be required whenever
                                         // the bootloader flashes itself.

typedef void(*KERN_SIG)(int r0, int r1); // Signature of the kernel entry point 
                                         // which the bootloader will be calling
                                         // whenever a "boot" command is executed.
typedef unsigned int MAGIC_t;

typedef struct {
//  MAGIC_t MAGIC_NUM;
  unsigned int START_OFFSET;
  unsigned int END_OFFSET;
} comp_flash_info_t;

// typedef unsigned short u16;
// typedef unsigned int u32;

typedef struct {
    u16 control;
    u16 mode;
    u16 count;
    u16 src_lo;
    u16 src_hi;
    u16 dest_lo;
    u16 dest_hi;
    u16 status;
} dma_reg_t;

#define MAX_DMA_SIZE 0x7FC

//#define MAGIC_DEF 0xbeefface

static comp_flash_info_t comp_flash_info[c_END] = {
  { /* c_BOOTLDR */
//    MAGIC_DEF,
    BSPCONF_BTLDR_FLASH_OFFSET,
    BSPCONF_BTLDR_FLASH_OFFSET+BSPCONF_BTLDR_FLASH_SIZE-1
  },
#if 0
  { /* c_PARAMS */
//    MAGIC_DEF,
    BSPCONF_PARAMS_FLASH_OFFSET,
    BSPCONF_PARAMS_FLASH_OFFSET+BSPCONF_PARAMS_FLASH_SIZE-1
  },
#endif
  { /* c_KERNEL */
//    MAGIC_DEF,
    BSPCONF_KERNEL_FLASH_OFFSET,
    BSPCONF_KERNEL_FLASH_OFFSET+BSPCONF_KERNEL_FLASH_SIZE-1
  },
  { /* c_FILESYS */
//    MAGIC_DEF,
    BSPCONF_FS_FLASH_OFFSET,
    BSPCONF_FS_FLASH_OFFSET+BSPCONF_FS_FLASH_SIZE-1
  },
};

typedef struct {
  unsigned int load_addr;
  unsigned int entry_addr;
  unsigned int num_bytes;
} FHEADER_t;

typedef struct {
  int is_SDRAM_resident;
  int is_FLASH_resident;
  FHEADER_t fheader;
} comp_info_t;

static comp_info_t comp_info[c_END];

//static params_t current_params;
static progress_meter_CBack_t prog_strobe;
static int strobe_cnt;
static int strobe_delta;

//#define CMDMAX 10
//static unsigned char cmd[CMDMAX];

extern void StageTwoBoardInit(void);

/******************************
 Routine:
 Description:
 ******************************/
void btldr_init(void)
{
  unsigned int i, start, end, bytes, f_type;

  io_init();
  
  f_type = flash_init(); // f_type set as either 160 or 128
                         // as underlying implementation detects
                         // the type of flash chips of the EVM.

  for (i=0; i<c_END; i++) {
    comp_info[i].is_SDRAM_resident = FALSE; // default; override below if necessary.
    comp_info[i].is_FLASH_resident = FALSE; // default; override below if necessary.
  }

  {
    // Has the user supplied filesystem location overrides
    // in the linker script? If so pick those up now.
    unsigned int FileSysStart;
    unsigned int FileSysEnd;
    FileSysStart = (unsigned int)(&FlashFileSysStart);
    FileSysEnd   = (unsigned int)(&FlashFileSysEnd);
    if (FileSysStart || FileSysEnd) {
      // Yes, at least one of them is defined to be
      // non zero which means the user wants to override
      // the normal hardcoded flash location used
      // when storing a filesytem (See rrload linker script
      // for more details).
      if (FileSysStart) {
        if (FileSysStart <= comp_flash_info[c_KERNEL].START_OFFSET) {
          util_printf("\n");
          util_printf("Error - Linker script FileSysStart too small and\n");
          util_printf("        leaves no room for storing a kernel image!\n");
          while (1) {}
        }
        else {
          comp_flash_info[c_KERNEL].END_OFFSET    = FileSysStart-1;
          comp_flash_info[c_FILESYS].START_OFFSET = FileSysStart;
        }
      }
      if (FileSysEnd) {
        comp_flash_info[c_FILESYS].END_OFFSET = FileSysEnd;
      }
    }
  }
  // Next, fill in information for the bootloader proper.
  start = (unsigned int)(&btldr_start);
  end   = (unsigned int)(&btldr_end);
  bytes = (end - start) + 1;
  comp_info[c_BOOTLDR].is_SDRAM_resident = TRUE;
  comp_info[c_BOOTLDR].fheader.load_addr = start;
  comp_info[c_BOOTLDR].fheader.num_bytes = bytes;
  comp_info[c_BOOTLDR].fheader.entry_addr =(unsigned int) &main;
  // Next, indicate that we at least have some default set of params.
//vincent  comp_info[c_PARAMS].is_SDRAM_resident = TRUE;
  // Next, Establish correct is_FLASH_resident setting for 
  // the various components.
  {
//    MAGIC_t magicn;
    comp_t comp;
    // Next, Do we have the bootloader stored?
    comp_info[c_BOOTLDR].is_FLASH_resident = TRUE; // no magic num associated
                                                   // with flash bootldr image.
                                                   // So just assign a default.
    // Next, Do we have a kernel stored?
    comp = c_KERNEL;
      comp_info[comp].is_FLASH_resident = TRUE;
    // Next, Do we have a filesystem stored?
    comp = c_FILESYS;
      comp_info[comp].is_FLASH_resident = TRUE;
  }

  // guess where some free memory is located that can be used 
  // by the gunzip uncompressor

  free_memory_start = (char *) 0x02280000;
  free_memory_size = 1024 * 1024; // 1 Mbyte
}

/******************************
 Routine:
 Description: Returns true if addr is in flash memory address space
 ******************************/

int btldr_AddressIsInFlashSpace(unsigned int addr)
{
  return( (addr >= BSPCONF_FLASH_BASE + comp_flash_info[c_BOOTLDR].START_OFFSET) &&
          (addr <= BSPCONF_FLASH_BASE + comp_flash_info[c_FILESYS].END_OFFSET) );
}


/******************************
 Routine:
 Description:
   Returns true if size is known, false otherwise
           Returns location in flash were component is stored
 ******************************/

int  get_flash_comp_addresses(comp_t c,
			      unsigned int *start_addr,
			      unsigned int *size,
			      unsigned int *max_addr)
{
  *start_addr=(unsigned int)(comp_flash_info[c].START_OFFSET) + 
              (unsigned int)(BSPCONF_FLASH_BASE);
  *max_addr=  (unsigned int)(comp_flash_info[c].END_OFFSET)   +
              (unsigned int)(BSPCONF_FLASH_BASE);
  if (comp_info[c].is_FLASH_resident) {
    *size=comp_info[c].fheader.num_bytes;
    return(true);
  } else {
    return(false);
  }
}

/******************************
 Routine:
 Description:
   Returns true if component resides in sdram, false otherwise
           Returns location in sdram were component is stored
 ******************************/

int  get_sdram_comp_addresses(comp_t c,
			      unsigned int *start_addr,
			      unsigned int *size,
			      unsigned int *entry_addr)
{
  if (comp_info[c].is_FLASH_resident) {
    *start_addr=comp_info[c].fheader.load_addr;
    *size      =comp_info[c].fheader.num_bytes;
    *entry_addr=comp_info[c].fheader.entry_addr;
    return(true);
  }

  return(false);
}

static void dma_flash_read(unsigned int offset, // Of flash.
                           unsigned int dest,   // Destination buffer
                           unsigned int num_bytes)
{
    //util_printf("\nin psuedo dma copy\n");
    //util_printf("%X %X %X\n", offset, dest, num_bytes);

    offset += BSPCONF_FLASH_BASE;

    if ((offset & 3) || (dest & 3))
    {
        flash_read(offset, (unsigned short *)dest, num_bytes);
        return;
    }
    if (num_bytes >= 24)
    {
        // Note:
        // Our objective below is to copy a chunk of bytes from a source
        // memory address to a new destination location. We'll be moving
        // 24 bytes at at time.
        // first set reg r1 == destination memory address,
        // next  set reg r2 == source memory address,
        // then  set reg r12 == num bytes to move,
        // lastly drop into a loop which copies 24 bytes at a time until
        // all bytes are copied. This will take advantage of the
        // "ldmia r2!,{r3-r8)" style assembly instruction which loads
        // in one fell swoop the six registers (r3-r8) with a total of
        // 24 continguous bytes read from the memory address reflected
        // by register r2 (and r2 is then automatically incremented by 24).
        // Now that the data registers are loaded, we can perform a
        // "stmia r1!,{r3-r8}" style instruction to move all 24
        // continuous bytes to the destination memory location reflected
        // by register r1 (and r1 is then automatically incremented by 24).
        // Finally, subtract our byte count by 24 and see if another pass
        // through the loop is needed. (I stepped through the code with
        // a debugger and that is how I know *all* the participating
        // registers used by the compiler for the algorithm below).
        num_bytes -= 24;
        asm volatile ("loop: \
                    \n ldmia   %0!,{r3 - r8} \
                    \n stmia   %1!,{r3 - r8} \
                    \n subs    %2, %2, #24 \
                    \n bge     loop"
                     : "+r" (offset), "+r" (dest), "+r" (num_bytes)
                     : "r" (offset), "r" (dest), "r" (num_bytes)
                     : "cc","r3","r4","r5","r6","r7","r8");
    }
    num_bytes +=24;
    if (num_bytes)
    {
        unsigned char *p1,*p2;

        p1 = (unsigned char *) offset;
        p2 = (unsigned char *) dest;

        while (num_bytes)
        {
            *p2++ = *p1++;
            num_bytes--;
        }
    }
}


/******************************
 Routine:
 Description:
 ******************************/
static void pull_from_flash(comp_t comp)
{
#if ( BSPCONF_KERNEL_COMPRESSED == 1 ) ||  ( BSPCONF_FS_COMPRESSED == 1 )
  char *image_start_in_flash;
#endif
  switch (comp) {
    case c_KERNEL:
    case c_FILESYS:
#if ( BSPCONF_KERNEL_COMPRESSED == 1 ) ||  ( BSPCONF_FS_COMPRESSED == 1 )
	image_start_in_flash = (char *)comp_flash_info[comp].START_OFFSET +
	                       BSPCONF_FLASH_BASE +
	                       sizeof(MAGIC_t) + sizeof(FHEADER_t);
	if (image_is_compressed(image_start_in_flash)) {
	  (void)decompress_comp(comp_info[comp].fheader.load_addr,
				free_memory_start,
				free_memory_size,
				image_start_in_flash,
				comp_info[comp].fheader.num_bytes);
	}
	else
#endif
	{

        // Next, move the image to SDRAM as per the
        // value of load_addr recorded with the image.
          if(comp == c_KERNEL){
		  comp_info[comp].fheader.load_addr=KERNEL_LOAD_ADDR;
		  comp_info[comp].fheader.entry_addr=KERNEL_ENTRY_ADDR;
		  comp_info[comp].fheader.num_bytes=KERNEL_NUM_BYTES;
	  }
	  else if(comp == c_FILESYS){
		  comp_info[comp].fheader.load_addr=FILESYS_LOAD_ADDR;
		  comp_info[comp].fheader.entry_addr=FILESYS_ENTRY_ADDR;
		  comp_info[comp].fheader.num_bytes=FILESYS_NUM_BYTES;
	  }
	  dma_flash_read(comp_flash_info[comp].START_OFFSET, 
			 comp_info[comp].fheader.load_addr,
			 comp_info[comp].fheader.num_bytes);        
//         util_printf("retrieved -- load_addr 0x%X, entry_addr 0x%X, num_bytes 0x%X\n",
//                     comp_info[comp].fheader.load_addr,
//                     comp_info[comp].fheader.entry_addr,
//                     comp_info[comp].fheader.num_bytes); // *debug* temp.

	  comp_info[comp].is_SDRAM_resident = TRUE;
      }
      break;
    case c_BOOTLDR:
      // Pull this to SDRAM, why? Ignore this request since
      // for the bootldr the very act of doing so would pull
      // bytes down on top of the very program running now,
      // presumably at that same SDRAM location. So, for this
      // particular implementation of the btldr_pi.h
      // programmer's interface, we will not support this type
      // of flash -> SDRAM copy request.
      break;
    default:
      SYSTEM_FATAL("Logic Error");
      break;
  }
}

/******************************
 Routine:
 Description:
   Note: See btldr_pi.h for description.
 ******************************/
void cmd_copy_comp(comp_t c,        //component
                   src_dest_t s,    //src
                   src_dest_t d,    //dest
                   port_format_t f, //format
                   progress_meter_CBack_t CBack,
                   int strobe_interval)
{
  // !!!!!!!!!!!
  // Note:
  // Of all the possible combinations of s and d parameters
  // only the following are currently supported by this particular
  // bootloader. Other bootloaders implementing the btldr_pi.h
  // interface may chose to support more combinations.
  // 
  //   1. copy from port to sdram
  //   2. copy from flash to sdram
  //   3. copy from sdram to flash
  //
  // !!!!!!!!!!!  
  
  prog_strobe = CBack;
  strobe_delta = strobe_interval;
  strobe_cnt = 0;
  switch (d) {
    default:
      // The d parameter is totally ignored since all
      // the currenlty supported modes can be deduced by
      // testing the s param only.
      break;
  }
  switch (s) {
    // Note: | Again, with the specified source (s) we'll assume the
    //       | destination -- independent of the dest (d) parameter.
    case sd_FLASH:
      pull_from_flash(c);
      break;
    default:
      SYSTEM_FATAL("Logic Error");
      break;
  }
}

/******************************
 Routine:
 Description:
   Note: See btldr_pi.h for description.
 ******************************/
void cmd_boot_kernel(int auto_xfer, // TRUE or FALSE,
                     progress_meter_CBack_t CBack,
                     int strobe_interval)
{
  if (auto_xfer) {
    // If the kernel is in RAM, check to see if it is compressed
    if (comp_info[c_KERNEL].is_SDRAM_resident) {
#if ( BSPCONF_KERNEL_COMPRESSED == 1 ) ||  ( BSPCONF_FS_COMPRESSED == 1 )
      if (image_is_compressed((char *)comp_info[c_KERNEL].fheader.load_addr)) {
	  util_printf("\nRAM kernel image is compressed.\n");
	  util_printf("You must first save the kernel to flash.\n\n");
	  util_printf("Press enter to return to main menu. ");
//	  util_gets(cmd,CMDMAX);
	  return;
      }
#endif
    }
    // If the kernel or file system is not in RAM and needs to be there,
    // copy them to RAM.  Data recorded with the
    // image indicates our course of action. If the
    // image is intended to be used *directly* from flash
    // then the following calls have no effect.
    else {
      cmd_copy_comp(c_KERNEL,sd_FLASH,sd_SDRAM,f_NA,CBack,strobe_interval);
    }
    if (!comp_info[c_FILESYS].is_SDRAM_resident) {
      cmd_copy_comp(c_FILESYS,sd_FLASH,sd_SDRAM,f_NA,CBack,strobe_interval);
    }
  }
  
  if (comp_info[c_KERNEL].is_SDRAM_resident ||
      comp_info[c_KERNEL].is_FLASH_resident) {
    // Next, Put our device ethernet MAC in the special memory location
    //       which the kernel will reference when initializing its
    //       ethernet driver.
    // First write a magic string that gives validity to this area of SDRAM.
    // The kernel will test validity prior to using the command line we are
    // going to deposit.
#if 0	  
    util_strcpy((unsigned char *)&__EtherMACMagicStr,"etherMAC-->");
    *((unsigned char *)&__EtherMAC) = (unsigned char) 0;
    util_strcpy((unsigned char *)&__EtherMAC,current_params.device_MAC);
#endif
    // Next, Put our kernel command line overrides in the special
    //       memory location which the kernel will reference on boot.
    // First write a magic string that gives validity to this area of SDRAM.
    // The kernel will test validity prior to using the command line we are
    // going to deposit.
#if 0    
    util_strcpy((unsigned char *)&__KernCommandLineMagicStr,"kcmdline-->");
    *((unsigned char *)&__KernCommandLineOverride) = (unsigned char) 0;
    util_strcpy((unsigned char *)&__KernCommandLineOverride,current_params.kernel_cmdline);
    util_printf("\nCommand line overrides = %s\n",(unsigned char *)&__KernCommandLineOverride);
#endif
    util_printf("Jumping to 0x%X\n",comp_info[c_KERNEL].fheader.entry_addr);
    io_delay(100); // kill some time to allow printf chars to get out
                   // the i/o port before we jump to the kernel.

    lock_flash(); // We'll keep flash unlocked the whole
                  // time the bootloader is running until
                  // it is time to transfer control to the 
                  // kernel at which time we first lock
                  // flash. Hopefully this will prevent
                  // corrupted flash that might otherwise
                  // result in a run-away program randomly
                  // jumping into the flash logic of the
                  // bootloader image still laying around
                  // in ram from a prior run.
   
	
    // Next, jump to it...
    // Note r0 and r1 values don't matter on entry.
    (*((KERN_SIG)(comp_info[c_KERNEL].fheader.entry_addr)))(0,0);
  }
}
/******************************
 Routine:
 Description:
   Note: See btldr_pi.h for description.
 ******************************/
void lock_flash(void)
{
  FlashLockMode = 0x00000000;
}

/******************************
 Routine:
 Description:
   Note: See btldr_pi.h for description.
 ******************************/
void unlock_flash(void)
{
  // Place a special pattern in a special
  // memory location. The idea here is that
  // stray program code running amock from
  // the kernel or a kernel driver that happens
  // to run itself into the flash logic of
  // bootloader (still left over in memory)
  // will not succeed in corrupting flash
  // since the low level routines will test
  // for this special pattern before modifying
  // any flash contents. In normal bootloader
  // operation the flash will remain unlocked
  // and then locked whenever it transfers
  // control to another program (e.g. the kernel)
  FlashLockMode = 0xBEEFFEED;
}
