/*! \file dm9000.c
 * \brief Ethernet device DM9000 related calls.
 *
 * Contains routines to support setting a MAC address and reading
 * the MAC address from the device's EEPROM. These functions are useful
 * only from the loader. The actual linux driver is in linux/drivers/net
 * directory.
 */
/*  Copyright  2001 Intel Corp. 
****************************************************************************
*                                                                          *
*  This software is the sole property of Intel Corporation and its         *
*  suppliers.  It is protected by the laws  and treaties of the United     *
*  States and other countries.  Unless you have entered into a license     *
*  agreement with Intel Corporation that specifically states otherwise:    *
*                                                                          *
*  (a) THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND      *
*      EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE       *
*      IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A          *
*      PARTICULAR PURPOSE;                                                 *
*  (b) no license is granted to you under any patent, copyright, or other  *
*      intellectual property right, and no ownership is transferred under  *
*      any such rights, by implication, estoppel, or otherwise; and        *
*  (c) this software may contain trade secrets of Intel Corporation and    *
*      you may not disclose or distribute this software to other parties   *
*      without the prior written consent of Intel Corporation.             *
*                                                                          *
****************************************************************************
*/
#include <stdio.h>
#include <string.h>
#include "types.h"
#include "flash.h"
#include "memconfig.h"
#include "prototype.h"
#include "io.h"

 /************************************************************************
  * Device Specific Macros 
  ************************************************************************/ 
/*! \def DM_Eth_Base_Addr
 * \brief the base address of the Ethernet chip.
 *
 * Note that the Virtual address is 0xF1000000, but we need
 * to use the physical address in the loader which is 0x10000000
 */

#define DM9000_PKT_MAX	1536	/* Received packet max size */
#define DM9000_PKT_RDY	0x01	/* Packet ready to receive */
#define DM9000_INT_MII	0x00
#define DM9000_EXT_MII	0x80
#define DM270_MACBASE	0x00110000
#define MAX_PACKET_SIZE 1514

#define DM_Eth_Base_Addr 0x07900000

/* Board/System/Debug information/definition ---------------- */
#define DM9000_ID	0x90000A46 /* The 4 bytes from the 4 ID registers */ 

#define DM9000_REG00	0x00
#define DM9000_REG05	0x30	/* SKIP_CRC/SKIP_LONG */
#define DM9000_REG08	0x27
#define DM9000_REG09	0x38
#define DM9000_REG0A	0x08
#define DM9000_REGFF	0x83	/* IMR */

#define DM9000_PHY	0x00	/* PHY address 0x00 */

#define DM9000_VID_L	0x28
#define DM9000_VID_H	0x29
#define DM9000_PID_L	0x2A
#define DM9000_PID_H	0x2B

#define DM9801_NOISE_FLOOR	0x08
#define DM9802_NOISE_FLOOR	0x05

#define DM9000_Data_Offset	4

/* The Physical Address Registers */ 
#define DM_PHY_ADDR_REG_0	    0x10
#define DM_PHY_ADDR_REG_1	    0x11
#define DM_PHY_ADDR_REG_2	    0x12
#define DM_PHY_ADDR_REG_3	    0x13
#define DM_PHY_ADDR_REG_4	    0x14
#define DM_PHY_ADDR_REG_5	    0x15
/* DM9000 Register */
#define NCR		0x00
#define NSR		0x01
#define TCR		0x02
#define RCR		0x05
#define RSR		0x06
#define GPCR		0x1e
#define GPR		0x1f
#define MRCMDX		0xf0
#define MRCMD		0xf2
#define MWCMDX		0xf6
#define MWCMD		0xf8
#define TXPLL		0xfc
#define TXPLH		0xfd
#define ISR		0xfe
#define IMR 		0xff
/* DM9000 Register End */

// typedef unsigned BOOL;

#define True	1
#define False	0

enum DM9000_PHY_mode {
	DM9000_10MHD = 0, DM9000_100MHD = 1, DM9000_10MFD = 4,
	DM9000_100MFD = 5, DM9000_AUTO = 8, DM9000_1M_HPNA =0x10 };

enum DM9000_NIC_TYPE {
	FASTETHER_NIC = 0, HOMERUN_NIC = 1, LONGRUN_NIC = 2 ,RTL8201_NIC = 3};

 /************************************************************************
  * Code 
  ************************************************************************/ 

static u8 nic_type;	/* NIC type */
static u8 nfloor = 0;
static u8 dmfe_ior(int reg);
static void dmfe_iow(int reg, u8 value);

/* Read a word from phyxcer */
static u16 phy_read(int reg)
{
	/* Fill the phyxcer register into REG_0C */
	dmfe_iow(0xc, DM9000_PHY | reg);

	dmfe_iow(0xb, 0xc); /* Issue phyxcer read command */
	io_delay(1000);	/* Wait read complete */
	dmfe_iow(0xb, 0x0); /* Clear phyxcer read command */

	/* The read data keeps on REG_0D & REG_0E */
	return ( dmfe_ior(0xe) << 8 ) | dmfe_ior(0xd);
}
/* Write a word to phyxcer */
static void phy_write(int reg, u16 value)
{
	/* Fill the phyxcer register into REG_0C */
	dmfe_iow(0xc, DM9000_PHY | reg);

	/* Fill the written data into REG_0D & REG_0E */
	dmfe_iow(0xd, (value & 0xff));
	dmfe_iow(0xe, ((value >> 8) & 0xff));

	dmfe_iow(0xb, 0xa);	/* Issue phyxcer write command */
	io_delay(500);		/* Wait write complete */
	dmfe_iow(0xb, 0x0);	/* Clear phyxcer write command */
}

/* Set PHY operationg mode */
static void set_PHY_mode(u8 mode)
{
    u16 phy_reg4 = 0x01e1, phy_reg0=0x1000;

    if ( !(mode & DM9000_AUTO) ) {
	switch(mode) {
	case DM9000_10MHD: phy_reg4 = 0x21; phy_reg0 = 0x0000; break;
	case DM9000_10MFD: phy_reg4 = 0x41; phy_reg0 = 0x1100; break;
	case DM9000_100MHD: phy_reg4 = 0x81; phy_reg0 = 0x2000; break;
	case DM9000_100MFD: phy_reg4 = 0x101; phy_reg0 =0x3100; break;
	}
    }
	phy_write(4, phy_reg4);	/* Set PHY media mode */
	phy_write(0, phy_reg0);	/*  Tmp */
}

/* Init RTL8201 */
static void program_rtl8201(void)
{
	/* To do */
}

/* Init HomeRun DM9801 */
static void program_dm9801(u16 HPNA_rev)
{
	u16 reg16, reg17, reg24, reg25;

	if ( !nfloor ) nfloor = DM9801_NOISE_FLOOR;

	reg16 = phy_read(16);
	reg17 = phy_read(17);
	reg24 = phy_read(24);
	reg25 = phy_read(25);

	switch(HPNA_rev) {
	case 0xb900: /* DM9801 E3 */
		reg16 |= 0x1000;
		reg25 = ( (reg24 + nfloor) & 0x00ff) | 0xf000;
		break;
	case 0xb901: /* DM9801 E4 */
		reg25 = ((reg24 + nfloor) & 0x00ff) | 0xc200;
		reg17 = (reg17 & 0xfff0) + nfloor + 3;
		break;
	case 0xb902: /* DM9801 E5 */
	case 0xb903: /* DM9801 E6 */
	default:
		reg16 |= 0x1000;
		reg25 = ( (reg24 + nfloor - 3) & 0x00ff) | 0xc200;
		reg17 = (reg17 & 0xfff0) + nfloor;
		break;
	}

	phy_write(16, reg16);
	phy_write(17, reg17);
	phy_write(25, reg25);
}

/* Init LongRun DM9802 */
static void program_dm9802(void)
{
	u16 reg25;

	if ( !nfloor ) nfloor = DM9802_NOISE_FLOOR;

	reg25 = phy_read(25);
	reg25 = (reg25 & 0xff00) + nfloor;
	phy_write(25, reg25);
}

/* Identify NIC type */
static void identify_nic(void)
{
	u16 phy_reg3;

	dmfe_iow(0, DM9000_EXT_MII);
	phy_reg3 = phy_read(3);

	switch(phy_reg3 & 0xfff0) {
	case 0xb900:
		if (phy_read(31) == 0x4404) {
			nic_type =  HOMERUN_NIC;
			program_dm9801(phy_reg3);
		} else {
			nic_type = LONGRUN_NIC;
			program_dm9802();
		}
		break;

	case 0x8200:
		nic_type = RTL8201_NIC;
		program_rtl8201();
		break;

	default: nic_type = FASTETHER_NIC; break;
	}

	dmfe_iow(0, DM9000_INT_MII);	
}

/*! \fn void dmfe_ior(int)
 * Performs a byte Read from the given register
 *
 * \brief Access the register with CMD low and then access the contents of the
 * register with CMD high. The address line A1 is connected to the CMD pin,
 * hence the second access will be base address plus 2.
 * To read a register location, first write the register address
 * into the io base address, then read from the io base address location
 * plus two.
 * \param int reg: the register to be read.
 * \return WORD: The contents of the register.
 */
static u8 dmfe_ior(int reg)
{
    *(volatile u8 *)DM_Eth_Base_Addr = reg;
    return *(volatile u8 *)(DM_Eth_Base_Addr + DM9000_Data_Offset);
}

static u16 dmfe_rx_data(void){

    return *(volatile u16 *)(DM_Eth_Base_Addr + DM9000_Data_Offset);
}

/*! \fn void dmfe_iow(int, u8)
 * Performs a byte write into the given register
 *
 * \brief Access the register with CMD low and then access the contents of the
 * register with CMD high. The address line A1 is connected to the CMD pin,
 * hence the second access will be base address plus 2.
 * Write the register number first into the io Address and then
 * issue a data write into io Base plus 2.
 * \param WORD reg: the register to be written
 * \param WORD val: The value to be written into the register
 * \return void: None.
 */
static void dmfe_iow(int reg, u8 value)
{
    *(volatile u8 *)DM_Eth_Base_Addr = reg;
    *(volatile u8 *)(DM_Eth_Base_Addr + DM9000_Data_Offset) = value;
}

static void dmfe_reg(int reg)
{
  *(volatile u8 *)DM_Eth_Base_Addr = reg;
}


static void dmfe_tx_data(u16 c)
{
    *(volatile u16 *)(DM_Eth_Base_Addr + DM9000_Data_Offset) = c;
}

/*! \fn dmfe_validateMacAddress(u8 *)
 * \brief Call to check if the given AC address is valid or not.
 *
 * \param u16 *addr -- Pointer to the address array.
 * \return BOOL retVal -- True for Valid and False for invalid address.
 */
BOOL dmfe_validateMacAddress(u8 *addr)
{
    u16	    i, j;
    u8	    *ptr = addr; 
    BOOL    retVal = True, macAddrSet = True;

    /* Check if it is valid address */ 
    for(i=0, j=0; i<6; i++, ptr++)
    {
	if ((*ptr == 0xFF) || (*ptr == 0))
	    j++;
	else
	{
	    macAddrSet = False;
	    break;
	}
    }
    if (macAddrSet || (j >= 5))
	retVal = False;

    return retVal;
}

/*! \fn BOOL chip_online(BOOL)
 * \brief Call to verify that it is a DM9000 ether device
 *
 * This call resets the device and makes sure that it is 
 * indeed a DM9000 part by reading the version register.
 * \return BOOL -- True if the device is found, False otherwise.
 */
int chip_online(void)
{
    u32 id_val;
    volatile u8 *iobase = (u8 *)DM_Eth_Base_Addr;
    u16 data_offset = 4; /* A1 is connected to the CMD pin */ 
//    int i;
//    u8 dev_addr[6];
//    BOOL    retVal;

    // printf("chip_online()\n");

    GpIOBitClear1(11);
    io_delay(2000);
    GpIOBitSet1(11);
    io_delay(2000);

    /* Check the DM9000 NIC ID */
    *iobase = DM9000_VID_L;
    id_val = *(iobase + data_offset);
    *iobase = DM9000_VID_L;
    id_val = *(iobase + data_offset);
    *iobase = DM9000_VID_H;
    id_val |= (*(iobase + data_offset) << 8);
    *iobase = DM9000_PID_L;
    id_val |= (*(iobase + data_offset) << 16);
    *iobase = DM9000_PID_H;
    id_val |= (*(iobase + data_offset) << 24);

    // util_printf("Ether device DM9000 got ID %x:%x\n", id_val>>16, id_val);
    if (id_val == DM9000_ID) 
    {
	// printf("<DM9000> I/O: %x, VID: %x \n", (int)iobase, (int)id_val);
	// printf("Resetting the device\n");
        dmfe_iow(0x0f,0x20);
        dmfe_iow(0x00,0x40);

	/* Reset the device */ 
	dmfe_iow(DM9000_REG00, 0x01);
	/* The reset is self clearing after 10 micro seconds */ 
	io_delay(2000); /* Delay for a while */ 
    }
    else 
    {
	util_printf("Ether device DM9000 not found %x\n", id_val);
	return False;
    }

    return True;
}

/*
 *  DM9000 initial 
 */

int dmfe_init(void){  
  int i;
  u16 oft;
  u16 hash_table[4];

  if(!chip_online())
	return False;

   dmfe_iow(NCR,0x03);/* Do twice software reset command with loopback mode */
   io_delay(20);      /* DavidCom solution */
   dmfe_iow(NCR,0x00);
   dmfe_iow(NCR,0x03);
   io_delay(20);
   dmfe_iow(NCR,0x00);

   dmfe_iow(NCR,0x01); /* Soft reset */
   io_delay(100);

   /* NIC Type: FASTETHER, HOMERUN, LONGRUN */
   identify_nic();	

   /* Set PHY */
   set_PHY_mode(DM9000_AUTO);

  dmfe_iow(NCR,0x80);

  //dm9000_hash_table();
  hash_table[0]=0x0;
  hash_table[1]=0x0;
  hash_table[2]=0x0;
  hash_table[3]=0x8000;

  for (i = 0, oft = 0x16; i < 4; i++) {
	dmfe_iow(oft++, hash_table[i] & 0xff);
	dmfe_iow(oft++, (hash_table[i] >> 8) & 0xff);
  }

  dmfe_iow(TCR,0); /* Tx Polling clear */
  dmfe_iow(0x08,0x3f);
  dmfe_iow(0x09,0x38);
  dmfe_iow(0x0a,0x08);
  dmfe_iow(0x2f,0);    /* Special Mode */
  dmfe_iow(NSR,0x2c);	/* Clear TX status */
  dmfe_iow(ISR,0x0f); /* Clear interrupt status */
  dmfe_iow(RCR,0x31); // RX enable
  dmfe_iow(IMR,DM9000_REGFF); //Disable Int

  return True;
}

/*
 * DM9000 Transmit function
 */
	
void dmfe_transmit(u16 *data,u16 len){
	int i;

	dmfe_reg(MWCMD);

    /* Disable all interrupt */
    // dmfe_iow(0xff,0x80);
	for (i=0;i< (len+1)/2 ;i++)
		dmfe_tx_data(*(data+i)); 

	dmfe_iow(TXPLL,len&0xff);
	dmfe_iow(TXPLH,((len&0xff00)>>8)&0xff);
	dmfe_iow(TCR,0x01);
    /* Enable interrupt mask */
    // dmfe_iow(0xff,0x83);
	dmfe_iow(ISR,dmfe_ior(0xfe));
}

/*
 * DM9000 Receive function
 */

void dmfe_receive(u16 *data, u16 *len) {
  	volatile u16 status;
  	u16 length=0;

 	while (1) {
   // Get the next frame off the wire which has a MAC dest that matches our MAC.
//    	dmfe_iow(ISR,dmfe_ior(ISR));	
    	status = dmfe_ior(MRCMDX); //dummy read
    	status = dmfe_ior(MRCMDX); 
	
    	if ( (status & 0xff) == 1) //packet ready
    	{
      		int i;
 
      		dmfe_reg(MRCMD);
      		status=dmfe_rx_data();	
      		length=dmfe_rx_data();	

      		if((status & 0x1f00) == 0) //status ok
		{
			for (i=0;i<(length+1)/2; i++)
				*(data+i)=dmfe_rx_data();
			*len=length;
			break;	
		}
		else //error packet
		{
#if debug	
    		util_printf("Packet error %x\n",status);
#endif
		for (i=0;i<(length)/2; i++)
			dmfe_rx_data(); //free error packet
	}
    }
    else //packet not ready
    {
#if debug 
    	printf("Packet not Ready%X\n",status);
#endif 
	*len=0;
	break;	
    }				
  } //end of while
}

BOOL dmfe_GetMACAddr(u8 *addr)
{
    flash_read((u32) FLASH_MAC_OFFSET, (u16*)addr, 6);
#if 0
    int i, len = 6; // 6 bytes of MAC address
    u8  *ptr = addr; 

    for (i = 0; i < len; i++, ptr++) 
        *ptr = dmfe_ior(DM_PHY_ADDR_REG_0 + i);
#endif
    return 0;
}

BOOL dmfe_SetMACAddr(u8 *macAddr)
{
    int	    i;

    for (i = 0; i < 6; i++) 
        dmfe_iow(DM_PHY_ADDR_REG_0 + i, macAddr[i]);

    return 0;
}

