/***********************************************************************
 * File: xflat/xflat.h
 *
 * Created 2002, Copyright (C) 2002 Cadenux, LLC.
 * Author: Gregory Nutt <greg.nutt@cadenux.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please report all bugs/problems to the author or <info@cadenux.com>
 ***********************************************************************/

#ifndef _XFLAT_H
#define _XFLAT_H

/***********************************************************************
 * Maximum Sizes
 ***********************************************************************/

/* To simplify the xFLT implementations, fixed maximum sizes of certain
 * things are hard-coded.  This eliminates certain memory management
 * operations that are necessary by the xFLT loader.
 */

#define XFLAT_MAX_STRING_SIZE 128 /* Largest size of string (incl. Z) */
#define XFLAT_MAXLIB_PATHES     8 /* Maximum number of search pathes  */
#define XFLAT_MAXLIB_NAMES     16 /* Maximum number of shared libraries
				   * per program. */

/***********************************************************************
 * The FLAT file header
 *
 * The elements within this structure are stored in network order (i.e.,
 * ntohs() and ntohl() should be used to access fields within the
 * header.
 ***********************************************************************/

struct xflat_hdr
{
  /* The "magic number identifying the FLT type.  For this FLT variant
   * this field should contain "xFLT"
   */

  char magic[4];

  /* Object file type and FLT revision number number. */

  unsigned short rev;
  unsigned short type;

  /* The following fields provide the memory map for the flat binary.
   *
   * entry      - Offset to the the first executable insruction from
   *              the beginning of the file.  For a program this is
   *              normally the entry point, _start.  For a shared
   *              library, this field may be NULL.  If it is non-NULL
   *              then it refers to the library initialization
   *              entry point and will be called by the loader after
   *              the library has been loaded.
   * data_start - Offset to the beginning of the data segment from
   *              the beginning of the file.  This field can also
   *              interpreted as the size of the ISpace segment.
   * data_end   - Offset to the end of the data segment from the
   *              beginning of  the file.
   * bss_end    - Offset to the end of bss segment from the beginning
   *              of the file.
   *
   * The text segment can be considered to be the contiguous (unrelocated)
   * address space range from address zero through (but not including)
   * data_start.
   *
   * The size of the data/bss segment includes (as a minimum) the data
   * and bss regions (bss_end - data_start) as well as the size of the
   * stack.  At run time, this region will also include program arguments
   * and environement variables.
   * 
   * The bss segment is data_end through bss_end.
   */

  unsigned long entry;
  unsigned long data_start;
  unsigned long data_end;
  unsigned long bss_end;

  /* Size of stack, in bytes */

  unsigned long stack_size;

  /* Relocation entries
   * reloc_start - Offset to the beginning of an array of relocation
   *               records (struct flat_reloc).  The offset is
   *               relative to the start of the file
   * reloc_count - The number of relocation records in the arry
   */

  unsigned long reloc_start;	   /* Offset of relocation records */
  unsigned long reloc_count;	   /* Number of relocation records */

  /* Loader Path
   *
   * This provides the full pathname of the xFLT loader that should
   * be used to load this file.  This name is used by binfmt_xflat
   * to start the process.  If this name is not present, the
   * binfmt_xflat will attempt to use a default loader path.
   */

  unsigned long loader;            /* Offset to loader pathname */

  /* Imported and exported symbol tables
   *
   * import_symbols - Offset to the beginning of an array of imported
   *                  symbol structures (struct flat_import).  The
   *                  import_symbols offset is relative to the
   *                  beginning of the file.  Each entry of the
   *                  array contains an unsigned long offset (again from
   *                  the beginning of the file) to the name of
   *                  a symbol string.  This string is null-terminated.
   * import_count   - The number of records in the export_symbols array.
   * export_symbols - Offset to the beginning of an array of export
   *                  symbol structures (struct flat_export).  The
   *                  import_symbols offset is relative to the
   *                  beginning of the file.  Each entry of the
   *                  array contains an unsigned long offset (again from
   *                  the beginning of the file) to the name of
   *                  a symbol string.  This string is null-terminated.
   * export_count   - The number of records in the export_symbols array.
   *
   * NOTE:  All of the arrays referenced in the header reside in the
   * the .text section.  This is possible because these arrays are
   * read-only and are only referenced by the load.  Residing in text
   * also guarantees that only one copy of the array is required.
   *
   * An exception is the import_import symbols array with will lie
   * in .data.  This array contains write-able data and must have
   * a single instance per process.  NOTE:  The string offset contained
   * within flat_import still refers to strings residing in the text
   * section.
   */

  unsigned long  import_symbols;   /* Offset to list of imported symbols */
  unsigned long  export_symbols;   /* Offset to list of exported symbols */
  unsigned short import_count;     /* Number of imported symbols */
  unsigned short export_count;     /* Number of imported symbols */

  /* Shared ibrary paths and file names.
   *
   * lib_pathes     - Offset to the beginning of an array of library
   *                  pathname offsets (unsigned long).  The lib_pathes
   *                  offset is relative to the beginning of the file
   *                  as is each pathname offset in the array.  The
   *                  pathname offset points to a null-terminated
   *                  string that provides library pathname. These
   *                  are the pathes that should be search by the
   *                  loader to locate the library files identified
   *                  by lib_names.
   * lpath_count    - The number of offsets in the lib_pathes array.
   * lib_names      - Offset to the beginning of an array of library
   *                  filename offsets (unsigned long).  The lib_names
   *                  offset is relative to the beginning of the file
   *                  as is each filename offset in the array.  The
   *                  filename offset points to a null-terminated
   *                  string that provides library filesname. These
   *                  are the shared library files that should be
   *                  loaded by the loader when the process is started.
   * lname_count    - The number of offsets in the lib_names array.
   */

  unsigned long  lib_pathes;       /* Offset to library search pathes */
  unsigned long  lib_names;        /* Offset to library names */
  unsigned short lpath_count;      /* Number of library pathes */
  unsigned short lname_count;      /* Number of library names */
};

/* Possible values for the type field: */

#define XFLAT_BINARY_TYPE_NONE    0       /* No file type */
#define XFLAT_BINARY_TYPE_REL     1       /* Relocatable file */
#define XFLAT_BINARY_TYPE_EXEC    2       /* Executable file */
#define XFLAT_BINARY_TYPE_DYN     3       /* Shared object file */
#define XFLAT_BINARY_TYPE_CORE    4       /* Core file */
#define XFLAT_BINARY_TYPE_NUM     5       /* Number of defined types */
#define XFLAT_BINARY_TYPE_LOOS    0xfe00  /* OS-specific range start */
#define XFLAT_BINARY_TYPE_HIOS    0xfeff  /* OS-specific range end */
#define XFLAT_BINARY_TYPE_LOPROC  0xff00  /* Processor-specific range start */
#define XFLAT_BINARY_TYPE_HIPROC  0xffff  /* Processor-specific range end */

/* Legal values for the version field.  */

#define XFLAT_VERSION_NONE         0      /* Invalid ELF version */
#define XFLAT_VERSION_CURRENT      1      /* Current version */
#define XFLAT_VERSION_NUM          2

/***********************************************************************
 * Entry Point Types
 ***********************************************************************/

/* The "entry" file offset points to the entry point of the module.
 * The prototype of the entry point differs with the type of the
 * module.  As typedef'ed below:
 */

typedef int  (*program_entry_t)(int argc, char **argv, char **envp);
typedef void (*library_entry_t)(void);
typedef void (*loader_entry_t) (unsigned long args);
typedef int  (*start_entry_t)(unsigned long args);

/***********************************************************************
 * Relocation types.
 *
 * The relocation records are an array of the following type.  The fields
 * in each element are stored in native machine order.
 ***********************************************************************/

#define XFLAT_RELOC_TYPE_NONE 0
#define XFLAT_RELOC_TYPE_TEXT 1
#define XFLAT_RELOC_TYPE_DATA 2
#define XFLAT_RELOC_TYPE_BSS  3
#define XFLAT_RELOC_TYPE_NUM  4

struct xflat_reloc
{
  signed   long offset : 30;
  unsigned long type   : 2; 
};

/***********************************************************************
 * Imported symbol type 
 *
 * The imported symbols are an array of the following type.  The fields
 * in each element are stored in native machine order.
 ***********************************************************************/

struct xflat_import
{
  unsigned long function_name;    /* Offset to name of imported function */
  unsigned long function_address; /* Resolved address of imported function */
  unsigned long data_segment;     /* Data segment value for imported func */
};

/***********************************************************************
 * Exported symbol type
 *
 * The exported symbols are an array of the following type.  The fields
 * in each element are stored in native machine order.
 ***********************************************************************/

struct xflat_export
{
  unsigned long function_name;    /* Offset to name of exported function */
  unsigned long function_address; /* Address of exported function */
};

#endif /* _XFLAT_H */
