/************************************************************************
 * linux/include/asm-armnommu/arch-dsc24/dsp_ioctl.h
 * DSC24 DSP Control/Interface Definitions
 *
 *   Copyright(c) 2003, Cadenux, LLC (http://www.cadenux.com)
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived from: linux/include/asm-armnommu/arch-dsc21/dsp_ioctl.h
 * DSC21 DSP Control/Interface Definitions
 *
 *   Copyright(c) 2002, Cadenux, LLC
 *   Author:  Gregory Nutt <greg.nutt@cadenux.com>
 *
 * Derived loosely from RidgeRun, Inc.'s, dspc_ioctl.h:
 *
 *   Derived from TI's DSPC.h
 *   Copyright (C) 2001, RidgeRun, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#ifndef __ASM_ARCH_DSP_IOCTL_H_
#define __ASM_ARCH_DSP_IOCTL_H_
    
/************************************************************************
 * Included Files
 ************************************************************************/

#include <asm/ioctl.h>

/************************************************************************
 * Public Types
 ************************************************************************/

/* Communication between the DSP and the ARM is performed through
 * mailboxes defined in the shared HPI memory.  Any number of ARM processes
 * may connect to mailboxes defined (assuming the some DSP logic is
 * aware of the mailbox location in the shared memory).  For the ARM
 * DSP driver, the mailbox location is defined via an ioctl call.
 */

/* Mailbox states.  All positive values are reserved for use by
 * the DSP driver.  The following enumeration describes the states of
 * a mailbox that are currently defined.
 * 
 * Mailbox error states.  Negative states in the range -1 through -4095
 * may be provided by the DSP to indicate error conditions:  instead of
 * the mailbox being set to DSPMB_COMPLETE, it may be set to an error
 * condition.  These error conditions will be reported to the caller
 * of the DSP driver's READ interface.
 *
 * A transaction consists of the following state changes:
 *
 * SENDER:   - Waits for state == DSPMB_READY, DSPMB_COMPLETE, or an
 *             error status.
 *           - Sets state == DSPMB_WAIT,
 *           - Writes data,
 *           - Sets state = DSPMB_AVAILABLE, and
 *           - Interrupts the receiver.
 * RECEIVER: - Detects DSPMB_AVAILABLE
 *           - Sets DSPMB_WAIT, 
 *           - Reads data,
 *           - Writes DSPMB_COMPLETE or an error status, and
 *           - Interrupts the sender.
 * SENDER:   - Marks DSPMB_READY
 */

enum dsp_mb_e
{
  DSPMB_NOT_INITIALIZED = 0, /* MB is not ready for use
                              * (Set by driver init logic) */
  DSPMB_READY           = 1, /* MB is initialized, but inactive
                              * (Set by sender) */
  DSPMB_AVAILABLE       = 2, /* MB contains new valid data
                              * (Set by sender) */
  DSPMB_WAIT            = 3, /* MB data is being processed
                              * (Set by receiver) */
  DSPMB_COMPLETE        = 4  /* MB transfer is complete
                              * (Set by receiver) */
};

/* Arm-centric macros associated with enum_mb_e */

#define _MAX_DSP_MB_E         DSPMB_COMPLETE
#define _MB_ERROR(m)          ((unsigned)((m)->state)>_MAX_DSP_MB_E)
#define _MB_READY(m)          ((m)->state==DSPMB_READY)
#define _MB_COMPLETE(m)       ((m)->state==DSPMB_COMPLETE)
#define _MB_AVAILABLE(m)      ((m)->state==DSPMB_AVAILABLE)
#define _MB_OUTBOX_NONFULL(m) ((_MB_COMPLETE(m))||(_MB_READY(m))||(_MB_ERROR(m)))
#define _MB_INBOX_NONEMPTY(m) _MB_AVAILABLE(m)

/* struct dsp_mb_s describes the form of mailbox as it appears in the
 * shared memory.  Note that any queue-ing logic must be provided by
 * logic outside of the mailbox structure.  Inter-processor interrupts
 * are used notify the ARM/DSP of the state changes changes (such as
 * the availability of new messages).
 */

struct dsp_mb_s
{
  signed   short state;  /* (see dsp_mb_e) */
  unsigned short used;   /* Number of valid 16-bit words in mb[] */
  unsigned short offset; /* Offset in words to first valid words */
  unsigned short size;   /* Maximum size of mb[] (# of 16-bit words) */
  unsigned short mb[1];  /* The beginning of the message buffer
			  * (Actual dimension determined by 'size') */
};

/* This structure defines one block of code or data to be loaded
 * into the DSP as part of the program load sequence.
 */

struct dsp_load_s
{
  unsigned short  to;      /* Destination address in the DSP address space */
  unsigned short *from;    /* Source address in the ARM address space */
  int             count;   /* Number of bytes to transfer. */
  int             cpu;     /* (For future multi-DSP support, set to 0) */
  int             is_data; /* 0:code space, 1:data space */
};

/* This structure defines parameters necessary to perform a remote
 * a memory clear as part of the program load sequence.
 */

struct dsp_memclr_s
{
  unsigned short  addr;    /* Start address in the DSP address space */
  int             count;   /* Number of bytes to clear. */
  int             cpu;     /* (For future multi-DSP support, set to 0) */
};

/* This structure is provided to define a mailbox to the driver.  This
 * structure is a 'contract' between the DSP logic and and ARM logic that
 * governs the communication:
 *
 * addr:   This is the address where a struct dsp_mbox_s resides in HPI
 *         shared memory.  This address is expressed in the DSP's address
 *         space.
 * size:   This is the size of the mb[] array in the 'struct dsp_mb_s'
 *         reference.  This is in units of 16-bit words.
 * wr_dir: This determines whether the mailbox will be used for DSP->ARM
 *         or for ARM->DSP communications.  The former is needed for
 *         driver 'read' operations to be successul; the latter are needed
 *         by driver 'write' operations.
 * cpu:    Reserved for future supported of multi-DSP platforms.
 */

struct dsp_mbspecifier_s
{
  unsigned short  addr;    /* Address of the mailbox in the DSP address space */
  unsigned short  size;    /* Size of the mb[] array (# of 16-bit words) */
  int             wr_dir;  /* 0:ARM reads (DSP->ARM), 1:ARM writes (ARM->DSP) */
  int             cpu;     /* (For future multi-DSP support, set to 0) */
};


/************************************************************************
 * Definitions
 ************************************************************************/

/* Major/minor numbers, 
 * (from Documentation/devices.txt local/experimental list).
 *
 * The major number specifies the DSP device driver.  The minor number
 * specifies a bi-di communication channel.  DSP_MINOR_MAX is, then the
 * the number of independent communication channels that can be supported
 * by the driver.
 */

#define DSP_CONTROL_MAJOR     120
#define DSP_MINOR_MAX           4

/* IOCTL numbers (See Documentation/ioctl_number.txt list) */

#define DSP_MAGIC_NUMBER 'D'

/* IOCTLs for loading DSP code ******************************************
 *
 * After the ARM takes DSP out of reset, the DSP will begin execution
 * at 0xFF80, a hard coded address in DSP ROM, which places the DSP into
 * IDLE 1 mode (a power down mode). This gives the ARM the opportunity
 * to download the DSP code. After it completes downloading the DSP
 * image, the ARM can send an interrupt to the DSP, which wakes it up
 * from IDLE 1 mode and jump to address 0x7F80 where it starts running
 * its application code loaded by the ARM.
 */

/* COMMAND: DSP_RESET.  This command resets the DSP.  At the conclusion
 * of the reset, the DSP will be in IDLE more ready for code download.
 *
 * NOTE:  Normal DSP driver reads and writes are prohibited after DSP_RESET
 * (or after power-up) and until the DSP_START has been called.
 *
 * ARGUMENT: A cpu number (zero for single DSP hardware).
 */

#define DSP_RESET    _IO(DSP_MAGIC_NUMBER, 0)

/* COMMAND: DSP_LOAD.  This command will load program data into DSP
 * memory while the DSP is in IDLE.  The DSP_RESET command should
 * have been issued prior to loading.
 *
 * ARGUMENTS: A reference to an instance of 'struct dsp_load_s'
 */

#define DSP_LOAD     _IO(DSP_MAGIC_NUMBER, 1)

/* COMMAND: DSP_MEMCLR.  This command will clear program data in the
 * DSP memory while the DSP is in IDLE.  The DSP_RESET command should
 * have been issued prior to loading.  This command may be used to
 * handle initializing of program BSS sections.
 *
 * ARGUMENTS: A reference to an instance of 'struct dsp_memclr_s'
 */

#define DSP_MEMCLR   _IO(DSP_MAGIC_NUMBER, 2)

/* COMMAND: DSP_START.  After the program has been loaded (via DSP_LOAD),
 * this command is used to start the DSP program.  This command will
 * interrupt the DSP and the DSP will branch to address 0x7F80 where the
 * new interrupt vector table is located. The ARM should have loaded this
 * location with at least a branch to the start of the new program.
 *
 * NOTE:  Normal DSP driver reads and writes are prohibited after DSP_RESET
 * (or after power-up) and until the DSP_START has been called.
 *
 * ARGUMENT: A cpu number (zero for single DSP hardware).
 */

#define DSP_START     _IO(DSP_MAGIC_NUMBER, 3)

/* IOCTLs for setting up communications  ********************************
 *
 * After code has been loaded and the DSP has been started (via DSP_START),
 * the following commands may be used to initialize for inter-processor
 * communications.
 */

/* COMMAND: DSP_MBOX.  Sets the address of a mailbox to be used by the
 * driver for reading or writing from the DSP.  This command also
 * initializes that mailbox instance for use in subsequence read() or
 * write() operations.
 *
 * ARGUMENT:  A reference to an instance of a struct dsp_mbspecifier_s
 * that describes the DSP mailbox to use.  The struct dsp_mbspecifier_s
 * contains (among other things):
 *
 *  addr:   The address of the mailbox in the DSP address space.  At this
 *          DSP address, the driver will assume the existence of a
 *          struct dsp_mb_s and will use this address for sending or
 *          receiving messages as determine by the wr_dir setting.
 *  wr_dir: If zero, the mailbox will be used for reading data sent from
 *          the DSP (only); if non-zero, the mailbox will be used for
 *          writing data to the DSP (only).
 *
 * The driver read operation will fail unless a read mailbox is specified
 * (wr_dir == 0); the write operation will fail if no write mailbox is
 * specified (wr_dir != 0).  Note that two mailboxes are required
 * to support full duplex communications; the application must make two
 * DSP_MBOX ioctl() calls for full duplex support.
 *
 * Each minor device specifies a DSP/ARM "channel."  These MBOX settings
 * apply to all users of a given "channel."
 *
 * As an implementation note, it may be necessary to define these mailboxes
 * prior to starting the DSP (via DSP_START).  This will guarantee that the
 * mailboxes are properly initialized when the DSP program runs.
 */

#define DSP_MBOX     _IO(DSP_MAGIC_NUMBER, 4)

#define DSP_IOC_MAXNR 5

#endif /* __ASM_ARCH_DSP_IOCTL_H_ */
